/*
 * finds calibration points of calibration pattern, maps 2d to 3d points and
 * determines intrinsic and extrinsic parameters using Tsai's method
 * for 3dscanner project
 *
 *		Marcel Lancelle 2002
 */

#include "CalibrateCam.h"
#include "libcamcalib.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h> // for debugging only
#include "mathext.h"

/*static*/ const int CalibrateCam::PTS_SET_NUM[CP_SET_NUM] = {20, 19}; // max value first, all values different!!
#define PTS_SET_NUM_0 20	// first value of array above (msvc doesn't eat it)

CalibrateCam::CalibrateCam() {
	// init member vars
}

void CalibrateCam::setDebugOutput(bool debugOutput) {
	this->debugOutput = debugOutput;
}

// matchCalibrationPoints and calcCamParams
bool CalibrateCam::matchAndCalcParams(MBmpImg* img, CalibrationPattern calibrationPattern) {
	if (!matchCalibrationPoints(img, calibrationPattern)) return false;
//	return calcCamParams();

	if (!calcCamParams()) return false;
camTrafo.calcTrafoMatrix();

	float worldDistUnit = 0.04;
	switch (calibrationPattern) {
		CAL_PAT_1:
			worldDistUnit = 0.02;
			break;
		CAL_PAT_2:
			worldDistUnit = 0.04;
			break;
	}
	for (int i=0; i<2; i++) for (int ptx=0; ptx<4; ptx++) for (int pty = 0; pty<5; pty++) {
		float wx, wy, wz;
		switch (i) {
			case 0:
				wx = 0;
				wy = (4-ptx) * worldDistUnit;
				wz = (pty+0.5) * worldDistUnit;
				break;
			case 1:
				wx = (ptx+1) * worldDistUnit;
				wy = 0;
				wz = (pty+0.5) * worldDistUnit;
				break;
		}
		double imX, imY;
		camTrafo.world2Img(wx*1000, wy*1000, wz*1000, &imX, &imY);
		*img->getDataXY((int) CLIP(imX, 0, img->getWidth()), (int) CLIP(imY, 0, img->getHeight())) = 255;
	}
	if (debugOutput) img->write2BmpFile("calibTest backprojection.bmp");
	return true;
}

// x, y, z: -1..1
void getHSLCoords(int h, int s, int l, float* x, float* y, float* z) {
	float norm = 1 / 255.0;
	float dist = (128-abs(l-127)) * 2 * norm * s * norm;
	*x = dist * cos(h*norm*2*PI);
	*y = dist * sin(h*norm*2*PI);
	*z = 2*norm*l - 1;
}

// finds and matches 2d image points to 3d object points
bool CalibrateCam::matchCalibrationPoints(MBmpImg* img, CalibrationPattern calibrationPattern) {
	imgWidth = img->getWidth();
	imgHeight = img->getHeight();
	if ((imgWidth < 50) || (imgHeight < 50)) return false;

	float worldDistUnit = 0.04;
	switch (calibrationPattern) {
		CAL_PAT_1:
			worldDistUnit = 0.02;
			break;
		CAL_PAT_2:
			worldDistUnit = 0.04;
			break;
	}




/*
	MBmpImg medianImg(imgWidth, imgHeight, 24);
	img->medianRGB3x3(&medianImg);
	//img->tmMedianRGB(&medianImg, 3.5);
medianImg.write2BmpFile("calibTest median.bmp");
*/
	// find black objects on white background

	MBmpImg blackObjImg(imgWidth, imgHeight, 8);
	MBmpImg blackObjImg2(imgWidth, imgHeight, 8);

////////////////////////////////////////////////////////////
/*
	// generate gray values
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) { // all pixel
			BYTE* bPtr = medianImg.getDataXY(x, y);
			int r = *bPtr++;
			int g = *bPtr++;
			int b = *bPtr;

			int max = r;
			if (g > max) max = g;
			if (b > max) max = b;
			int min = r;
			if (g < min) min = g;
			if (b < min) min = b;
		
			int l = (BYTE) ((min + max)/2);
			int dist = max-min - 40;
			if (dist < 0) dist = 0;

			*blackObjImg.getDataXY(x, y) = CLIP(l + dist, 0, 255);
		}
blackObjImg.write2BmpFile("calibTest gray.bmp");

	// calc histogram of gray img
	int histo[256], smoothHisto[256];
	blackObjImg.fillHisto(&histo[0]);

	// smooth histogram
	static const int SMOOTHING_WIDTH = 35; // difference between white and black regions
	float factors[SMOOTHING_WIDTH];
	for (int i=0; i<SMOOTHING_WIDTH; i++)
		factors[i] = (cos(2 * PI * abs(SMOOTHING_WIDTH/2 - i) / (float) SMOOTHING_WIDTH) + 1.04) / SMOOTHING_WIDTH;

	for (int i=0; i<256; i++) {
		float sum = 0;
		for (int j=0; j<SMOOTHING_WIDTH; j++) sum += histo[CLIP(i+j-SMOOTHING_WIDTH/2, 0, 255)] * factors[j];
		smoothHisto[i] = (int) sum;
	}

	// find first max, then first min
	int threshold = 0;
	int thresholdHistoVal = 0;
	while ((threshold < 250) && (smoothHisto[threshold+3] >= thresholdHistoVal)) {
		threshold++;
		thresholdHistoVal = smoothHisto[threshold];
	}
	threshold += 3;
	thresholdHistoVal = smoothHisto[threshold];
	while ((threshold < 250) && (smoothHisto[threshold+3] <= thresholdHistoVal)) {
		threshold++;
		thresholdHistoVal = smoothHisto[threshold];
	}
	threshold++;

	if (threshold > 250) return false;

	// threshold
	for (int y=0; y<imgHeight; y++) {
		for (int x=0; x<imgWidth; x++) { // all pixel
			BYTE* bPtr = blackObjImg.getDataXY(x, y);
			*bPtr = (*bPtr > threshold) ? 0 : 255;
		}
	}
*/
////////////////////////////////////////////////////////////
/*
	// clear
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) *blackObjImg.getDataXY(x, y) = 0;

	int delta = 2; // distance between white and black region in pixel (both directions)
//	int delta = 4; // distance between white and black region in pixel (both directions)
//	int wbDiff = 30; // difference between white and black regions
//	int wbDiffReset = 20; // same but for reset
	int wbDiff = 40; // difference between white and black regions
	int wbDiffReset = 35; // same but for reset
	int initVal = (imgWidth + imgHeight) / 10;

	// from left to right
	for (int y=0; y<imgHeight; y++) {
		int blObjFound = 0;
		for (int x=delta; x<imgWidth-delta; x++) { // nearly all pixel
			BYTE* vPtr = medianImg.getDataXY(x + delta, y);
			int rd = *vPtr++;
			int gd = *vPtr++;
			int bd = *vPtr;
			vPtr = medianImg.getDataXY(x - delta, y);
			rd -= *vPtr++;
			gd -= *vPtr++;
			bd -= *vPtr;

			if ((rd < -wbDiff) && (gd < -wbDiff) && (bd < -wbDiff)) blObjFound = initVal;
			if ((rd > wbDiffReset) && (gd > wbDiffReset) && (bd > wbDiffReset)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
		}
	}
	// from top to bottom
	for (int x=0; x<imgWidth; x++) {
		int blObjFound = 0;
		for (int y=delta; y<imgHeight-delta; y++) { // nearly all pixel
			BYTE* vPtr = medianImg.getDataXY(x, y + delta);
			int rd = *vPtr++;
			int gd = *vPtr++;
			int bd = *vPtr;
			vPtr = medianImg.getDataXY(x, y - delta);
			rd -= *vPtr++;
			gd -= *vPtr++;
			bd -= *vPtr;

			if ((rd < -wbDiff) && (gd < -wbDiff) && (bd < -wbDiff)) blObjFound = initVal;
			if ((rd > wbDiffReset) && (gd > wbDiffReset) && (bd > wbDiffReset)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
		}
	}
	// from right to left
	for (int y=0; y<imgHeight; y++) {
		int blObjFound = 0;
		for (int x=imgWidth-1-delta; x>=delta; x--) { // nearly all pixel
			BYTE* vPtr = medianImg.getDataXY(x - delta, y);
			int rd = *vPtr++;
			int gd = *vPtr++;
			int bd = *vPtr;
			vPtr = medianImg.getDataXY(x + delta, y);
			rd -= *vPtr++;
			gd -= *vPtr++;
			bd -= *vPtr;

			if ((rd < -wbDiff) && (gd < -wbDiff) && (bd < -wbDiff)) blObjFound = initVal;
			if ((rd > wbDiffReset) && (gd > wbDiffReset) && (bd > wbDiffReset)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
		}
	}
	// from top to bottom
	for (int x=0; x<imgWidth; x++) {
		int blObjFound = 0;
		for (int y=imgHeight-1-delta; y>=delta; y--) { // nearly all pixel
			BYTE* vPtr = medianImg.getDataXY(x, y - delta);
			int rd = *vPtr++;
			int gd = *vPtr++;
			int bd = *vPtr;
			vPtr = medianImg.getDataXY(x, y + delta);
			rd -= *vPtr++;
			gd -= *vPtr++;
			bd -= *vPtr;

			if ((rd < -wbDiff) && (gd < -wbDiff) && (bd < -wbDiff)) blObjFound = initVal;
			if ((rd > wbDiffReset) && (gd > wbDiffReset) && (bd > wbDiffReset)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
		}
	}

blackObjImg.write2BmpFile("calibTest blackobj before max.bmp");
	// generate binary img with max values
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) { // all pixel
			*blackObjImg.getDataXY(x, y) = (*blackObjImg.getDataXY(x, y) == 4);
		}

*/
////////////////////////////////////////////////////////////

	// generate gray values
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) { // all pixel
//			BYTE* bPtr = medianImg.getDataXY(x, y);
			BYTE* bPtr = img->getDataXY(x, y);
			int r = *bPtr++;
			int g = *bPtr++;
			int b = *bPtr;

			*blackObjImg2.getDataXY(x, y) = (r + g + b) / 3;
		}

	// clear
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) *blackObjImg.getDataXY(x, y) = 0;

//	static const int FACTOR_NUM = 11;
//	int factors[FACTOR_NUM] = {-1, -8, -28, -50, -43, 0, 43, 50, 28, 8, 1}; // Gau low pass filter (1 7 21 30 21 7 1) convoluted with gradient (-1 -1 0 1 1)
//	static const int FACT_MAX = 130;

	static const int FACTOR_NUM = 15;
	int factors[FACTOR_NUM] = {-1, -7, -21, -30, -21, -7, -1, 0, 1, 7, 21, 30, 21, 7, 1}; // Gau low pass filter (1 7 21 30 21 7 1) convoluted with gradient (-1 -1 0 0 0 1 1)

//--> falten mit Gau-TP, dann: bevor (Pixel -8 .. -2) min, danach (2 .. 8) max; auch nicht gut (Lokalisation!)

	static const int FACT_MAX = 88;

	static const int wbDiff = 50 * FACT_MAX; // difference between white and black regions
	static const int wbDiffReset = 40 * FACT_MAX; // same but for reset
	int initVal = (imgWidth + imgHeight) / 8; // max length of beam

	// from left to right
	for (int y=0; y<imgHeight; y++) {
		int blObjFound = 0;
		int prevV = 0;
		for (int x=0; x<imgWidth; x++) { // nearly all pixel
			int v = 0;
			for (int i=0; i<FACTOR_NUM; i++) v += *blackObjImg2.getDataXY(CLIP(x+i - FACTOR_NUM/2, 0, imgWidth-1), y) * factors[i];

			if ((-prevV > wbDiff) && (-prevV > -v)) blObjFound = initVal;
			if ((prevV > wbDiffReset) && (prevV > v)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
			prevV = v;
		}
	}

	// from top to bottom
	for (int x=0; x<imgWidth; x++) {
		int blObjFound = 0;
		int prevV = 0;
		for (int y=0; y<imgHeight; y++) { // nearly all pixel
			int v = 0;
			for (int i=0; i<FACTOR_NUM; i++) v += *blackObjImg2.getDataXY(x, CLIP(y+i - FACTOR_NUM/2, 0, imgHeight-1)) * factors[i];

			if ((-prevV > wbDiff) && (-prevV > -v)) blObjFound = initVal;
			if ((prevV > wbDiffReset) && (prevV > v)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
			prevV = v;
		}
	}
	// from right to left
	for (int y=0; y<imgHeight; y++) {
		int blObjFound = 0;
		int prevV = 0;
		for (int x=imgWidth-1; x>=0; x--) { // nearly all pixel
			int v = 0;
			for (int i=0; i<FACTOR_NUM; i++) v += *blackObjImg2.getDataXY(CLIP(x+i - FACTOR_NUM/2, 0, imgWidth-1), y) * factors[i];

			if ((prevV > wbDiff) && (prevV > v)) blObjFound = initVal;
			if ((-prevV > wbDiffReset) && (-prevV > -v)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
			prevV = v;
		}
	}
	// from top to bottom
	for (int x=0; x<imgWidth; x++) {
		int blObjFound = 0;
		int prevV = 0;
		for (int y=imgHeight-1; y>=0; y--) { // nearly all pixel
			int v = 0;
			for (int i=0; i<FACTOR_NUM; i++) v += *blackObjImg2.getDataXY(x, CLIP(y+i - FACTOR_NUM/2, 0, imgHeight-1)) * factors[i];

			if ((prevV > wbDiff) && (prevV > v)) blObjFound = initVal;
			if ((-prevV > wbDiffReset) && (-prevV > -v)) blObjFound = 0;

			if (blObjFound) {
				(*blackObjImg.getDataXY(x, y))++;
				blObjFound--;
			}
			prevV = v;
		}
	}

	if (debugOutput) blackObjImg.write2BmpFile("calibTest blackobj before max.bmp");

	// generate binary img with max values
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) { // all pixel
			*blackObjImg.getDataXY(x, y) = (*blackObjImg.getDataXY(x, y) == 4);
		}

////////////////////////////////////////////////////////////

	if (debugOutput) blackObjImg.write2BmpFile("calibTest blackobj.bmp");

	// dilate 3x3
	blackObjImg.dilateGray(&blackObjImg2, 3);

/* TODO (#2#): better: closing or geodaetic erosion */
//blackObjImg2.erodeGray(&blackObjImg, 1);

	if (debugOutput) blackObjImg2.write2BmpFile("calibTest blackobj closed.bmp");

	int regNum = blackObjImg2.label16();

	// regions are now merged; background is still 0, colors 1 to regNum used for regNum regions
	if (debugOutput) blackObjImg2.write2BmpFile("calibTest enum merged regions.bmp");

/*
char buf[200];
sprintf((char*) &buf, "%i", regNum);
MessageBox(NULL, buf, "region number", MB_OK);
*/
	if (regNum != CALIB_POINT_NUM) return false; // wrong number of points

	// calc center of each region

	int cn[CALIB_POINT_NUM];
	for (int i=0; i<CALIB_POINT_NUM; i++) {
		calibPoints[i].x = 0.0;
		calibPoints[i].y = 0.0;
		cn[i] = 0;
	}

	for (int y=1; y<imgHeight; y++)
		for (int x=1; x<imgWidth; x++) { // all pixel
			if (int v = *blackObjImg2.getDataXY(x, y)) {
				v--;
				cn[v]++;
				calibPoints[v].x += x;
				calibPoints[v].y += y;
			}
		}
	for (int i=0; i<CALIB_POINT_NUM; i++) {
		calibPoints[i].x /= cn[i];
		calibPoints[i].y /= cn[i];
	}

	if (debugOutput) {
		for (int i=0; i<CALIB_POINT_NUM; i++) *blackObjImg2.getDataXY((int) calibPoints[i].x, (int) calibPoints[i].y) = 255;
		blackObjImg2.write2BmpFile("calibTest points.bmp");
	}

	// generate MST without last CP_SET_NUM (longest) edges to separate the CP_SET_NUM sets

	// generate edge list
	static const int EDGE_NUM = ((CALIB_POINT_NUM-1)*CALIB_POINT_NUM) / 2;
	edgeStruct edges[EDGE_NUM+1];
	int edgInd = 0;
	for (int i=0; i<CALIB_POINT_NUM; i++) for (int j=0; j<CALIB_POINT_NUM; j++) {
		if (i >= j) continue;
		edges[edgInd].v1 = i;
		edges[edgInd].v2 = j;
		edges[edgInd].w = sqr(calibPoints[i].x - calibPoints[j].x)
				+ sqr(calibPoints[i].y - calibPoints[j].y);
		edgInd++;
	}
	// sort edge list
	qsort(&edges, edgInd, sizeof(edgeStruct), compareEdges);

	int setNum[CALIB_POINT_NUM];
	for (int i=0; i<CALIB_POINT_NUM; i++) setNum[i] = 0;
	int edgeSetLut[CALIB_POINT_NUM];
	edgeSetLut[0] = 0;
	int addedEdgeNum = 0;
	int actEdge = -1;
	int tmpSetNum = 0;

	do {
		actEdge++;
		int sne1 = setNum[edges[actEdge].v1];
		int sne2 = setNum[edges[actEdge].v2];
		if ((edgeSetLut[sne1] == edgeSetLut[sne2]) && sne1) continue; // no circles (both same set)

		if (sne1 && sne2) { // different sets:
			int repVal = edgeSetLut[sne2];
			for (int i=1; i<=tmpSetNum; i++) // update lut entry
				if (edgeSetLut[i] == repVal) edgeSetLut[i] = edgeSetLut[sne1];
			addedEdgeNum++;
			continue;
		}
		int sn = sne1 + sne2;
		if (!sn) { // no point in set: add new one
			tmpSetNum++;
			edgeSetLut[tmpSetNum] = tmpSetNum;
			sn = tmpSetNum;
		}

		setNum[edges[actEdge].v1] = edgeSetLut[sn];
		setNum[edges[actEdge].v2] = edgeSetLut[sn];
		addedEdgeNum++;
	} while (addedEdgeNum < CALIB_POINT_NUM - CP_SET_NUM);
	// merge sets using lut's
	//for (int i=0; i<CALIB_POINT_NUM; i++) setNum[i] = edgeSetLut[setNum[i]];
	int edgeSetLutLut[CALIB_POINT_NUM];
	for (int i=0; i<CALIB_POINT_NUM; i++) edgeSetLutLut[i] = 0;
	int ert = 0;
	for (int i=1; i<=tmpSetNum; i++) {
		if (edgeSetLutLut[edgeSetLut[i]] == 0) {
			ert++;
			edgeSetLutLut[edgeSetLut[i]] = ert;
		}
	}
	// replace sets
	for (int i=0; i<CALIB_POINT_NUM; i++) setNum[i] = edgeSetLutLut[edgeSetLut[setNum[i]]]-1;

	if (debugOutput) {
		for (int i=0; i<CALIB_POINT_NUM; i++) *blackObjImg2.getDataXY((int) calibPoints[i].x, (int) calibPoints[i].y) = 150 + setNum[i]*100;
		blackObjImg2.write2BmpFile("calibTest sets.bmp");
	}

	// now setNum[vertex number] contains number of set the vertex belongs to (0 or 1)
	// now build list sorted by sets

	int ptNumPerSet[CP_SET_NUM];
	for (int i=0; i<CP_SET_NUM; i++) ptNumPerSet[i] = 0;
	for (int i=0; i<CALIB_POINT_NUM; i++) ptNumPerSet[setNum[i]]++;
	for (int i=0; i<CP_SET_NUM; i++) if (ptNumPerSet[i] > PTS_SET_NUM[0]) return false; // set too big
	int setPt[CP_SET_NUM][PTS_SET_NUM_0];
	int ptNumPerSet2[CP_SET_NUM];
	for (int i=0; i<CP_SET_NUM; i++) ptNumPerSet2[i] = 0;
	for (int i=0; i<CALIB_POINT_NUM; i++)
	 for (int j=0; j<CP_SET_NUM; j++) {
			if (PTS_SET_NUM[j] == ptNumPerSet[setNum[i]]) {
				if (ptNumPerSet2[j] < PTS_SET_NUM[0]) setPt[j][ptNumPerSet2[j]] = i;
				ptNumPerSet2[j]++;
			}
		}
	for (int i=0; i<CP_SET_NUM; i++) if (ptNumPerSet2[i] != PTS_SET_NUM[i]) return false; // set has wrong size

	// setPt contains point indices now with sets sorted as in PT_SET_NUM

	// now search corners of each set
	cornerStruct cornerInd[CP_SET_NUM][4]; // will contain index of corners in setPt
	for (int i=0; i<CP_SET_NUM; i++) { // all sets
		int cornersFound = 0;
		for (int j=0; j<PTS_SET_NUM[i]; j++) { // all points of set
			// build a histogram containing angles of each other point of set
			int angleHisto[72];
			for (int k=0; k<72; k++) angleHisto[k] = 0;
			for (int k=0; k<PTS_SET_NUM[i]; k++) if (k != j) {
				int angle = (int) (180/PI*atan2(calibPoints[setPt[i][k]].y - calibPoints[setPt[i][j]].y,
						calibPoints[setPt[i][k]].x - calibPoints[setPt[i][j]].x));
				if (angle < 0) angle += 360;
				angleHisto[angle / 5]++;
			}
			// check if there is a space of at least 190 degrees in histogram
			int subseqZeros = 0;
			for (int k=0; k<72 *3/2; k++)
				if (angleHisto[k % 72] == 0) {
					subseqZeros++;
					if (subseqZeros >= 72 * 190/360) {
						// this point j is a corner, so mark it
						cornersFound++;
						if (cornersFound > 4) return false; // max. 4 corners
						cornerInd[i][cornersFound-1].ind = j;
						break;
					}
				} else subseqZeros = 0;
		}
		if (cornersFound != 4) return false; // allways 4 corners needed!
	}

	if (debugOutput) {
		for (int i=0; i<CP_SET_NUM; i++) for (int j=0; j<4; j++) *blackObjImg2.getDataXY((int) calibPoints[setPt[i][cornerInd[i][j].ind]].x, (int) calibPoints[setPt[i][cornerInd[i][j].ind]].y) = 110;
		blackObjImg2.write2BmpFile("calibTest corners.bmp");
	}

	// sort corners (counterclockwise in image)

	for (int i=0; i<CP_SET_NUM; i++) { // all sets
		// calc center of corners
		float cx = 0;
		float cy = 0;
		for (int j=0; j<4; j++) {
			cx += 0.25 * calibPoints[setPt[i][cornerInd[i][j].ind]].x;
			cy += 0.25 * calibPoints[setPt[i][cornerInd[i][j].ind]].y;
		}
		// calc angle of center to each corner
		for (int j=0; j<4; j++)
			cornerInd[i][j].angle = atan2(cy - calibPoints[setPt[i][cornerInd[i][j].ind]].y,
				calibPoints[setPt[i][cornerInd[i][j].ind]].x - cx);
		// sort edge list
		qsort(&cornerInd[i], 4, sizeof(cornerStruct), compareCorners);
	}

	if (debugOutput) {
		for (int i=0; i<CP_SET_NUM; i++) for (int j=0; j<4; j++) *blackObjImg2.getDataXY((int) calibPoints[setPt[i][cornerInd[i][j].ind]].x, (int) calibPoints[setPt[i][cornerInd[i][j].ind]].y) = 110 + 3*j;
		blackObjImg2.write2BmpFile("calibTest corners sorted.bmp");
	}

	// find corner near to hole in second set (down left)

	float maxDistVal = 0;
	int maxDistRotInd = 0;
	for (int j=0; j<4; j++) { // all corner rotations
		float px = 1 - 0.33; // point coords with origin 0 , (1,0) 1 and (0,1) 3
		float py = 1 - 0.25;
		float px1m = 1 - px;
		float py1m = 1 - py;
/* TODO (#2#): instead of max (nearest point to hole) use:
               - min sum of distances (nearest point to not-hole)
               - or: square pattern */
		// interpolate bilinear only - hopefully suffisent
		float holeX = px*py*calibPoints[setPt[1][cornerInd[1][(0+j)%4].ind]].x
				+ px1m*py*calibPoints[setPt[1][cornerInd[1][(1+j)%4].ind]].x
				+ px1m*py1m*calibPoints[setPt[1][cornerInd[1][(2+j)%4].ind]].x
				+ px*py1m*calibPoints[setPt[1][cornerInd[1][(3+j)%4].ind]].x;
		float holeY = px*py*calibPoints[setPt[1][cornerInd[1][(0+j)%4].ind]].y
				+ px1m*py*calibPoints[setPt[1][cornerInd[1][(1+j)%4].ind]].y
				+ px1m*py1m*calibPoints[setPt[1][cornerInd[1][(2+j)%4].ind]].y
				+ px*py1m*calibPoints[setPt[1][cornerInd[1][(3+j)%4].ind]].y;
		// find nearest point to estimated hole
		float nearestDist = 10000000;
		for (int k=0; k<PTS_SET_NUM[1]; k++) {
			float dx = calibPoints[setPt[1][k]].x - holeX;
			float dy = calibPoints[setPt[1][k]].y - holeY;
			float thisDist = sqr(dx) + sqr(dy);
			if (thisDist < nearestDist) nearestDist = thisDist;
		}
		if (nearestDist > maxDistVal) {
			maxDistVal = nearestDist;
			maxDistRotInd = j;
		}
	}
	int rotInd[CP_SET_NUM];
	rotInd[1] = maxDistRotInd;

	// now other set (set 0: the one with 20 points)
	float minDistVal = 10000000;
	int minDistRotInd = 0;
	for (int j=0; j<4; j++) { // all corner rotations
		float dx = calibPoints[setPt[0][cornerInd[0][(1+j)%4].ind]].x - calibPoints[setPt[1][cornerInd[1][(0+rotInd[1])%4].ind]].x;
		float dy = calibPoints[setPt[0][cornerInd[0][(1+j)%4].ind]].y - calibPoints[setPt[1][cornerInd[1][(0+rotInd[1])%4].ind]].y;
		float thisDist = sqr(dx) + sqr(dy);
		if (thisDist < minDistVal) {
			minDistVal = thisDist;
			minDistRotInd = j;
		}
	}
	rotInd[0] = minDistRotInd;

	bool calibPointsFound[CALIB_POINT_NUM];
	for (int i=0; i<CALIB_POINT_NUM; i++) calibPointsFound[i] = false;
	for (int i=0; i<CP_SET_NUM; i++) { // all sets
		for (int ptx=0; ptx<4; ptx++)
		 for (int pty=0; pty<5; pty++) { // all 3d-calib pattern points
			// find closest image point to bilinearly interpolated
			if ((i == 1) && (ptx == 1) && (pty == 1)) continue; // skip hole

			float px = 1 - ptx / 3.0; // point coords with origin 0 , (1,0) 1 and (0,1) 3
			float py = 1 - pty / 4.0;
			float px1m = 1 - px;
			float py1m = 1 - py;
			// interpolate bilinear only - hopefully suffisent
			float iX = px*py*calibPoints[setPt[i][cornerInd[i][(0+rotInd[i])%4].ind]].x
					+ px1m*py*calibPoints[setPt[i][cornerInd[i][(1+rotInd[i])%4].ind]].x
					+ px1m*py1m*calibPoints[setPt[i][cornerInd[i][(2+rotInd[i])%4].ind]].x
					+ px*py1m*calibPoints[setPt[i][cornerInd[i][(3+rotInd[i])%4].ind]].x;
			float iY = px*py*calibPoints[setPt[i][cornerInd[i][(0+rotInd[i])%4].ind]].y
					+ px1m*py*calibPoints[setPt[i][cornerInd[i][(1+rotInd[i])%4].ind]].y
					+ px1m*py1m*calibPoints[setPt[i][cornerInd[i][(2+rotInd[i])%4].ind]].y
					+ px*py1m*calibPoints[setPt[i][cornerInd[i][(3+rotInd[i])%4].ind]].y;
			// find nearest point to interpolated one
			float nearestDist = 10000000;
			int nearestDistInd = 0;
			for (int j=0; j<PTS_SET_NUM[i]; j++) { // all points of set
				float dx = calibPoints[setPt[i][j]].x - iX;
				float dy = calibPoints[setPt[i][j]].y - iY;
				float thisDist = sqr(dx) + sqr(dy);
				if (thisDist < nearestDist) {
					nearestDist = thisDist;
					nearestDistInd = setPt[i][j];
				}
			}

			calibPointsFound[nearestDistInd] = true;

			switch (i) {
				case 0:
					calibPoints[nearestDistInd].wx = 0;
					calibPoints[nearestDistInd].wy = (4-ptx) * worldDistUnit;
					calibPoints[nearestDistInd].wz = (pty+0.5) * worldDistUnit;
					break;
				case 1:
					calibPoints[nearestDistInd].wx = (ptx+1) * worldDistUnit;
					calibPoints[nearestDistInd].wy = 0;
					calibPoints[nearestDistInd].wz = (pty+0.5) * worldDistUnit;
					break;
			}
		}
	}

	for (int i=0; i<CALIB_POINT_NUM; i++) if (!calibPointsFound[i]) return false; // not all points could be matched

	return true;
}

bool CalibrateCam::writeMatchToFile(char* fileName) {
	FILE* mFile = fopen(fileName, "w");
	if (mFile == NULL) return false;

	for (int i=0; i<CALIB_POINT_NUM; i++)
		fprintf(mFile, "%f %f %f %f %f\n", calibPoints[i].wx, calibPoints[i].wy, calibPoints[i].wz, calibPoints[i].x, calibPoints[i].y);

	fclose(mFile);
}

// calculate intrinsic and extrinsic cam parameters
bool CalibrateCam::calcCamParams() {
	extern struct calibration_data cd;
	extern struct calibration_constants cc;
	extern struct camera_parameters cp;

	if (CALIB_POINT_NUM < 11) return false;
	if (CALIB_POINT_NUM >= MAX_POINTS) return false;

	// copy and scale calibration data points
    cd.point_count = CALIB_POINT_NUM;
	for (int i=0; i<CALIB_POINT_NUM; i++) {
		cd.xw[i] = calibPoints[i].wx * 1000; // [mm]
		cd.yw[i] = calibPoints[i].wy * 1000; // [mm]
		cd.zw[i] = calibPoints[i].wz * 1000; // [mm]
		cd.Xf[i] = calibPoints[i].x; // [pixel]
		cd.Yf[i] = calibPoints[i].y; // [pixel]
	}

	// run calibration using Tsai's method
	calibNonCoplanar(imgWidth, imgHeight, 0/*false*/);

	camTrafo.cx = cp.Cx;
	camTrafo.cy = cp.Cy;

	camTrafo.f = cc.f;
	camTrafo.k1 = cc.kappa1;
	camTrafo.tx = cc.Tx;
	camTrafo.ty = cc.Ty;
	camTrafo.tz = cc.Tz;
	camTrafo.rx = cc.Rx;
	camTrafo.ry = cc.Ry;
	camTrafo.rz = cc.Rz;

	camTrafo.imgWidth = imgWidth;
	camTrafo.imgHeight = imgHeight;

	return true;
}

bool CalibrateCam::writeParamsToFile(char* fileName) {
	FILE* mFile = fopen(fileName, "w");
	if (mFile == NULL) return false;

	fprintf(mFile, "image center\n cx: %f pixel\n", camTrafo.cx);
	fprintf(mFile, " cy: %f pixel\n", camTrafo.cy);
	fprintf(mFile, "focal length\n f: %f mm\n", camTrafo.f);
	fprintf(mFile, "lens distortion\n k1: %f 1/mm^2\n", camTrafo.k1);
	fprintf(mFile, "translation\n tx: %f mm\n", camTrafo.tx);
	fprintf(mFile, " ty: %f mm\n", camTrafo.ty);
	fprintf(mFile, " tz: %f mm\n", camTrafo.tz);
	fprintf(mFile, "rotation euler angles\n rx: %f rad\n", camTrafo.rx);
	fprintf(mFile, " ry: %f rad\n", camTrafo.ry);
	fprintf(mFile, " rz: %f rad\n", camTrafo.rz);

	fclose(mFile);
}

////////////////////////////////////////////////////////////////////////////////
// compare functions for sorting

/*static*/ int CalibrateCam::compareEdges(const void *e1, const void *e2) {
	return (int) (((edgeStruct*)e1)->w - ((edgeStruct*)e2)->w);
}

/*static*/ int CalibrateCam::compareCorners(const void *c1, const void *c2) {
	float d = ((cornerStruct*)c1)->angle - ((cornerStruct*)c2)->angle;
	if (d < 0) return -1;
	return (d > 0);
}
