/**
 * Camera calibration class.
 * Finds calibration points of calibration pattern, maps 2d to 3d points and
 * determines intrinsic and extrinsic parameters using Tsai's method
 * for 3dscanner project
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef CALIBRATECAM_HEADER_INCLUDED
#define CALIBRATECAM_HEADER_INCLUDED

#include "MBmpImg.h"
#include "CamTrafo.h"

class CalibrateCam {
  public:
	CalibrateCam();
//	~CalibrateCam();

	enum CalibrationPattern{CAL_PAT_1, CAL_PAT_2}; ///< available calibration patterns
	/// matchCalibrationPoints and calcCamParams
	bool matchAndCalcParams(MBmpImg* img, CalibrationPattern calibrationPattern);

	///////////////////////////////////////////////////////////////////////////

	/// finds and matches 2d image points to 3d object points
	bool matchCalibrationPoints(MBmpImg* img, CalibrationPattern calibrationPattern);

	/// calculate intrinsic and extrinsic cam parameters
	bool calcCamParams();

	/// writes a file with all world coordinates and the corresponding image coordinates
	bool writeMatchToFile(char* fileName);

	/// writes a file with all estimated intrinsic and extrinsic camera parameters
	bool writeParamsToFile(char* fileName);

	void setDebugOutput(bool debugOutput);

	static const int CALIB_POINT_NUM = 39; ///< number for both calibration patterns
	struct {float x, y; float wx, wy, wz;} calibPoints[CALIB_POINT_NUM]; ///< for matching list

	CamTrafo camTrafo;

  private:
	static const int CP_SET_NUM = 2;
	static const int PTS_SET_NUM[CP_SET_NUM];// = {20, 19}; // max value first, all values different!!

	int imgWidth, imgHeight;
	bool debugOutput;

	struct edgeStruct {
		int v1, v2;
		float w; // weight is squared length
	};
	struct cornerStruct {
		int ind;
		float angle;
	};

	// for sorting
	static int compareEdges(const void *e1, const void *e2);
	static int compareCorners(const void *c1, const void *c2);
};

#endif // CALIBRATECAM_HEADER_INCLUDED
