#include "CamAccess.h"
#include "ScannerApp.h"
#include "frontend.h"

CamAccess::CamAccess() {
	hwnd = NULL;
	logFilePath = NULL;
}

CamAccess::~CamAccess() {
	if (logFilePath) delete[] logFilePath;
}

void CamAccess::init(HWND hwnd, char* logFilePath) {
	this->hwnd = hwnd;

	if (this->logFilePath) delete[] this->logFilePath;
	this->logFilePath = new char[strlen(logFilePath) + 1];
	strcpy(this->logFilePath, logFilePath);
}

bool CamAccess::initOk() {
//	nextFileName = NULL;
	char* LFN = "3DScanner Twain log file.txt";
	char* logFileName = new char[strlen(logFilePath) + strlen(LFN) + 1];
	strcpy(logFileName, logFilePath);
	strcat(logFileName, LFN);
	startLog(logFileName);
	delete[] logFileName;
	if (initialize(hwnd, "3DScanner",
		"Marcel Lancelle, TU Braunschweig")) return true;

	// else: show error message
	MessageBox(NULL, "Twain initialisation failed!", "3DScanner",
		MB_ICONEXCLAMATION | MB_OK);
	return false;
}

void CamAccess::onImageTransferDone(int width, int height, char bitsPerPixel) {
	waitingForAquisition = false;

//MBmpImg::writeBmp2File(imageDataPtr, "_tmp_.bmp");
	acquiredBmp = new MBmpImg(imageDataPtr);
/*
	//MessageBox(NULL, "image just arrived in application :-)", "testMTwain", MB_OK);
	if (nextFileName == NULL) return;
	//MessageBox(NULL, nextFileName, "saving as", MB_OK);
	MBmpImg::writeBmp2File((BYTE*) imageDataPtr, (char *) nextFileName);
	delete nextFileName;
	nextFileName = NULL;

	// seems to work better with Phillips Cam ?
	if (!camUiControllable) {
		disableSource();
	}

	//MessageBox (NULL, "Posting WM_USER+0..." , "3DScanner", MB_OK);
	PostMessage(hwnd, WM_USER_GET_IMG, GET_IMG_CONTINUE, 0);
*/
}

MBmpImg* CamAccess::getSingleImgTwainModal(bool showUI) {
	imageDataPtr = NULL;
	waitingForAquisition = true;

	if (!openSource()) {
		MessageBox(NULL, "Twain source initialisation failed!", "3DScanner",
			MB_ICONEXCLAMATION | MB_OK);
		backToState(STATE_SOURCE_MANAGER_LOADED);
		return NULL;
	}

	camUiControllable = getCapabilityBool(CAP_UICONTROLLABLE);

	// get 1 image
	TW_CAPABILITY cap;
	pTW_ONEVALUE pval;

	cap.Cap = CAP_XFERCOUNT;
	cap.ConType = TWON_ONEVALUE;

	// alloc the container
	if ((cap.hContainer = GlobalAlloc(GHND, sizeof(TW_ONEVALUE))) != NULL) {
		pval = (pTW_ONEVALUE)GlobalLock(cap.hContainer);
		pval->ItemType = TWTY_UINT16;
		pval->Item = 1;
		GlobalUnlock(cap.hContainer);
		setCapability(&cap);
		GlobalFree((HANDLE)cap.hContainer);
	}

	enableSource(showUI || !camUiControllable, true);

	MSG msg;
	while (waitingForAquisition && GetMessage((LPMSG)&msg, NULL, 0, 0)) {
		if (!preProcessMessage(&msg)) {
			TranslateMessage ((LPMSG)&msg);
			DispatchMessage ((LPMSG)&msg);
		}
	}

	if (getState() >= STATE_SOURCE_ENABLED) disableSource();
	closeSource();

	return acquiredBmp;
}

MBmpImg* CamAccess::acquireImg(ImageAquisitionType imgAquisitionType, bool showUI, char* contentDescription) {
	switch (imgAquisitionType) {
		case IAT_FILE: {
			static const int MAX_FILENAME_LENGTH = 500;
			char fileNameBuffer[MAX_FILENAME_LENGTH];
			if (!getFileNameOpen(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"Bitmap files\0*.bmp\0", (char*) contentDescription, NULL)) return NULL;
			return new MBmpImg(fileNameBuffer);}
		case IAT_TWAIN:
			if (!initOk()) return NULL;
			return getSingleImgTwainModal(showUI);
		case IAT_VFW:
return NULL;
	}
	return NULL;
}

bool CamAccess::acquireImgAndSave2File(ImageAquisitionType imgAquisitionType, bool showUI, char* contentDescription, char* imgFileName) {
	MBmpImg* bmpPtr = acquireImg(imgAquisitionType, showUI, contentDescription);
	if (!bmpPtr) return false;
	bool retVal = bmpPtr->write2BmpFile(imgFileName);
	delete bmpPtr;
	return retVal;
}

void CamAccess::selectSource(ImageAquisitionType imageAquisitionType) {
	switch (imageAquisitionType) {
		case IAT_TWAIN:
			if (!initOk()) return;
			showSelectSourceDialog();
			break;
		case IAT_VFW:
MessageBox(NULL, "not yet implemented", "3DScanner", MB_ICONWARNING | MB_OK);
			break;
	}
}

void CamAccess::configSource(ImageAquisitionType imageAquisitionType) {
	switch (imageAquisitionType) {
		case IAT_TWAIN:
			if (!initOk()) return;
			if (!openSource()) return;
			if (!showConfigSourceUI()) return;
			if (!closeSource()) return;
			break;
		case IAT_VFW:
MessageBox(NULL, "not yet implemented", "3DScanner", MB_ICONWARNING | MB_OK);
			break;
	}
}

void CamAccess::onStateChange(char oldState, char newState) {
    MTwain::onStateChange(oldState, newState);
/*
// for debugging:
	// suppress MessageBox in following case so it won't eat MSG_XFERREADY message
	if ((newState == 5) && (oldState == 4)) return;

	char buffer[100];
	sprintf(buffer, "stateChange: %i --> %i", oldState, newState);
	MessageBox(NULL, buffer, "stateChange", MB_OK);
*/
}

/* (old:)
bool CamAccess::prepareGetMultiImgNoUI() {
	if (openSource()) {
		camUiControllable = getCapabilityBool(CAP_UICONTROLLABLE);

if (camUiControllable) {
MessageBox(NULL, "Ui Controllable", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
} else {
MessageBox(NULL, "Ui NOT Controllable", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
}

		// if supported, don't show user interface
		//if (enableSource(!camUiControllable, false)) return true;
		return true;
	}

	// else: error case
	MessageBox(NULL, "Twain source initialisation failed!", "3DScanner",
		MB_ICONEXCLAMATION | MB_OK);
	backToState(STATE_SOURCE_MANAGER_LOADED);
	return false;
}

bool CamAccess::acquireImgNoUIAndSave2File(char* imgFileName) {
	if (getState() >= STATE_SOURCE_ENABLED) disableSource();

	// copy image file name
	if (nextFileName) {
		delete nextFileName;
		nextFileName = NULL;
	}
	nextFileName = new char[strlen(imgFileName) + 1];
	strcpy(nextFileName, imgFileName);
	//MessageBox(NULL, nextFileName, "cpy", MB_OK);

	// ??? don't knwo why, but when taking several images with a Phillips cam driver crashes when enabling source twice
	// perhaps this cam doesn't support transfer of multiple images ???
	// with this code it works:
	if (!camUiControllable) {
		closeSource();
		openSource();
	}

	// if supported, don't show user interface
	return enableSource(!camUiControllable, false);
}

void CamAccess::stopGetMultiImgNoUI() {
	if (getState() >= STATE_SOURCE_ENABLED) disableSource();
	closeSource();
}

MBmpImg* CamAccess::getSingleImgNoUIModal() {
	if (!openSource()) {
		MessageBox(NULL, "Twain source initialisation failed!", "3DScanner",
			MB_ICONEXCLAMATION | MB_OK);
		backToState(STATE_SOURCE_MANAGER_LOADED);
		return NULL;
	}

	camUiControllable = getCapabilityBool(CAP_UICONTROLLABLE);

if (camUiControllable) {
MessageBox(NULL, "Ui Controllable", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
} else {
MessageBox(NULL, "Ui NOT Controllable", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
}

	// get 1 image
	TW_CAPABILITY cap;
	pTW_ONEVALUE pval;

	cap.Cap = CAP_XFERCOUNT;
	cap.ConType = TWON_ONEVALUE;

	// alloc the container
	if ((cap.hContainer = GlobalAlloc(GHND, sizeof(TW_ONEVALUE))) != NULL) {
		pval = (pTW_ONEVALUE)GlobalLock(cap.hContainer);
		pval->ItemType = TWTY_UINT16;
		pval->Item = 1;
		GlobalUnlock(cap.hContainer);
		setCapability(&cap);
		GlobalFree((HANDLE)cap.hContainer);
	}

	char* tmpFileName = (char*) &"tmp.bmp";
	// copy image file name
	if (nextFileName) {
		delete nextFileName;
		nextFileName = NULL;
	}
	nextFileName = new char[strlen(tmpFileName) + 1];
	strcpy(nextFileName, tmpFileName);

enableSource(!camUiControllable, false);
	MSG msg;
	while (nextFileName && GetMessage((LPMSG)&msg, NULL, 0, 0)) {
		if (!preProcessMessage(&msg)) {
			TranslateMessage ((LPMSG)&msg);
			DispatchMessage ((LPMSG)&msg);
		}
	}



	if (getState() >= STATE_SOURCE_ENABLED) disableSource();
	closeSource();

	MBmpImg *bmp = new MBmpImg(tmpFileName);
	return bmp;
}*/
