#include "MBmpImg.h"
#include "stdio.h"
#include <math.h>
#include "mathext.h"

/*

bmp file format (wichtigstes)

BITMAPFILEHEADER
	WORD	bfType			"BM" = 0x4D42
	DWORD	bfSize			Dateigre; nicht benutzen! da einige Programme anders rechnen (exkl. Gre des BITMAPFILEHEADER oder so???)
	WORD	bfReserved1
	WORD	bfReserved2
	DWORD	bfOffBits		Offset von Anf. BITMAPFILEHEADER bis Daten
BITMAPINFO
	BITMAPINFOHEADER bmiHeader;
		DWORD	biSize		Gre von BITMAPINFOHEADER (40)
		LONG	biWidth
		LONG	biHeight	(angeblich v. oben nach unten, wenn negativ; untersttzt aber kein Programm...)
		WORD	biPlanes	immer 1
		WORD	biBitCount	1, 4, 8, 16, 24, 32
		DWORD	biCompression
		DWORD	biSizeImage	Gre des Bildes in Byte oder 0, wenn unkomprimiert
		LONG	biXPelsPerMeter	horizontale Auflsung: Pixel pro Meter
		LONG	biYPelsPerMeter vertikale Auflsung
		DWORD	biClrUsed	Anzahl benutzter Palettenfarben; 0: max
		DWORD	biClrImportant	0: alle
	RGBQUAD	bmiColors[1]
		Palette...
	Daten... (vertikal gespiegelt; aligned to 4 byte)

*/

MBmpImg::MBmpImg(int width, int height, char bitsPerPixel) {
	this->width = width;
	this->height = height;
	this->bitsPerPixel = bitsPerPixel;
	rowSize = bitsPerPixel * width / 8;
	dataPtr = new BYTE[rowSize * height];
	palSize = 0;
	palPtr = NULL;
}

MBmpImg::MBmpImg(char* bmpFileName) {
	width = 0;
	height = 0;
	bitsPerPixel = 0;
	palSize = 0;
	palPtr = NULL;

	FILE* bmpFile = fopen(bmpFileName, "rb"); // !! das b steht fr binary, wichtig!
	if (bmpFile == NULL) return;

	BITMAPFILEHEADER bitmapFileHeader;
	fread((void*) &bitmapFileHeader, 1, sizeof(BITMAPFILEHEADER), bmpFile);
	if (bmpFile == NULL) return;

	if (bitmapFileHeader.bfType != 0x4D42) return;

	BITMAPINFOHEADER bmpInfoHeader;
	fread((void*) &bmpInfoHeader, 1, sizeof(BITMAPINFOHEADER), bmpFile);

	if (bmpInfoHeader.biWidth < 1) return;
	if (bmpInfoHeader.biHeight < 1) return;

	width = bmpInfoHeader.biWidth;
	height = bmpInfoHeader.biHeight;
	bitsPerPixel = bmpInfoHeader.biBitCount;
	rowSize = bitsPerPixel * width / 8;
	dataPtr = new BYTE[rowSize * height];

	// palette
	if (bitsPerPixel <= 8) {
		palSize = bmpInfoHeader.biClrUsed;
		if (palSize == 0) palSize = 1 << bitsPerPixel;
		BYTE* tmpPalPtr = new BYTE[palSize * 4];
		fread((void*) tmpPalPtr, 1, palSize * 4, bmpFile);

		palPtr = new BYTE[palSize * 3];
		for (int i=0; i<palSize; i++) for (int p=0; p<3; p++) palPtr[i*3+p] = tmpPalPtr[i*4+p];

		delete[] tmpPalPtr;
	}

	// data
	int lineSize = bitsPerPixel * width / 8;
	while (lineSize % 4) lineSize++; // align to 4 byte boundary
	BYTE* lineBufferPtr = new BYTE[lineSize];
	BYTE* destPtr;
	for (int y=height-1; y>=0; y--) {
		fread((void*) lineBufferPtr, 1, lineSize, bmpFile);
		BYTE* srcPtr = lineBufferPtr;
		destPtr = getDataXY(0, y);
		switch (bitsPerPixel) {
			case 24:
				for (int x=0; x<width; x++) {
					// exchange r and b
					BYTE b = *srcPtr++;
					BYTE g = *srcPtr++;
					BYTE r = *srcPtr++;
					*destPtr++ = r;
					*destPtr++ = g;
					*destPtr++ = b;
				}
				break;
			default: for (int ax=0; ax<width*bitsPerPixel/8; ax++) *destPtr++ = *srcPtr++;
		}
    }

	delete[] lineBufferPtr;
}

// points to BITMAPINFO / BITMAPINFOHEADER
MBmpImg::MBmpImg(BYTE* bmpPtr) {
	palSize = 0;
	palPtr = NULL;
	width = 0;
	height = 0;
	if (bmpPtr == NULL) return;

    BITMAPINFOHEADER* bmpInfoHeaderPtr = (BITMAPINFOHEADER*) bmpPtr;
	if (bmpInfoHeaderPtr->biWidth < 1) return;
	if (bmpInfoHeaderPtr->biHeight < 1) return;

	width = bmpInfoHeaderPtr->biWidth;
	height = bmpInfoHeaderPtr->biHeight;
	bitsPerPixel = bmpInfoHeaderPtr->biBitCount;
	rowSize = bitsPerPixel * width / 8;
	dataPtr = new BYTE[rowSize * height];
	bmpPtr += sizeof(BITMAPINFOHEADER);

	// palette
	if (bitsPerPixel <= 8) {
		palSize = bmpInfoHeaderPtr->biClrUsed;
		if (palSize == 0) palSize = 1 << bitsPerPixel;
		BYTE* tmpPalPtr = bmpPtr;

		palPtr = new BYTE[palSize * 3];
		for (int i=0; i<palSize; i++) for (int p=0; p<3; p++) palPtr[i*3+p] = tmpPalPtr[i*4+p];
		bmpPtr += palSize*4;
	}

	// data
	int lineSize = bitsPerPixel * width / 8;
	while (lineSize % 4) lineSize++; // align to 4 byte boundary
	BYTE* destPtr;
	for (int y=height-1; y>=0; y--) {
		BYTE* srcPtr = bmpPtr;
		bmpPtr += lineSize;
		destPtr = getDataXY(0, y);
		switch (bitsPerPixel) {
			case 24:
				for (int x=0; x<width; x++) {
					// exchange r and b
					BYTE b = *srcPtr++;
					BYTE g = *srcPtr++;
					BYTE r = *srcPtr++;
					*destPtr++ = r;
					*destPtr++ = g;
					*destPtr++ = b;
				}
				break;
			default: for (int ax=0; ax<width*bitsPerPixel/8; ax++) *destPtr++ = *srcPtr++;
		}
    }
}

MBmpImg::~MBmpImg() {
	if (dataPtr) delete[] dataPtr;
	if (palPtr) delete[] palPtr;
}

// bmpPtr points to BITMAPINFO / BITMAPINFOHEADER
/*static*/ bool MBmpImg::writeBmp2File(BYTE* bmpPtr, char* bmpFileName) {
    BITMAPINFOHEADER* bmpInfoHeaderPtr = (BITMAPINFOHEADER*) bmpPtr;
	if (bmpInfoHeaderPtr->biWidth < 1) return false;
	if (bmpInfoHeaderPtr->biHeight < 1) return false;
	int width = bmpInfoHeaderPtr->biWidth;
	int height = bmpInfoHeaderPtr->biHeight;
	int bitsPerPixel = bmpInfoHeaderPtr->biBitCount;
	int rowSize = bitsPerPixel * width / 8;
	int dataSize = rowSize * height;

	FILE* bmpFile = fopen(bmpFileName, "wb"); // !! das b steht fr binary, wichtig!
	if (bmpFile == NULL) return false;

	BITMAPFILEHEADER bitmapFileHeader;
	bitmapFileHeader.bfType = 0x4D42;
	bitmapFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER); // Offset von Anf. BITMAPFILEHEADER bis Daten
	bitmapFileHeader.bfSize = dataSize + bitmapFileHeader.bfOffBits; // Dateigre; einige Programme rechnen anders (exkl. Gre des BITMAPFILEHEADER oder so???)
	fwrite((void*) &bitmapFileHeader, 1, sizeof(BITMAPFILEHEADER), bmpFile);

	fwrite((void*) bmpInfoHeaderPtr, 1, sizeof(BITMAPINFOHEADER), bmpFile);

	bmpPtr += sizeof(BITMAPINFOHEADER);

	// palette
	int palSize = 0;
	if (bitsPerPixel <= 8) {
		palSize = bmpInfoHeaderPtr->biClrUsed;
		if (palSize == 0) palSize = 1 << bitsPerPixel;
		palSize *= 4;
	}

	fwrite((void*) bmpPtr, 1, dataSize + palSize, bmpFile);
	fclose(bmpFile);
}

bool MBmpImg::write2BmpFile(char* bmpFileName) {
	FILE* bmpFile = fopen(bmpFileName, "wb"); // !! das b steht fr binary, wichtig!
	if (bmpFile == NULL) return false;

	// if in 8 bit mode and no palette is specified, use gray palette
	int bmpPalMemSize = ((bitsPerPixel == 8) && (palSize == 0))?1024:palSize*4;

	BITMAPFILEHEADER bitmapFileHeader;
	bitmapFileHeader.bfType = 0x4D42; // "BM"
	bitmapFileHeader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)
		+ palSize * 4 + rowSize * height;
	bitmapFileHeader.bfReserved1 = 0;
	bitmapFileHeader.bfReserved2 = 0;
	bitmapFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bmpPalMemSize;
	fwrite((void*) &bitmapFileHeader, 1, sizeof(BITMAPFILEHEADER), bmpFile);

	BITMAPINFOHEADER bitmapInfoHeader;
	bitmapInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
	bitmapInfoHeader.biWidth = width;
	bitmapInfoHeader.biHeight = height;
	bitmapInfoHeader.biPlanes = 1;
	bitmapInfoHeader.biBitCount = bitsPerPixel;
	bitmapInfoHeader.biCompression = BI_RGB;
	bitmapInfoHeader.biSizeImage = 0;
	bitmapInfoHeader.biXPelsPerMeter = 1000;
	bitmapInfoHeader.biYPelsPerMeter = 1000;
	bitmapInfoHeader.biClrUsed = 0;
	bitmapInfoHeader.biClrImportant = 0;
	fwrite((void*) &bitmapInfoHeader, 1, sizeof(BITMAPINFOHEADER), bmpFile);

	// palette
	// if in 8 bit mode and no palette is specified, use gray palette
	if ((bitsPerPixel == 8) && (palSize == 0)){
		BYTE grayPal[256][4];
		for (int i=0; i<256; i++) {
			grayPal[i][0] = i;
			grayPal[i][1] = i;
			grayPal[i][2] = i;
			grayPal[i][3] = 0;
		}
		fwrite((void*) &grayPal, 1, 1024, bmpFile);
	} else if (bitsPerPixel <= 8) {
/*
Palette schreiben: noch implementieren; hier der Code beim Lesen:
	int bmpPalMemSize = ((bitsPerPixel == 8) && (palSize == 0))?256:palSize*4;

	bmpPalMemSize
	if (bitsPerPixel <= 8) {
		palSize = bmpInfoHeader.biClrUsed;
		if (palSize == 0) palSize = 1 << bitsPerPixel;
		BYTE* tmpPalPtr = new BYTE[palSize * 4];
		fread((void*) tmpPalPtr, 1, palSize * 4, bmpFile);

		palPtr = new BYTE[palSize * 3];
		for (int i=0; i<palSize; i++) for (int p=0; p<3; p++) palPtr[i*3+p] = tmpPalPtr[i*4+p];

		delete[] tmpPalPtr;
	}
*/
	}

	// data
	int lineSize = rowSize;
	while (lineSize % 4) lineSize++; // align to 4 byte boundary
	BYTE* lineBufferPtr = new BYTE[lineSize];
	for (int i=rowSize; i<lineSize; i++) lineBufferPtr[i] = 0; // fill unused with 0

	BYTE* srcPtr;
	for (int y=height-1; y>=0; y--) {
		BYTE* destPtr = lineBufferPtr;
		srcPtr = getDataXY(0, y);
		switch (bitsPerPixel) {
			case 24: for (int x=0; x<width; x++) {
					// exchange r and b
					BYTE b = *srcPtr++;
					BYTE g = *srcPtr++;
					BYTE r = *srcPtr++;
					*destPtr++ = r;
					*destPtr++ = g;
					*destPtr++ = b;
				}
				break;
			default: for (int ax=0; ax<width*bitsPerPixel/8; ax++) *destPtr++ = *srcPtr++;
		}
		fwrite((void*) lineBufferPtr, 1, lineSize, bmpFile);
    }

	fclose(bmpFile);
	delete[] lineBufferPtr;
}

/*
BYTE* MBmpImg::write2BmpMem() {
}
*/

int MBmpImg::getWidth() {
	return width;
}

int MBmpImg::getHeight() {
	return height;
}

char MBmpImg::getBitsPerPixel() {
	return bitsPerPixel;
}

BYTE* MBmpImg::getData() {
	return dataPtr;
}

BYTE* MBmpImg::getDataXY(int x, int y) {
	return dataPtr + y*rowSize + x*bitsPerPixel/8;
}

// returns linearly interpolated gray value
BYTE MBmpImg::getGrayLinIntPolXY(float x, float y) {
	int flx = (int) floor(x);
	float fx2 = x - flx;
	float fx1 = 1 - fx2;

	int fly = (int) floor(y);
	float fy2 = y - fly;
	float fy1 = 1 - fy2;

	return (int) (fx1*fy1* *getDataXY(flx, fly)
			+ fx2*fy1* *getDataXY(flx+1, fly)
			+ fx1*fy2* *getDataXY(flx, fly+1)
			+ fx2*fy2* *getDataXY(flx+1, fly+1));
}

// linearly interpolated rgb-values
void MBmpImg::getRGBLinIntPolXY(float x, float y, BYTE* r, BYTE* g, BYTE* b) {
	int flx = (int) floor(x);
	float fx2 = x - flx;
	float fx1 = 1 - fx2;

	int fly = (int) floor(y);
	float fy2 = y - fly;
	float fy1 = 1 - fy2;

	BYTE* px1y1 = getDataXY(flx, fly);
	BYTE* px2y1 = getDataXY(flx+1, fly);
	BYTE* px1y2 = getDataXY(flx, fly+1);
	BYTE* px2y2 = getDataXY(flx+1, fly+1);

	*r = (int) (fx1*fy1 * *px1y1++
			+ fx2*fy1 * *px2y1++
			+ fx1*fy2 * *px1y2++
			+ fx2*fy2 * *px2y2++);
	*g = (int) (fx1*fy1 * *px1y1++
			+ fx2*fy1 * *px2y1++
			+ fx1*fy2 * *px1y2++
			+ fx2*fy2 * *px2y2++);
	*b = (int) (fx1*fy1 * *px1y1
			+ fx2*fy1 * *px2y1
			+ fx1*fy2 * *px1y2
			+ fx2*fy2 * *px2y2);
}

int MBmpImg::getPalSize() {
	return palSize;
}

BYTE* MBmpImg::getPalData() {
	return palPtr;
}



BYTE MBmpImg::getDataXYGrayInRGB(int x, int y) {
	BYTE* datPtr = getDataXY(x, y);
	BYTE r = *datPtr++;
	BYTE g = *datPtr++;
	BYTE b = *datPtr;
	return ((int) r + (int) g + (int) b)/3;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
///////////////////      image manipulation routines     ///////////////////
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


// h, s, l: 0..255
/*static*/ void MBmpImg::getHSL(int r, int g, int b, int* h, int* s, int* l) {
	int max = r;
	if (g > max) max = g;
	if (b > max) max = b;
	int min = r;
	if (g < min) min = g;
	if (b < min) min = b;

	*l = (BYTE) ((min + max)/2); //(r + g + b) / 3;

	//*s = (*l == 0) ? 0 : (255*(max-min)) / *l; //abs(r - *l) + abs(g - *l) + abs(b - *l);
	int dn = 128 - abs(*l - 127);
	int sInt = (dn) ? (max-min)*128 / dn : 0;
	*s = CLIP(sInt, 0, 255); //abs(r - *l) + abs(g - *l) + abs(b - *l);

	float cx = r;
	float cy = 0;
	cx +=  -0.5 * g; // -0.5 = cos(120*3.14159/180)
	cy +=  0.866 * g; // 0.866 = sin(120*3.14159/180) = sqrt(3)/2;
	cx +=  -0.5 * b; // -0.5 = cos(240*3.14159/180)
	cy +=  -0.866 * b; // -0.866 = sin(240*3.14159/180) = -sqrt(3)/2;

	if ((r == g) && (r == b)) {
		*h = 0;
		return;
	}

	*h = (BYTE) (255.0/(2*PI)*(atan2(cy, cx))); // negative values become positive by casting to BYTE
}

// h, s, l: 0..255
/*static*/ void MBmpImg::getRGB(int h, int s, int l, int* r, int* g, int* b) {
	*r = CLIP((int) (l + s * 0.5 * cos(h * 2 * PI / 255)), 0, 255);
	*g = CLIP((int) (l + s * 0.5 * cos((85 - h) * 2 * PI / 255)), 0, 255);
	*b = CLIP((int) (l + s * 0.5 * cos((170 - h) * 2 * PI / 255)), 0, 255);
}

void MBmpImg::RGB2HSL(MBmpImg* dest) {
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			BYTE* datPtr = getDataXY(x, y);
			BYTE r = *datPtr++;
			BYTE g = *datPtr++;
			BYTE b = *datPtr;
			int h, s, l;
			getHSL(r, g, b, &h, &s, &l);
			datPtr = dest->getDataXY(x, y);
			*datPtr++ = h;
			*datPtr++ = s;
			*datPtr = l;
		}
}

// line functions: no clipping -> paint lines only within image borders!!
void MBmpImg::lineGray(int x1, int y1, int x2, int y2, BYTE gray) {
	int dx = x2 - x1;
	int dy = y2 - y1;
	if (abs(dx) > abs(dy)) {
		if (x2 < x1) {
			int tmp = x1; x1 = x2; x2 = tmp;
			tmp = y1; y1 = y2; y2 = tmp;
		}
		int x = x1;
		float y = y1;
		float yIncr = (y2-y1) / (float) (x2-x1);
		while (x <= x2) {
			*getDataXY(x, (int) y) = gray;
			x++;
			y += yIncr;
		}
	} else {
		if (y2 < y1) {
			int tmp = x1; x1 = x2; x2 = tmp;
			tmp = y1; y1 = y2; y2 = tmp;
		}
		int y = y1;
		float x = x1;
		float xIncr = (x2-x1) / (float) (y2-y1);
		while (y <= y2) {
			*getDataXY((int) x, y) = gray;
			x += xIncr;
			y++;
		}
	}
}

void MBmpImg::lineRGB(int x1, int y1, int x2, int y2, BYTE r, BYTE g, BYTE b) {
	int dx = x2 - x1;
	int dy = y2 - y1;
	if (abs(dx) > abs(dy)) {
		if (x2 < x1) {
			int tmp = x1; x1 = x2; x2 = tmp;
			tmp = y1; y1 = y2; y2 = tmp;
		}
		int x = x1;
		float y = y1;
		float yIncr = (y2-y1) / (float) (x2-x1);
		while (x <= x2) {
			BYTE* bPtr = getDataXY(x, (int) y);
			*bPtr++ = r;
			*bPtr++ = g;
			*bPtr = b;
			x++;
			y += yIncr;
		}
	} else {
		if (y2 < y1) {
			int tmp = x1; x1 = x2; x2 = tmp;
			tmp = y1; y1 = y2; y2 = tmp;
		}
		int y = y1;
		float x = x1;
		float xIncr = (x2-x1) / (float) (y2-y1);
		while (y <= y2) {
			BYTE* bPtr = getDataXY((int) x, y);
			*bPtr++ = r;
			*bPtr++ = g;
			*bPtr = b;
			x += xIncr;
			y++;
		}
	}
}

void MBmpImg::HSL2RGB(MBmpImg* dest) {
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			BYTE* datPtr = getDataXY(x, y);
			BYTE h = *datPtr++;
			BYTE s = *datPtr++;
			BYTE l = *datPtr;
			int r, g, b;
			getRGB(h, s, l, &r, &g, &b);
			datPtr = dest->getDataXY(x, y);
			*datPtr++ = r;
			*datPtr++ = g;
			*datPtr = b;
		}
}

void MBmpImg::tmMedianRGB(MBmpImg* dest, float radius) {
	// generate median filtered image (round window shape, trimmed mean)
	static const int MAX_RADIUS = 6;
	static const int RAD_SQR = MAX_RADIUS * MAX_RADIUS;
	BYTE rgbGrayList[RAD_SQR*4][4];

	if (radius > MAX_RADIUS) radius = MAX_RADIUS;
	int radCeil = (int) ceil(radius);
	float radSqr = radius * radius;

	int listSize;
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			// collect pixel values inside the radius
			listSize = 0;
			for (int i=-radCeil; i<=radCeil; i++)
				for (int j=-radCeil; j<=radCeil; j++)
					if ((i*i + j*j <= radSqr) && (x+i >= 0) && (x+i < width)
							&& (y+j >= 0) && (y+j < height)) {
						BYTE* datPtr = getDataXY(x+i, y+j);
						BYTE r = *datPtr++;
						BYTE g = *datPtr++;
						BYTE b = *datPtr;
						rgbGrayList[listSize][0] = (((int) r) + g + b)/3;
						rgbGrayList[listSize][1] = r;
						rgbGrayList[listSize][2] = g;
						rgbGrayList[listSize][3] = b;
						listSize++;
					}

			// calculate number of entries to ignore
			// (average 10% in the middle)
			int delMin = (int) (listSize * 0.45);
			if (listSize - 2*delMin < 1) delMin++;
			int delMax = delMin;

			// kick out delMin smallest values
			while (delMin > 0) {
				int min = 255;
				int minInd = 0;
				// find min
				for (int i=0; i<listSize; i++)
					if (rgbGrayList[i][0] < min) {
						min = rgbGrayList[i][0];
						minInd = i;
					}
				// kick min
				listSize--;
				for (int li=0; li<4; li++) {
					rgbGrayList[minInd][li] = rgbGrayList[listSize][li];
				}
				delMin--;
			}

			// kick out delMax biggest values
			while (delMax > 0) {
				int max = 0;
				int maxInd = 0;
				// find max
				for (int i=0; i<listSize; i++)
					if (rgbGrayList[i][0] > max) {
						max = rgbGrayList[i][0];
						maxInd = i;
					}
				// kick max
				listSize--;
				for (int li=0; li<4; li++) {
					rgbGrayList[maxInd][li] = rgbGrayList[listSize][li];
				}
				delMax--;
			}

			// average of medium values
			int r = 0;
			int b = 0;
			int g = 0;
			for (int li=0; li<listSize; li++) {
				r += rgbGrayList[li][1];
				g += rgbGrayList[li][2];
				b += rgbGrayList[li][3];
			}
			r /= listSize;
			g /= listSize;
			b /= listSize;

			BYTE* datPtr = dest->getDataXY(x, y);
			*datPtr++ = (BYTE) r;
			*datPtr++ = (BYTE) g;
			*datPtr   = (BYTE) b;
		}
}

void MBmpImg::medianRGB3x3(MBmpImg* dest) { // faster than other version above (and slightly different)
	BYTE rgbList[9][3];

	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			// collect pixel values inside the radius
			int ind = 0;
			for (int i=-1; i<=1; i++)
				for (int j=-1; j<=1; j++) {
						BYTE* datPtr = getDataXY(CLIP(x+i, 0, width-1), CLIP(y+j, 0, height-1));
						BYTE r = *datPtr++;
						BYTE g = *datPtr++;
						BYTE b = *datPtr;
						rgbList[ind][0] = r;
						rgbList[ind][1] = g;
						rgbList[ind][2] = b;
						ind++;
					}

			int rgb[3];
			for (int i=0; i<3; i++) {
				bool val[9];
				for (int j=0; j<9; j++) val[j] = true;

				for (int j=0; j<4; j++) {
					// kick out min and max
					int min=255;	
					int minInd = 0;
					for (int j=0; j<9; j++) if (val[j] && (rgbList[j][i] < min)) {
						min = rgbList[j][i];
						minInd = j;
					}
					val[minInd] = false;

					int max=0;	
					int maxInd = 0;
					for (int j=0; j<9; j++) if (val[j] && (rgbList[j][i] > max)) {
						max = rgbList[j][i];
						maxInd = j;
					}
					val[maxInd] = false;
				}
				for (int j=0; j<9; j++) if (val[j]) rgb[i] = rgbList[j][i];
			}

			BYTE* datPtr = dest->getDataXY(x, y);
			*datPtr++ = (BYTE) rgb[0];
			*datPtr++ = (BYTE) rgb[1];
			*datPtr   = (BYTE) rgb[2];
		}
}

void MBmpImg::sobelRGB(MBmpImg* dest, float factor /*= 1.0*/) {
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) {
			// smoothing convoluted with edge detection: 1 2 1, 0 0 0, -1 -2 -1 matrix: horizontal + vertical
			int r = 0;
			int g = 0;
			int b = 0;
			if ((x>0) && (x<width-1) && (y>0) && (y<height-1)) {
				r = abs(((int) *getDataXY(x-1, y-1))
						  + ((int) *getDataXY(x-1,  y )) *2
						  + ((int) *getDataXY(x-1, y+1))
						  - ((int) *getDataXY(x+1, y-1))
						  - ((int) *getDataXY(x+1,  y )) *2
						  - ((int) *getDataXY(x+1, y+1)))
					  + abs(((int) *getDataXY(x-1, y-1))
						  + ((int) *getDataXY( x , y-1)) *2
						  + ((int) *getDataXY(x+1, y-1))
						  - ((int) *getDataXY(x-1, y+1))
						  - ((int) *getDataXY( x , y+1)) *2
						  - ((int) *getDataXY(x+1, y+1)));
				g = abs(((int) *(getDataXY(x-1, y-1)+1))
						  + ((int) *(getDataXY(x-1,  y )+1)) *2
						  + ((int) *(getDataXY(x-1, y+1)+1))
						  - ((int) *(getDataXY(x+1, y-1)+1))
						  - ((int) *(getDataXY(x+1,  y )+1)) *2
						  - ((int) *(getDataXY(x+1, y+1)+1)))
					  + abs(((int) *(getDataXY(x-1, y-1)+1))
						  + ((int) *(getDataXY( x , y-1)+1)) *2
						  + ((int) *(getDataXY(x+1, y-1)+1))
						  - ((int) *(getDataXY(x-1, y+1)+1))
						  - ((int) *(getDataXY( x , y+1)+1)) *2
						  - ((int) *(getDataXY(x+1, y+1)+1)));
				b = abs(((int) *(getDataXY(x-1, y-1)+2))
						  + ((int) *(getDataXY(x-1,  y )+2)) *2
						  + ((int) *(getDataXY(x-1, y+1)+2))
						  - ((int) *(getDataXY(x+1, y-1)+2))
						  - ((int) *(getDataXY(x+1,  y )+2)) *2
						  - ((int) *(getDataXY(x+1, y+1)+2)))
					  + abs(((int) *(getDataXY(x-1, y-1)+2))
						  + ((int) *(getDataXY( x , y-1)+2)) *2
						  + ((int) *(getDataXY(x+1, y-1)+2))
						  - ((int) *(getDataXY(x-1, y+1)+2))
						  - ((int) *(getDataXY( x , y+1)+2)) *2
						  - ((int) *(getDataXY(x+1, y+1)+2)));
			}

			r = (int) (r*factor);
			g = (int) (g*factor);
			b = (int) (b*factor);

			if (r > 255) r=255;
			if (g > 255) g=255;
			if (b > 255) b=255;

			BYTE* datPtr = dest->getDataXY(x, y);
			*datPtr++ = (BYTE) r;
			*datPtr++ = (BYTE) g;
			*datPtr   = (BYTE) b;
		}
}

void MBmpImg::sobelRGB2Gray(MBmpImg* dest, float factor /*= 1.0*/) {
	factor *= 0.3;
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) {
			// smoothing convoluted with edge detection: 1 2 1, 0 0 0, -1 -2 -1 matrix: horizontal + vertical
			int r = 0;
			int g = 0;
			int b = 0;
			if ((x>0) && (x<width-1) && (y>0) && (y<height-1)) {
				r = abs(((int) *getDataXY(x-1, y-1))
						  + ((int) *getDataXY(x-1,  y )) *2
						  + ((int) *getDataXY(x-1, y+1))
						  - ((int) *getDataXY(x+1, y-1))
						  - ((int) *getDataXY(x+1,  y )) *2
						  - ((int) *getDataXY(x+1, y+1)))
					  + abs(((int) *getDataXY(x-1, y-1))
						  + ((int) *getDataXY( x , y-1)) *2
						  + ((int) *getDataXY(x+1, y-1))
						  - ((int) *getDataXY(x-1, y+1))
						  - ((int) *getDataXY( x , y+1)) *2
						  - ((int) *getDataXY(x+1, y+1)));
				g = abs(((int) *(getDataXY(x-1, y-1)+1))
						  + ((int) *(getDataXY(x-1,  y )+1)) *2
						  + ((int) *(getDataXY(x-1, y+1)+1))
						  - ((int) *(getDataXY(x+1, y-1)+1))
						  - ((int) *(getDataXY(x+1,  y )+1)) *2
						  - ((int) *(getDataXY(x+1, y+1)+1)))
					  + abs(((int) *(getDataXY(x-1, y-1)+1))
						  + ((int) *(getDataXY( x , y-1)+1)) *2
						  + ((int) *(getDataXY(x+1, y-1)+1))
						  - ((int) *(getDataXY(x-1, y+1)+1))
						  - ((int) *(getDataXY( x , y+1)+1)) *2
						  - ((int) *(getDataXY(x+1, y+1)+1)));
				b = abs(((int) *(getDataXY(x-1, y-1)+2))
						  + ((int) *(getDataXY(x-1,  y )+2)) *2
						  + ((int) *(getDataXY(x-1, y+1)+2))
						  - ((int) *(getDataXY(x+1, y-1)+2))
						  - ((int) *(getDataXY(x+1,  y )+2)) *2
						  - ((int) *(getDataXY(x+1, y+1)+2)))
					  + abs(((int) *(getDataXY(x-1, y-1)+2))
						  + ((int) *(getDataXY( x , y-1)+2)) *2
						  + ((int) *(getDataXY(x+1, y-1)+2))
						  - ((int) *(getDataXY(x-1, y+1)+2))
						  - ((int) *(getDataXY( x , y+1)+2)) *2
						  - ((int) *(getDataXY(x+1, y+1)+2)));
			}

			int val = (int) ((r+g+b) * factor);
			if (val > 255) val = 255;
			*dest->getDataXY(x, y) = val;
		}
}

void MBmpImg::erodeGray(MBmpImg* dest, int size) {
	for (int y=0; y<height; y++)
	 for (int x=0; x<width; x++) { // all pixel
		BYTE min = 255;
		for (int i=-size/2; i<=size/2; i++)
		 for (int j=-size/2; j<=size/2; j++) {
			int xl = x + i; if (xl < 0) xl = 0; if (xl > width-1) xl = width-1;
			int yl = y + j; if (yl < 0) yl = 0; if (yl > height-1) yl = height-1;
			BYTE b = *getDataXY(xl, yl);
			if (b < min) min = b;
		}
		*dest->getDataXY(x, y) = min;
	}
}

void MBmpImg::dilateGray(MBmpImg* dest, int size) {
	for (int y=0; y<height; y++)
	 for (int x=0; x<width; x++) { // all pixel
		BYTE max = 0;
		for (int i=-size/2; i<=size/2; i++)
		 for (int j=-size/2; j<=size/2; j++) {
			int xl = x + i; if (xl < 0) xl = 0; if (xl > width-1) xl = width-1;
			int yl = y + j; if (yl < 0) yl = 0; if (yl > height-1) yl = height-1;
			BYTE b = *getDataXY(xl, yl);
			if (b > max) max = b;
		}
		*dest->getDataXY(x, y) = max;
	}
}

// convolutes rows and columns with given 1d signal
void MBmpImg::convolute(MBmpImg* tmpImg, float* factors, int num) {
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) {
			float sum = 0;
			for (int i=0; i<num; i++) sum += CLIP(*getDataXY(CLIP(x+i-num/2, 0, width-1), y) * factors[i], 0, 255);
			if (sum > 255) sum = 255;
			if (sum < 0) sum = 0;
			*tmpImg->getDataXY(x, y) = (BYTE) sum;
		}
	for (int y=0; y<height; y++)
		for (int x=0; x<=width-num; x++) {
			float sum = 0;
			for (int i=0; i<num; i++) sum += CLIP(*tmpImg->getDataXY(x, CLIP(y+i-num/2, 0, height-1)) * factors[i], 0, 255);
			if (sum > 255) sum = 255;
			if (sum < 0) sum = 0;
			*getDataXY(x, y) = (BYTE) sum;
		}
}

void MBmpImg::blurGaussian121(MBmpImg* tmpImg) {
	float gauss121[3] = {0.25, 0.5, 0.25};
	convolute(tmpImg, &gauss121[0], 3);
}

// label (enumerate) regions of 8 bit image (non-zero pixels; regions must not touch image borders!)
// return number of labels found
// work on same 8 bit image
int MBmpImg::label8() {
	int regNum = 0;
	int regLut[256];
	for (int i=0; i<256; i++) regLut[i] = i;
	for (int y=1; y<height; y++)
		for (int x=1; x<width; x++) { // all pixel
			if (*getDataXY(x, y)) {
				BYTE vNorth = *getDataXY(x, y-1);
				BYTE vWest = *getDataXY(x-1, y);
				BYTE v;

				if (vNorth && vWest) {
					v = regLut[vNorth];

					if (vNorth != vWest) {
						// merge regions
						int repVal = regLut[vWest];
						for (int i = 1; i<= regNum; i++)
							if (regLut[i] == repVal) regLut[i] = regLut[vNorth];
					}
				} else if (vNorth || vWest) {
					// same region
					v = vNorth + vWest; // one of them == 0
				} else {
					// found new region
					regNum++;
					v = regNum;
				}
				*getDataXY(x, y) = v;
			}
		}

	// regions are ready but not yet merged

	// merge regions using lut's
	int regLutLut[256];
	for (int i=0; i<256; i++) regLutLut[i] = 0;
	int rt = -1;
	for (int i=0; i<=regNum; i++) {
		if (regLutLut[regLut[i]] == 0) {
			rt++;
			regLutLut[regLut[i]] = rt;
		}
	}
	// replace pixel
	for (int y=1; y<height; y++)
		for (int x=1; x<width; x++) { // all pixel
			*getDataXY(x, y) = regLutLut[regLut[*getDataXY(x, y)]];
		}
	return rt;
}

// label (enumerate) regions of 8 bit image (non-zero pixels; regions must not touch image borders!)
// return number of labels found
// use temporary 16 bit image
int MBmpImg::label16() {
	MBmpImg tmp16(width, height, 16);

	int regNum = 0;
	/* TODO (#1#): could be optimized: normally not all 65535 values used 
	               but only a small part */
	int regLut[65536];
	for (int i=0; i<65536; i++) regLut[i] = i;
	for (int y=1; y<height; y++)
		for (int x=1; x<width; x++) { // all pixel
			if (*getDataXY(x, y)) {
				WORD vNorth = *(WORD*) tmp16.getDataXY(x, y-1);
				WORD vWest = *(WORD*) tmp16.getDataXY(x-1, y);
				WORD v;

				if (vNorth && vWest) {
					v = regLut[vNorth];

					if (vNorth != vWest) {
						// merge regions
						int repVal = regLut[vWest];
						for (int i = 1; i<= regNum; i++)
							if (regLut[i] == repVal) regLut[i] = regLut[vNorth];
					}
				} else if (vNorth || vWest) {
					// same region
					v = vNorth + vWest; // one of them == 0
				} else {
					// found new region
					regNum++;
					v = regNum;
				}

				*(WORD*) tmp16.getDataXY(x, y) = v;
			}
		}

	// regions are ready but not yet merged

	// merge regions using lut's
	int regLutLut[65536];
	for (int i=0; i<65536; i++) regLutLut[i] = 0;
	int rt = -1;
	for (int i=0; i<=regNum; i++) {
		if (regLutLut[regLut[i]] == 0) {
			rt++;
			regLutLut[regLut[i]] = rt;
		}
	}
	// replace pixel
	for (int y=1; y<height; y++)
		for (int x=1; x<width; x++) { // all pixel
			*getDataXY(x, y) = regLutLut[regLut[*(WORD*) tmp16.getDataXY(x, y)]];
		}
	return rt;
}

void MBmpImg::fillHisto(int *histoPtr) {
	for (int i=0; i<256; i++) histoPtr[i] = 0;
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) histoPtr[*getDataXY(x, y)]++;
}
