#include "MLInputDialog.h"
#include <stdlib.h>
#include "frontend.h"

MLInputDialog::MLInputDialog(HWND hwndParent, HINSTANCE hinst) {
	this->hwndParent = hwndParent;
	this->hinst = hinst;
}

bool MLInputDialog::getString(char* msg, char* defaultStr, char** answerPtr) {
	description = msg;
	value = defaultStr;
	if (doModal() != IDOK) return FALSE;
	*answerPtr = value;
	return TRUE;
}

bool MLInputDialog::getDoubleMinMax(char* msg, char* defaultStr, double min, double max, double* answerPtr) {
	char minStr[15];
	_gcvt(min, 10, minStr); // max 10 digits
	char maxStr[15];
	_gcvt(max, 10, maxStr); // max 10 digits

	description = new char[200];
	description[0] = 0; // clear string
	strcat(description, msg);
	strcat(description, "\n[");
	strcat(description, minStr);
	strcat(description, " .. ");
	strcat(description, maxStr);
	strcat(description, "]");
	BOOL parseError;
	do {
		value = defaultStr;
		if (doModal() != IDOK) {
			delete[] description;
			return FALSE;
		}
		char* stopStr;
		*answerPtr = strtod(value, &stopStr);
		parseError = (*stopStr != 0);
	} while ((parseError) || (*answerPtr < min) || (*answerPtr > max));
	delete[] description;
	return TRUE;
}

bool MLInputDialog::getLongMinMax(char* msg, char* defaultStr, long min, long max, long* answerPtr) {
	char minStr[20];
	_ltoa(min, minStr, 10); // base 10 (decimal)
	char maxStr[20];
	_ltoa(max, maxStr, 10); // base 10 (decimal)

	description = new char[300];
	description[0] = 0; // clear string
	strcat(description, msg);
	strcat(description, "\n[");
	strcat(description, minStr);
	strcat(description, " .. ");
	strcat(description, maxStr);
	strcat(description, "]");
	BOOL parseError;
	do {
		value = defaultStr;
		if (doModal() != IDOK) {
			delete[] description;
			return FALSE;
		}
		char* stopStr;
		*answerPtr = strtol(value, &stopStr, 10); // base 10 (decimal)
		parseError = (*stopStr != 0);
	} while ((parseError) || (*answerPtr < min) || (*answerPtr > max));
	delete[] description;
	return TRUE;
}

bool MLInputDialog::getBool(char* msg, BOOL* answerPtr) {
	int result = MessageBox(NULL, msg, "input box", MB_ICONQUESTION | MB_YESNOCANCEL);
	if (result == IDCANCEL) return FALSE;
	*answerPtr = (result == IDYES);
	return TRUE;
}

// using list box; -1 is returned when cancelled
int MLInputDialog::getIndexFromList(char* msg, char* stringList) {
	description = msg;
	value = stringList;

	doModalList();

	return listIndex;
}

char szItemName[80]; // receives name of item
BOOL CALLBACK MLInputDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_INITDIALOG: {
			MLInputDialog* mlid = (MLInputDialog*) lParam;
			SetDlgItemText(hwndDlg, IDC_STATIC_DESCRIPTION, mlid->description);
			SetDlgItemText(hwndDlg, IDC_EDIT1, mlid->value);
			/* TODO (#1#): SetFocus doesn't work */
			SetFocus(GetDlgItem(hwndDlg, IDC_EDIT1));
			return TRUE;}
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
					if (!GetDlgItemText(hwndDlg, IDC_EDIT1, szItemName, 80)) *szItemName=0;
					// fall through
				case IDCANCEL:
					EndDialog(hwndDlg, wParam);
					return TRUE;
			}
	}
	return FALSE;
}

int returnIndex; // receives index of list item
BOOL CALLBACK MLInputDialogProcList(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_INITDIALOG: {
			MLInputDialog* mlid = (MLInputDialog*) lParam;
			SetDlgItemText(hwndDlg, IDC_STATIC_DESCRIPTION, mlid->description);
			HWND hwndList = GetDlgItem(hwndDlg, IDC_LIST1);
			// fill list box
			char* tmp = mlid->value;
			while (strlen(tmp)) {
				SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM) tmp);
				tmp += strlen(tmp) + 1;
			}
			//SendMessage(hwndList, LB_SETITEMDATA, i, (LPARAM) i);
            /* TODO (#1#): SetFocus doesn't work */
            SetFocus(hwndList);
			return TRUE;}
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_LIST1:
					if (HIWORD(wParam) != CBN_DBLCLK) break;
					// else: double click: fall through
				case IDOK: {
					HWND hwndList = GetDlgItem(hwndDlg, IDC_LIST1);
					returnIndex = SendMessage(hwndList, LB_GETCURSEL, 0, 0);
					EndDialog(hwndDlg, wParam);
					return TRUE;}
				case IDCANCEL:
					returnIndex = -1;
					EndDialog(hwndDlg, wParam);
					return TRUE;
			}
	}
	return FALSE;
}

int MLInputDialog::doModal() {
	int r = DialogBoxParam(hinst, MAKEINTRESOURCE(IDD_DIALOG_ML_INPUT_DIALOG),
		hwndParent, (DLGPROC) MLInputDialogProc, (LPARAM) this);
// TODO: now uses global buffer --> several inputboxes confirmed at the same time may get confused
	value = (char*) &szItemName;
	return r;
}

int MLInputDialog::doModalList() {
	int r = DialogBoxParam(hinst, MAKEINTRESOURCE(IDD_DIALOG_ML_INPUT_DIALOG_LIST),
		hwndParent, (DLGPROC) MLInputDialogProcList, (LPARAM) this);
// TODO: now uses global buffer --> several inputboxes confirmed at the same time may get confused
	listIndex = returnIndex;
	return r;
}
