/*
 * Marching Cubes extension with color and normal per vertex
 * for 3dscanner project
 *
 *		Marcel Lancelle 2002
 */

#include "MarchingCubesVertexExt.h"
#include <windows.h>
#include <gl/gl.h>
#include "mathext.h"
#include <io.h>

void MarchingCubesVertexExt::drawFlatShaded(bool colored) {
	glEnable(GL_COLOR_MATERIAL);

	glVertexPointer(3, GL_FLOAT, 0, &coords[0]); // 3 coords, type, stride = 0 (-> tightly packed), data pointer
	glEnableClientState(GL_VERTEX_ARRAY);
	if (colored) {
		glColorPointer(3, GL_UNSIGNED_BYTE, 0, &vertexColors[0]); // 3 colors, type (color format), stride = 0 (-> tightly packed), data pointer
		glEnableClientState(GL_COLOR_ARRAY);
	} else {
		glColor3f(0.8, 0.8, 0.8);
	}

	glBegin(GL_TRIANGLES);
	for (int t=0; t<triangles.size(); t++) { // all triangles
		glNormal3fv(&faceNormals[t].n[0]);
		for (int v=0; v<3; v++) // all three vertices
			glArrayElement(triangles[t].t[v]);
	}
	glEnd();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

	glDisable(GL_COLOR_MATERIAL);
}

void MarchingCubesVertexExt::drawSmoothShaded(bool colored) {
	glEnable(GL_COLOR_MATERIAL);

	glVertexPointer(3, GL_FLOAT, 0, &coords[0]); // 3 coords, type, stride = 0 (-> tightly packed), data pointer
	glEnableClientState(GL_VERTEX_ARRAY);
	glNormalPointer(GL_FLOAT, 0, &vertexNormals[0]); // type, stride = 0 (-> tightly packed), data pointer
	glEnableClientState(GL_NORMAL_ARRAY);
	if (colored) {
		glColorPointer(3, GL_UNSIGNED_BYTE, 0, &vertexColors[0]); // 3 colors, type (color format), stride = 0 (-> tightly packed), data pointer
		glEnableClientState(GL_COLOR_ARRAY);
	} else {
		glColor3f(0.8, 0.8, 0.8);
	}

	glBegin(GL_TRIANGLES);
	for (int t=0; t<triangles.size(); t++) { // all triangles
		for (int v=0; v<3; v++) // all three vertices
			glArrayElement(triangles[t].t[v]);
	}
	glEnd();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

	glDisable(GL_COLOR_MATERIAL);

/*
	for (int t=0; t<triangles.size(); t++) { // all triangles
	glBegin(GL_TRIANGLES);

		for (int v=0; v<3; v++) { // all three vertices
			int ind = triangles[t].t[v];
			glNormal3fv(&vertexNormals[ind].n[0]);
if (colored) glColor3ubv(&vertexColors[ind].r);
//if (colored) glMaterial(GL_FRONT, );
			glVertex3f(coords[ind].c[0], coords[ind].c[1], coords[ind].c[2]);
		}

	glEnd();
/ *
glDisable(GL_LIGHTING);
glDisable(GL_CULL_FACE);
glEnable(GL_POLYGON_OFFSET_LINE);
glPolygonOffset(-1.0, -1.0);
glBegin(GL_LINE_LOOP);
for (int v=0; v<3; v++) { // all three vertices
int ind = triangles[t].t[v];
glVertex3f(coords[ind].c[0], coords[ind].c[1], coords[ind].c[2]);
}
glEnd();
glEnable(GL_LIGHTING);
glEnable(GL_CULL_FACE);
glDisable(GL_POLYGON_OFFSET_LINE);
* /
	}*/
}

// writes Model to the .wrl VRML file format (VRML97)
bool MarchingCubesVertexExt::write2VRMLFile(char* vrmlFileName, bool exportColor /*= true*/) {
	FILE* vrmlFile;
	vrmlFile = fopen(vrmlFileName, "w");
	if (vrmlFile == NULL) return false;

	fprintf(vrmlFile, "#VRML V2.0 utf8\n\n");
	fprintf(vrmlFile, "#3DScanner Marching Cubes output\n\n");

	fprintf(vrmlFile, "Shape {\n");
	fprintf(vrmlFile, " geometry IndexedFaceSet {\n");
	fprintf(vrmlFile, "  solid TRUE\n");
	fprintf(vrmlFile, "  ccw TRUE\n");
	if (exportColor) fprintf(vrmlFile, "  colorPerVertex TRUE\n");
	fprintf(vrmlFile, "  normalPerVertex TRUE\n");

	fprintf(vrmlFile, "  coord Coordinate {\n");
	fprintf(vrmlFile, "   point [\n");
	// vertices
	for (int i=0; i<coords.size(); i++) {
		fprintf(vrmlFile, "%f %f %f,\n", coords[i].c[0], coords[i].c[1], coords[i].c[2]);
	}
	fprintf(vrmlFile, "   ]  # end point\n");
	fprintf(vrmlFile, "  } # end coord\n");

	if (exportColor) {
		// color values
		fprintf(vrmlFile, "  color Color {\n");
		fprintf(vrmlFile, "  color [\n");
		for (int i=0; i<vertexColors.size(); i++) { // all colors
			fprintf(vrmlFile, "%f %f %f\n", vertexColors[i].r/255.0, vertexColors[i].g/255.0, vertexColors[i].b/255.0);
		}
		fprintf(vrmlFile, "  ] # end inner color group\n");
		fprintf(vrmlFile, " } # end color node\n");
	}

	// normals
	fprintf(vrmlFile, "  normal Normal {\n");
	fprintf(vrmlFile, "  vector [\n");
	for (int i=0; i<vertexNormals.size(); i++) { // all normals
		fprintf(vrmlFile, "%f %f %f,\n", vertexNormals[i].n[0], vertexNormals[i].n[1], vertexNormals[i].n[2]);
	}
	fprintf(vrmlFile, "  ] # end inner normal group\n");
	fprintf(vrmlFile, " } # end normal node\n");

	// triangle faces: index list
	fprintf(vrmlFile, "  coordIndex [\n");
	for (int i=0; i<triangles.size(); i++) { // all triangles
		fprintf(vrmlFile, "%i, %i, %i, -1,\n", triangles[i].t[0], triangles[i].t[1], triangles[i].t[2]);
	}
	fprintf(vrmlFile, "  ] # end coordIndex\n");


	fprintf(vrmlFile, " } # end geometry\n");

	if (!exportColor) {
		fprintf(vrmlFile, " appearance Appearance {\n");
		fprintf(vrmlFile, "  material Material { diffuseColor 0.7 0.7 0.7 }\n");
		fprintf(vrmlFile, " }\n");
	}

	fprintf(vrmlFile, "} # end shape\n");

	fclose(vrmlFile);
	return true;
}

void MarchingCubesVertexExt::terminate() {
	MarchingCubes::terminate();

	faceNormals.resize(triangles.size());
	vertexNormals.resize(coords.size());
	vertexColors.resize(coords.size());

	calcFaceNormals();
	calcVertexNormals();
}

void MarchingCubesVertexExt::calcFaceNormals() {
	for (int t=0; t<triangles.size(); t++) { // all triangles
		// calc and set normal
		float a[3], b[3], n[3];
		int ind0 = triangles[t].t[0];
		int ind1 = triangles[t].t[1];
		int ind2 = triangles[t].t[2];
		for (int i=0; i<3; i++) {
			a[i] = coords[ind1].c[i] - coords[ind0].c[i];
			b[i] = coords[ind2].c[i] - coords[ind0].c[i];
		}
		n[0] = a[1]*b[2] - a[2]*b[1];
		n[1] = a[2]*b[0] - a[0]*b[2];
		n[2] = a[0]*b[1] - a[1]*b[0];

		// normalize
		float l = sqrt(sqr(n[0]) + sqr(n[1]) + sqr(n[2]));
		if (l != 0) {
			faceNormals[t].n[0] = n[0] / l;
			faceNormals[t].n[1] = n[1] / l;
			faceNormals[t].n[2] = n[2] / l;
		} else {
			faceNormals[t].n[0] = 0;
			faceNormals[t].n[1] = 0;
			faceNormals[t].n[2] = 0;
		}
	}
}

/* TODO (#1#): unweighted sum is used --> introduce some weights! (see nelson's method) */
void MarchingCubesVertexExt::calcVertexNormals() {
	// all normals = 0
	for (int v=0; v<coords.size(); v++) { // all vertices
		for (int i=0; i<3; i++) vertexNormals[v].n[i] = 0;
	}

	// add normal of each triangle to each of its points (better would be to introduce a weight: face area?)
	for (int t=0; t<triangles.size(); t++) { // all triangles
		// set normal
		int ind0 = triangles[t].t[0];
		int ind1 = triangles[t].t[1];
		int ind2 = triangles[t].t[2];
		for (int i=0; i<3; i++) {
			vertexNormals[ind0].n[i] += faceNormals[t].n[i];
			vertexNormals[ind1].n[i] += faceNormals[t].n[i];
			vertexNormals[ind2].n[i] += faceNormals[t].n[i];
		}
	}

	// normalize normals
	for (int v=0; v<coords.size(); v++) { // all vertices
		float l = sqrt(sqr(vertexNormals[v].n[0]) + sqr(vertexNormals[v].n[1]) + sqr(vertexNormals[v].n[2]));
		for (int i=0; i<3; i++) vertexNormals[v].n[i] /= l;
	}
}
