/*
 * These functions provide port access in Win9x via assembler calls
 * and in Win NT/2K/ME via the PortTalk driver.
 *
 * If you don't know which system your app runs on, try with io driver first
 * and check the portAccessLastError code; when using direct access under
 * Win NT/2K/ME, your app will crash.
 *
 * This implementation uses the recommended but slow port access method;
 * if you need fast access, use the PortTalk methods modifing the io permission
 * bitmap.
 *
 * See PortTalk doc for more info.
 * don't forget to include 'porttalk.sys' in your distribution package
 *
 *
 *		Marcel Lancelle 2002
 */

#include "PortAccess.h"
#include "buildOptions.h"

/*
 * much code taken from IoExample of PortTalk:
 *
 * IoExample for PortTalk V2.1
 * Version 2.1, 12th January 2002
 * http://www.beyondlogic.org
 *
 */

#include <windows.h>
#include <stdio.h>
#include <winioctl.h>
#include "PortTalk_IOCTL.h"

bool portAccessUseIODriver;
HANDLE PortTalk_Handle; // Handle for PortTalk Driver
int portAccessLastError = PORTACCES_OK;

bool portAccessBegin(bool useIODriver) {
	portAccessUseIODriver = useIODriver;

	if (!useIODriver) {
		portAccessLastError = PORTACCES_OK;
		return true;
	}

	// open PortTalk driver
	PortTalk_Handle = CreateFile("\\\\.\\PortTalk", GENERIC_READ, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (PortTalk_Handle != INVALID_HANDLE_VALUE) {
		portAccessLastError = PORTACCES_OK;
		return true;
	}

	// else try to start driver
	SC_HANDLE SchSCManager;
	SC_HANDLE schService;
	BOOL ret;
	DWORD err;

	// Open Handle to Service Control Manager
    SchSCManager = OpenSCManager(NULL,			// machine (NULL == local)
						NULL,					// database (NULL == default)
						SC_MANAGER_ALL_ACCESS); // access required

	if ((SchSCManager == NULL) && (GetLastError() == ERROR_ACCESS_DENIED)) {
		// We do not have enough rights to open the SCM, therefore we must
		// be a poor user with only user rights.
		//printf("PortTalk: You do not have rights to access the Service Control Manager and\n");
		//printf("PortTalk: the PortTalk driver is not installed or started. Please ask \n");
		//printf("PortTalk: your administrator to install the driver on your behalf.\n");
		portAccessLastError = PORTACCES_ACCESS_DENIED;
		return false;
	}

	// Open a Handle to the PortTalk Service Database
	schService = OpenService(SchSCManager,	// handle to service control manager database
					"PortTalk",				// pointer to name of service to start
					SERVICE_ALL_ACCESS);	// type of access to service

	if (schService == NULL)
		switch (GetLastError()){
			case ERROR_ACCESS_DENIED:
				//printf("PortTalk: You do not have rights to the PortTalk service database\n");
				portAccessLastError = PORTACCES_ACCESS_DENIED;
				return false;
			case ERROR_INVALID_NAME:
				//printf("PortTalk: The specified service name is invalid.\n");
				portAccessLastError = PORTACCES_UNKNOWN_ERROR;
				return false;
			case ERROR_SERVICE_DOES_NOT_EXIST:
				//printf("PortTalk: The PortTalk driver does not exist. Installing driver.\n");
				//printf("PortTalk: This can take up to 30 seconds on some machines . .\n");
				//InstallPortTalkDriver();
				portAccessLastError = PORTACCES_PORTTALK_NOT_INSTALLED;
				return false;
			case ERROR_CALL_NOT_IMPLEMENTED: // ML
				//printf("PortTalk: you are not running Win NT/2K/ME, so you don't need PortTalk!\n");
				portAccessLastError = PORTACCES_IOCTL_NOT_IMPLEMENTED;
				return false;
		}

	// Start the PortTalk Driver. Errors will occur here if PortTalk.SYS file doesn't exist

	ret = StartService(schService,	// service identifier
				0,				// number of arguments
				NULL);			// pointer to arguments

    if (ret) {
		//printf("PortTalk: The PortTalk driver has been successfully started.\n");
	} else {
		err = GetLastError();
		if (err == ERROR_SERVICE_ALREADY_RUNNING) {
			//printf("PortTalk: The PortTalk driver is already running.\n");
		} else {
			//printf("PortTalk: Unknown error while starting PortTalk driver service.\n");
			//printf("PortTalk: Does PortTalk.SYS exist in your \\System32\\Drivers Directory?\n");
			portAccessLastError = PORTACCES_UNKNOWN_ERROR;
			return false;
		}
	}

	// Close handle to Service Control Manager
	CloseServiceHandle (schService);
	return(TRUE);

	// try to open once more
	PortTalk_Handle = CreateFile("\\\\.\\PortTalk", GENERIC_READ, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (PortTalk_Handle == INVALID_HANDLE_VALUE) {
		//printf("PortTalk: Couldn't access PortTalk Driver, Please ensure driver is loaded.\n\n");
		portAccessLastError = PORTACCES_UNKNOWN_ERROR;
		return false;
	}

	portAccessLastError = PORTACCES_OK;
	return true;
}

void portAccessEnd() {
	// close PortTalk driver
	if (portAccessUseIODriver) CloseHandle(PortTalk_Handle);
}

//-----------------------------------------------------------------------
//  functions
//      inport  - inputs a word from a hardware port
//      inportb - inputs a byte from a hardware port
//      inp     - inputs a byte from a hardware port
//      inpw    - inputs a word from a hardware port
//
//      outport  - outputs a word to a hardware port
//      outportb - outputs a byte to a hardware port
//      outp     - outputs a byte to a hardware port
//      outpw    - outputs a word to a hardware port
//-----------------------------------------------------------------------


// use these for Intel assembler syntax
/*
unsigned short inport(unsigned short port)
{
    asm     mov     dx,port
    asm     in      ax,dx
        return _AX;
}

unsigned char inportb(unsigned short port)
{
    asm     mov     dx,port
    asm     in      al,dx
    asm     xor     ah,ah
        return _AX;
}

unsigned short inpw(unsigned short port)
{
    asm     mov     dx,port
    asm     in      ax,dx
        return _AX;
}

short int inp(unsigned short port)
{
    asm     mov     dx,port
    asm     in      al,dx
    asm     xor     ah,ah
        return _AX;
}

void outport(unsigned short port, unsigned short val)
{
    asm     mov     dx, port
    asm     mov     ax, val
    asm     out     dx, ax
}

void outportb(unsigned short port, unsigned char val)
{
    asm     mov     dx, port
    asm     mov     al, val
    asm     out     dx, al
}

unsigned short outpw(unsigned short port, unsigned short val)
{
    asm     mov     dx, port
    asm     mov     ax, val
    asm     out     dx, ax
        return (_AX);
}

short int outp(unsigned short port, unsigned char val)
{
    asm     mov     dx, port
    asm     mov     al, val
    asm     out     dx, al
    asm     xor     ah, ah
        return (_AX);
}
*/

// use these for AT&T assembler syntax
/*
unsigned short inport(unsigned short port)
{
    asm     mov     dx,port
    asm     in      ax,dx
        return _AX;
}

unsigned char inportb(unsigned short port)
{
    asm     mov     dx,port
    asm     in      al,dx
    asm     xor     ah,ah
        return _AX;
}

unsigned short inpw(unsigned short port)
{
    asm     mov     dx,port
    asm     in      ax,dx
        return _AX;
}

short int inp(unsigned short port)
{
    asm     mov     dx,port
    asm     in      al,dx
    asm     xor     ah,ah
        return _AX;
}

void outport(unsigned short port, unsigned short val)
{
    asm     mov     dx, port
    asm     mov     ax, val
    asm     out     dx, ax
}

void outportb(unsigned short port, unsigned char val)
{
    asm     mov     dx, port
    asm     mov     al, val
    asm     out     dx, al
}

unsigned short outpw(unsigned short port, unsigned short val)
{
    asm     mov     dx, port
    asm     mov     ax, val
    asm     out     dx, ax
        return (_AX);
}
*/

// quick and dirty (and slow) workaround that only global variables can be accessed in the AT&T basic syntax
unsigned short global_port;
unsigned short global_val;

void outp(unsigned short port, unsigned char value) {
	if (portAccessUseIODriver) {
		unsigned int errorVal;
		DWORD BytesReturned;
		unsigned char Buffer[3];
		unsigned short * pBuffer;
		pBuffer = (unsigned short *)&Buffer[0];
		*pBuffer = port;
		Buffer[2] = value;

		errorVal = DeviceIoControl(PortTalk_Handle,
					IOCTL_WRITE_PORT_UCHAR,
					&Buffer,
					3,
					NULL,
					0,
					&BytesReturned,
					NULL);

		if (errorVal) {
			portAccessLastError = PORTACCES_OK;
		} else {
			//printf("Error occured during outportb while talking to PortTalk driver %d\n",GetLastError());
			portAccessLastError = PORTACCES_IOCTL_ERROR;
		}
	} else {
		global_port = port;
		global_val = value;
#ifdef INLINE_ASM_INTEL
		asm mov dx, _global_port
		asm mov ax, _global_val
		asm out dx, ax
#elseif
	    __asm("mov _global_port,%dx");
	    __asm("mov _global_val,%al");
	    __asm("out %al,%dx");
#endif
	}
}

//void outp(unsigned short PortAddress, unsigned char byte);
//unsigned char inp(unsigned short PortAddress);

//#define    inp(PortAddress)         inportb(PortAddress)
//#define    outp(PortAddress, Value) outportb(PortAddress, Value)

/*
unsigned char inp(unsigned short PortAddress)
{
    unsigned int error;
    DWORD BytesReturned;
    unsigned char Buffer[3];
    unsigned short * pBuffer;
    pBuffer = (unsigned short *)&Buffer;
    *pBuffer = PortAddress;

    error = DeviceIoControl(PortTalk_Handle,
                            IOCTL_READ_PORT_UCHAR,
                            &Buffer,
                            2,
                            &Buffer,
                            1,
                            &BytesReturned,
                            NULL);

    if (!error) printf("Error occured during inportb while talking to PortTalk driver %d\n",GetLastError());
    return(Buffer[0]);
}
*/

bool InstallPortTalkDriver() {
	SC_HANDLE SchSCManager;
	SC_HANDLE schService;
	DWORD err;
	CHAR  DriverFileName[80];

    /* Get Current Directory. Assumes PortTalk.SYS driver is in this directory.    */
    /* Doesn't detect if file exists, nor if file is on removable media - if this  */
    /* is the case then when windows next boots, the driver will fail to load and  */
    /* a error entry is made in the event viewer to reflect this */

    /* Get System Directory. This should be something like c:\windows\system32 or  */
    /* c:\winnt\system32 with a Maximum Character lenght of 20. As we have a       */
    /* buffer of 80 bytes and a string of 24 bytes to append, we can go for a max  */
    /* of 55 bytes */

	if (!GetSystemDirectory(DriverFileName, 55)) {
		//printf("PortTalk: Failed to get System Directory. Is System Directory Path > 55 Characters?\n");
		//printf("PortTalk: Please manually copy driver to your system32/driver directory.\n");
		return false;
	}

	// Append our Driver Name
	lstrcat(DriverFileName,"\\Drivers\\PortTalk.sys");
	//printf("PortTalk: Copying driver to %s\n",DriverFileName);

	// Copy Driver to System32/drivers directory. This fails if the file doesn't exist.

    if (!CopyFile("PortTalk.sys", DriverFileName, FALSE)) {
		//printf("PortTalk: Failed to copy driver to %s\n",DriverFileName);
		//printf("PortTalk: Please manually copy driver to your system32/driver directory.\n");
		return false;
	}

	// Open Handle to Service Control Manager
    SchSCManager = OpenSCManager (NULL,		// machine (NULL == local)
                                  NULL,		// database (NULL == default)
                                  SC_MANAGER_ALL_ACCESS);	// access required

	// Create Service/Driver - This adds the appropriate registry keys in
	// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services - It doesn't
	// care if the driver exists, or if the path is correct.

    schService = CreateService (SchSCManager,                      /* SCManager database */
                                "PortTalk",                        /* name of service */
                                "PortTalk",                        /* name to display */
                                SERVICE_ALL_ACCESS,                /* desired access */
                                SERVICE_KERNEL_DRIVER,             /* service type */

// ML Bugfix
                                SERVICE_AUTO_START, //SERVICE_DEMAND_START,              /* start type */

                                SERVICE_ERROR_NORMAL,              /* error control type */
                                "System32\\Drivers\\PortTalk.sys", /* service's binary */
                                NULL,                              /* no load ordering group */
                                NULL,                              /* no tag identifier */
                                NULL,                              /* no dependencies */
                                NULL,                              /* LocalSystem account */
                                NULL                               /* no password */
                                );

	if (schService == NULL) {
		err = GetLastError();
		//if (err == ERROR_SERVICE_EXISTS)
		//	printf("PortTalk: Driver already exists. No action taken.\n");
		//	else printf("PortTalk: Unknown error while creating Service.\n");
		return false;
    }
//	else printf("PortTalk: Driver successfully installed.\n");

	// Close Handle to Service Control Manager
    CloseServiceHandle(schService);
	return true;
}

bool UninstallPortTalkDriver() {
	/******************************************************************************/
	/*                                                                            */
	/*                       Remove Driver for PortTalk  V2.0                     */
	/*                        Version 2.0, 12th January 2002                      */
	/*                          http://www.beyondlogic.org                        */
	/*                                                                            */
	/* Copyright  2002 Craig Peacock. Craig.Peacock@beyondlogic.org              */
	/* Any publication or distribution of this code in source form is prohibited  */
	/* without prior written permission of the copyright holder. This source code */
	/* is provided "as is", without any guarantee made as to its suitability or   */
	/* fitness for any particular use. Permission is herby granted to modify or   */
	/* enhance this sample code to produce a derivative program which may only be */
	/* distributed in compiled object form only.                                  */
	/******************************************************************************/

    SC_HANDLE       SchSCManager;
    SC_HANDLE       schService;
    BOOL            ret;
    SERVICE_STATUS  serviceStatus;

    //printf("Uninstall for PortTalk V2.0\nCopyright 2002 Craig Peacock\nhttp://www.beyondlogic.org\n");

    /* Open Handle to Service Control Manager */
    SchSCManager = OpenSCManager (NULL,                   // machine (NULL == local)
                                  NULL,                   // database (NULL == default)
                                  SC_MANAGER_ALL_ACCESS); // access required

    /* Open Handle to PortTalk Service Database */
    schService = OpenService (SchSCManager,
                              "PortTalk",
                              SERVICE_ALL_ACCESS);

    if (schService == NULL) {
        //printf("Error while opening PortTalk service, has PortTalk been installed?\n");
        return false;
    }

    /* Stop Service */
    ret = ControlService (schService,
                          SERVICE_CONTROL_STOP,
                          &serviceStatus);

//    if (ret) printf("PortTalk service has been successfully stopped.\n");
//    else     printf("Unknown error while stopping PortTalk service.\n");

    /* Delete Service */
    ret = DeleteService (schService);

//    if (ret) printf("PortTalk service has been successfully deleted.\n");
//    else     printf("Error removing PortTalk service - PortTalk has NOT been successfully removed.");
	if (!ret) return false;

    //printf("You may now re-install PortTalk.\n");

    /* Close Handle to Porttalk Service Database */
    CloseServiceHandle (schService);

    /* Close Handle to Service Control Manager */
    CloseServiceHandle(SchSCManager);

	return true;
}
