/**
 * These functions provide port access in Win9x via assembler calls
 * and in Win NT/2K/ME via the PortTalk driver.
 *
 * If you don't know which system your app runs on, try with io driver first
 * and check the portAccessLastError code; when using direct access under
 * Win NT/2K/ME, your app will crash.
 *
 * This implementation uses the recommended but slow port access method;
 * if you need fast access, use the PortTalk methods modifing the io permission
 * bitmap.
 *
 * See PortTalk doc for more info.
 * don't forget to include 'porttalk.sys' in your distribution package
 *
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef PORTACCESS_HEADER_INCLUDED
#define PORTACCESS_HEADER_INCLUDED

#define _outp outp

#define PORTACCES_OK						0 /// no error
#define PORTACCES_IOCTL_ERROR				1 /// writing to port via ioctl not successfull
#define PORTACCES_PORTTALK_NOT_INSTALLED	2 /// porttalk driver required but not installed
#define PORTACCES_ACCESS_DENIED				3 /// you need admin rights to install porttalk
#define PORTACCES_IOCTL_NOT_IMPLEMENTED 	4 /// probably using 9x, so don't use io driver
#define PORTACCES_UNKNOWN_ERROR				5 /// unknown error case
/// last error code stored here:
extern int portAccessLastError;

// following functions all set portAccessLastError
extern bool portAccessBegin(bool useIODriver); /// useIODriver = true: use PortTalk; false: use direct assembler port access
extern void portAccessEnd();
extern bool InstallPortTalkDriver();
extern bool UninstallPortTalkDriver();

void outp(unsigned short port, unsigned char value);

//unsigned char  inportb( unsigned short __portid );
//unsigned short inport ( unsigned short __portid );
//short int      inp( unsigned short __portid );
//unsigned short inpw( unsigned short __portid );
//void           outportb( unsigned short __portid, unsigned char __value );
//void           outport ( unsigned short __portid, unsigned short __value );
//unsigned short outpw( unsigned short __portid, unsigned short __value );

#endif // PORTACCESS_HEADER_INCLUDED
