#include "QuaternionRotation.h"
#include <math.h>

QuaternionRotation::QuaternionRotation() {
	reset();
}

void QuaternionRotation::reset() { // getRotMatrix will return identity matrix afterwards
	a = 1.0f;
	b = 0.0f;
	c = 0.0f;
	d = 0.0f;
}

void QuaternionRotation::rotate(float angle, float vx, float vy, float vz) { // v is rotation unit vector
	// calculate r which is rotation Quaternion
	float ra = (float) cos(angle / 2);
	float f = (float) sin(angle / 2);
	float rb = vx * f;
	float rc = vy * f;
	float rd = vz * f;

	// save old Quaternion
	float oa = a;
	float ob = b;
	float oc = c;
	float od = d;

	// multiply the two Quaternions: quat = quatRotation * quat
	a = ra*oa - rb*ob - rc*oc - rd*od;
	b = ra*ob + rb*oa + rc*od - rd*oc;
	c = ra*oc - rb*od + rc*oa + rd*ob;
	d = ra*od + rb*oc - rc*ob + rd*oa;
}

void QuaternionRotation::getRotMatrix(float m[4][4]) { // in OpenGL format (transposed)
	m[0][0] = 1.0 - 2.0*(c*c + d*d); //a*a + b*b - c*c - d*d;
	m[1][0] = 2.0 * (b*c - a*d);
	m[2][0] = 2.0 * (b*d + a*c);
	m[3][0] = 0.0;

	m[0][1] = 2.0f * (b*c + a*d);
	m[1][1] = 1.0 - 2.0*(b*b + d*d); //a*a - b*b + c*c - d*d;
	m[2][1] = 2.0f * (c*d - a*b);
	m[3][1] = 0.0f;

	m[0][2] = 2.0f * (b*d - a*c);
	m[1][2] = 2.0f * (c*d + a*b);
	m[2][2] = 1.0 - 2.0*(b*b + c*c); //a*a - b*b - c*c + d*d;
	m[3][2] = 0.0f;

	m[0][3] = 0.0;
	m[1][3] = 0.0;
	m[2][3] = 0.0;
	m[3][3] = 1.0;
}
