#include "ScanProject.h"
#include <string>
#include <stdio.h>
#include <gl/gl.h>
#include "frontend.h"

const int MAX_LINESIZE = 400; // max size of a single line

ScanProject::ScanProject(char* fileName) {
	// copy file name
	this->fileName = new char[strlen(fileName)+1];
	strcpy(this->fileName, fileName);
	isTextured = false;

	// fill projectPath and projectName
	char* ext;
	splitPathNameExt(fileName, &projectPath, &projectName, &ext);
	delete[] ext;

	// reset all member variables
	modified = false;
	lastErrorLine = 0;
	minX = 0;
	maxX = 0;
	minY = 0;
	maxY = 0;
	minZ = 0;
	maxZ = 0;
	resolution = 0.01;
	gridSizeX = 3;
	gridSizeY = 3;
	gridSizeZ = 3;

//	photoNum = 0;
//	photosTaken = 0;
	mcPtr = NULL;
}

ScanProject::~ScanProject() {
	if ((modified) && (MessageBox(NULL, "Project has been changed. Save now?", "3DScanner", MB_ICONQUESTION | MB_YESNOCANCEL) == IDYES)) write();
	if (mcPtr) delete mcPtr;
	delete[] fileName;
}

// check 'lastErrorLine' if not successful (0 means: file couldn't be read)
bool ScanProject::read() {
	modified = false;
	lastErrorLine = 0;
	FILE* sFile;
	if ((sFile = fopen(fileName, "r")) == NULL) {
		MessageBox(NULL, "Couldn't open project file!\n", "3DScanner", MB_ICONERROR | MB_OK);
		return false;
	}

	char buffer[MAX_LINESIZE]; // buffer for a single line
	char* bufferStr = (char *) &buffer;
	int lineNum = 0;
	Photo* currentPhotoPtr = NULL;
	CamDef* currentCamDefPtr = NULL;
	ParseState parseState = PSTATE_NULL;

	while (fgets(bufferStr, MAX_LINESIZE, sFile) != NULL) { // while not error / eof
		lineNum++;
		if (strcmp(bufferStr, "\n") != 0) { // ignore empty lines
			// delete '\n' at end of line
			bufferStr[strlen(bufferStr)-1] = 0;

			// seek for '=' character
			char* tmpPtr = bufferStr;
			while ((*tmpPtr) && (*tmpPtr != '=')) tmpPtr++;

			bool processed = false;

			if (*tmpPtr == '=') { // if '=' is in line
				*tmpPtr = 0; // replace '=' with string separator character
				tmpPtr++;
				char* valPtr = tmpPtr;

				switch (parseState) {
					case PSTATE_3DSCAN:
						if (strcmp(bufferStr, "minX") == 0)
							processed = sscanf(valPtr, "%f", &minX);
						if (strcmp(bufferStr, "maxX") == 0)
							processed = sscanf(valPtr, "%f", &maxX);
						if (strcmp(bufferStr, "minY") == 0)
							processed = sscanf(valPtr, "%f", &minY);
						if (strcmp(bufferStr, "maxY") == 0)
							processed = sscanf(valPtr, "%f", &maxY);
						if (strcmp(bufferStr, "minZ") == 0)
							processed = sscanf(valPtr, "%f", &minZ);
						if (strcmp(bufferStr, "maxZ") == 0)
							processed = sscanf(valPtr, "%f", &maxZ);
						break;
					case PSTATE_3DSCAN_PHOTO:
						if (strcmp(bufferStr, "mask") == 0) {
							currentPhotoPtr->setMaskFileName(valPtr);
							processed = true;
						}
						if (strcmp(bufferStr, "name") == 0) {
							currentPhotoPtr->setFileName(valPtr);
							processed = true;
						}
						if (strcmp(bufferStr, "cam") == 0) {
							for (int i=0; i<camDefs.size(); i++)
								if (strcmp(valPtr, camDefs[i].name) == 0) currentPhotoPtr->camDefInd = i;
							processed = (currentPhotoPtr->camDefInd != -1);
						}
						if (strcmp(bufferStr, "camRotationAngle") == 0)
							processed = sscanf(valPtr, "%f", &currentPhotoPtr->scanParams.camRotationAngle);
						break;
					case PSTATE_3DSCAN_CAM:
						if (strcmp(bufferStr, "name") == 0) {
							currentCamDefPtr->setName(valPtr);
							processed = true;
						}
						if (strcmp(bufferStr, "f") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.f);
						if (strcmp(bufferStr, "k1") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.k1);
						if (strcmp(bufferStr, "tx") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.tx);
						if (strcmp(bufferStr, "ty") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.ty);
						if (strcmp(bufferStr, "tz") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.tz);
						if (strcmp(bufferStr, "rx") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.rx);
						if (strcmp(bufferStr, "ry") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.ry);
						if (strcmp(bufferStr, "rz") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.rz);
						if (strcmp(bufferStr, "ax") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->rotationAxis.v[0]);
						if (strcmp(bufferStr, "ay") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->rotationAxis.v[1]);
						if (strcmp(bufferStr, "az") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->rotationAxis.v[2]);
						if (strcmp(bufferStr, "cx") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.cx);
						if (strcmp(bufferStr, "cy") == 0)
							processed = sscanf(valPtr, "%f", &currentCamDefPtr->camTrafo.cy);
						if (strcmp(bufferStr, "w") == 0)
							processed = sscanf(valPtr, "%i", &currentCamDefPtr->camTrafo.imgWidth);
						if (strcmp(bufferStr, "h") == 0)
							processed = sscanf(valPtr, "%i", &currentCamDefPtr->camTrafo.imgHeight);
						break;
				}
			} else {
				// no '=' is in line
				switch (parseState) {
					case PSTATE_NULL:
						if (strcmp(bufferStr, "<3dscan>") == 0) {
							parseState = PSTATE_3DSCAN;
							processed = true;
						}
						break;
					case PSTATE_3DSCAN:
						if (strcmp(bufferStr, "<cam>") == 0) {
							currentCamDefPtr = new CamDef();
							parseState = PSTATE_3DSCAN_CAM;
							processed = true;
						}
						if (strcmp(bufferStr, "<photo>") == 0) {
							currentPhotoPtr = new Photo();
							parseState = PSTATE_3DSCAN_PHOTO;
							processed = true;
						}
						if (strcmp(bufferStr, "</3dscan>") == 0) {
							parseState = PSTATE_NULL;
							processed = true;
						}
						break;
					case PSTATE_3DSCAN_CAM:
						if (strcmp(bufferStr, "</cam>") == 0) {
							if (!currentCamDefPtr) break; // error: no opening <cam> before!
							currentCamDefPtr->camTrafo.calcTrafoMatrix();
							camDefs.push_back(*currentCamDefPtr);
							delete currentCamDefPtr;
							currentCamDefPtr = NULL;
							parseState = PSTATE_3DSCAN;
							processed = true;
						}
						break;
					case PSTATE_3DSCAN_PHOTO:
						if (strcmp(bufferStr, "</photo>") == 0) {
							if (!currentPhotoPtr) break; // error: no opening <photo> before!
							currentPhotoPtr->calcTrafo(&camDefs[0]);
							photos.push_back(*currentPhotoPtr);
							delete currentPhotoPtr;
							currentPhotoPtr = NULL;
							parseState = PSTATE_3DSCAN;
							processed = true;
						}
						break;
				}
			}

			if (!processed) {
				lastErrorLine = lineNum;
				fclose(sFile);
				return false;
			}
		}
	} // while not error / eof

	fclose(sFile);

	if (minX >= maxX) return false;
	if (minY >= maxY) return false;
	if (minZ >= maxZ) return false;

/*
	// fill transformation matrices
	for (int i=0; i<photos.size(); i++) {
		photos[i].calcTrafo();
	}
*/
	return true;
}

bool ScanProject::write() {
	if (camDefs.size() < 1) {
		MessageBox(NULL, "Project can't be saved without camera definition!", "3DScanner", MB_ICONINFORMATION | MB_OK);
		return false;
	}

	modified = false;

	FILE* sFile;
	if ((sFile = fopen(fileName, "w")) == NULL) {
		MessageBox(NULL, "Couldn't open file!\n", "3DScanner", MB_ICONERROR | MB_OK);
		return false;
	}

	fprintf(sFile, "<3dscan>\n\n");

	fprintf(sFile, "minX=%f\n", minX);
	fprintf(sFile, "maxX=%f\n", maxX);
	fprintf(sFile, "minY=%f\n", minY);
	fprintf(sFile, "maxY=%f\n", maxY);
	fprintf(sFile, "minZ=%f\n", minZ);
	fprintf(sFile, "maxZ=%f\n", maxZ);
	fprintf(sFile, "\n");

	for(int i=0; i<camDefs.size(); i++) {
		fprintf(sFile, "<cam>\n");
		fprintf(sFile, "name=%s\n", camDefs[i].name);
		fprintf(sFile, "f=%f\n", camDefs[i].camTrafo.f);
		fprintf(sFile, "k1=%f\n", camDefs[i].camTrafo.k1);
		fprintf(sFile, "tx=%f\n", camDefs[i].camTrafo.tx);
		fprintf(sFile, "ty=%f\n", camDefs[i].camTrafo.ty);
		fprintf(sFile, "tz=%f\n", camDefs[i].camTrafo.tz);
		fprintf(sFile, "rx=%f\n", camDefs[i].camTrafo.rx);
		fprintf(sFile, "ry=%f\n", camDefs[i].camTrafo.ry);
		fprintf(sFile, "rz=%f\n", camDefs[i].camTrafo.rz);
		fprintf(sFile, "ax=%f\n", camDefs[i].rotationAxis.v[0]);
		fprintf(sFile, "ay=%f\n", camDefs[i].rotationAxis.v[1]);
		fprintf(sFile, "az=%f\n", camDefs[i].rotationAxis.v[2]);
		fprintf(sFile, "cx=%f\n", camDefs[i].camTrafo.cx);
		fprintf(sFile, "cy=%f\n", camDefs[i].camTrafo.cy);
		fprintf(sFile, "w=%i\n", camDefs[i].camTrafo.imgWidth);
		fprintf(sFile, "h=%i\n", camDefs[i].camTrafo.imgHeight);
		fprintf(sFile, "</cam>\n\n");
	}

	for(int i=0; i<photos.size(); i++) {
		fprintf(sFile, "<photo>\n");
		if (photos[i].fileName) fprintf(sFile, "name=%s\n", photos[i].fileName);
		if (photos[i].maskFileName) fprintf(sFile, "mask=%s\n", photos[i].maskFileName);
		if (photos[i].camDefInd != -1) fprintf(sFile, "cam=%s\n", camDefs[photos[i].camDefInd].name);
		fprintf(sFile, "camRotationAngle=%f\n", photos[i].scanParams.camRotationAngle);
		fprintf(sFile, "</photo>\n\n");
	}

	fprintf(sFile, "</3dscan>\n");

	fclose(sFile);
}

void ScanProject::showInfo(HWND hwnd) {
	char buf[300];
	sprintf((char*) buf, "%s\n\nCamera Definitions: %i\nPhotos: %i", fileName, camDefs.size(), photos.size());
	if (mcPtr) {
		char buf2[50];
		sprintf((char*) buf2, "\n\nVertices: %i\nTriangles: %i", mcPtr->coords.size(), mcPtr->triangles.size());
		strcat((char*) &buf, (char*) &buf2);
	}
	MessageBox(hwnd, buf, "3DScanner Project Info", MB_ICONINFORMATION | MB_OK);
}

////////////////////////////////////////////////////////////////////////
//////// CamDef

ScanProject::CamDef::CamDef() {
	name = NULL;
	rotationAxis.v[0] = 0.0;
	rotationAxis.v[1] = 0.0;
	rotationAxis.v[2] = 1.0;
	rotationAxis.v[3] = 1.0;
}

ScanProject::CamDef::CamDef(const ScanProject::CamDef& cd) {
	name = NULL;
	if (cd.name) setName(cd.name);
	camTrafo = cd.camTrafo;
	rotationAxis = cd.rotationAxis;
}

ScanProject::CamDef::~CamDef() {
	if (name) delete[] name;
	name = NULL;
}

void ScanProject::CamDef::setName(char* newName) {
	if (name) delete[] name;
	name = new char[strlen(newName)+1];
	strcpy(name, newName);
}

////////////////////////////////////////////////////////////////////////
//////// Photo

ScanProject::Photo::Photo() {
	fileName = NULL;
	maskFileName = NULL;
//	resetMat(&camParams.trafo);
	camDefInd = -1;
	scanParams.camRotationAngle = 0;
}

ScanProject::Photo::Photo(const ScanProject::Photo& ph) {
	fileName = NULL;
	maskFileName = NULL;
	if (ph.fileName) setFileName(ph.fileName);
	if (ph.maskFileName) setMaskFileName(ph.maskFileName);

	camTrafo = ph.camTrafo;
	camDefInd = ph.camDefInd;
	scanParams = ph.scanParams;

	for (int i=0; i<5; i++) for (int j=0; j<3; j++)
	camVis[i][j] = ph.camVis[i][j]; // for visualisation only
}

ScanProject::Photo::~Photo() {
	if (maskFileName) delete[] maskFileName;
	maskFileName = NULL;
	if (fileName) delete[] fileName;
	fileName = NULL;
}

void ScanProject::Photo::setFileName(char* newName) {
	if (fileName) delete[] fileName;
	fileName = new char[strlen(newName)+1];
	strcpy(fileName, newName);
}

void ScanProject::Photo::setMaskFileName(char* newName) {
	if (maskFileName) delete[] maskFileName;
	maskFileName = new char[strlen(newName)+1];
	strcpy(maskFileName, newName);
}

void ScanProject::Photo::calcTrafo(CamDef* camDefs) {
	camTrafo.copyFrom(&camDefs[camDefInd].camTrafo);
	camTrafo.calcTrafoMatrixRotateAxis(&camDefs[camDefInd].rotationAxis, scanParams.camRotationAngle);

/*
	// copy fxy
	camParams.fxy = camDefPtr->scanParams.fxy;

	// calculate trafo matrix
	resetMat(&camParams.trafo);
	VEC4 t;
	resetVec(&t);
	t.v[0] = -camDefPtr->scanParams.camPos[0];
	t.v[1] = -camDefPtr->scanParams.camPos[1]; // distance
	t.v[2] = -camDefPtr->scanParams.camPos[2]; // height
	rotateZ(&camParams.trafo, -scanParams.camRotationAngle); // alpha
	translate(&camParams.trafo, &t);
	rotateX(&camParams.trafo, 90-camDefPtr->scanParams.camVAngle); // beta
*/

	// calc camera visualisation points
	int iw = camTrafo.imgWidth - 1;
	int ih = camTrafo.imgHeight - 1;

	for (int x=0; x<2; x++) for (int y=0; y<2; y++) {
		int ind = x + 2*y + 1;
		camTrafo.img2World(x*iw, y*ih, &camVis[0][0], &camVis[0][1], &camVis[0][2], &camVis[ind][0], &camVis[ind][1], &camVis[ind][2]);
		VEC4 a, b;
		resetVec(&a);
		resetVec(&b);
		for (int i=0; i<3; i++) {
			a.v[i] = camVis[0][i] * 0.001;
			b.v[i] = camVis[ind][i];
		}
		normalizeVec3(&b);
		scaleVec(&b, getVec3Length(&a) * 0.4);
		addVec(&b, &a);
		for (int i=0; i<3; i++) {
			camVis[0][i] = a.v[i];
			camVis[ind][i] = b.v[i];
		}
	}
}

void ScanProject::visualizeCams() {
	for (int i=0; i<photos.size(); i++) {
/*		double ax, ay, az, bx[4], by[4], bz[4];
		int iw = project->photos[i].camTrafo.imgWidth - 1;
		int ih = project->photos[i].camTrafo.imgHeight - 1;

		for (int x=0; x<2; x++) for (int y=0; y<2; y++) {
			int ind = x + 2*y;
			project->photos[i].camTrafo.img2World(x*iw, y*ih, &ax, &ay, &az, &bx[ind], &by[ind], &bz[ind]);
			VEC4 a, b;
			resetVec(&a);
			resetVec(&b);
			a.v[0] = ax * 0.001;
			a.v[1] = ay * 0.001;
			a.v[2] = az * 0.001;
			b.v[0] = bx[ind];
			b.v[1] = by[ind];
			b.v[2] = bz[ind];
			normalizeVec3(&b);
			scaleVec(&b, getVec3Length(&a) * 0.4);
			addVec(&b, &a);
			ax = a.v[0];
			ay = a.v[1];
			az = a.v[2];
			bx[ind] = b.v[0];
			by[ind] = b.v[1];
			bz[ind] = b.v[2];
		}
*/
		glBegin(GL_TRIANGLES);
			glColor4f(0.4, 0.4, 0.4, 0.5);
			glVertex3f(photos[i].camVis[0][0], photos[i].camVis[0][1], photos[i].camVis[0][2]);
			glColor4f(0.5, 0.5, 0.5, 0.0);
			glVertex3f(photos[i].camVis[1][0], photos[i].camVis[1][1], photos[i].camVis[1][2]);
			glVertex3f(photos[i].camVis[2][0], photos[i].camVis[2][1], photos[i].camVis[2][2]);
			
			glColor4f(0.4, 0.4, 0.4, 0.5);
			glVertex3f(photos[i].camVis[0][0], photos[i].camVis[0][1], photos[i].camVis[0][2]);
			glColor4f(0.5, 0.5, 0.5, 0.0);
			glVertex3f(photos[i].camVis[2][0], photos[i].camVis[2][1], photos[i].camVis[2][2]);
			glVertex3f(photos[i].camVis[4][0], photos[i].camVis[4][1], photos[i].camVis[4][2]);
			
			glColor4f(0.4, 0.4, 0.4, 0.5);
			glVertex3f(photos[i].camVis[0][0], photos[i].camVis[0][1], photos[i].camVis[0][2]);
			glColor4f(0.5, 0.5, 0.5, 0.0);
			glVertex3f(photos[i].camVis[3][0], photos[i].camVis[3][1], photos[i].camVis[3][2]);
			glVertex3f(photos[i].camVis[4][0], photos[i].camVis[4][1], photos[i].camVis[4][2]);
			
			glColor4f(0.4, 0.4, 0.4, 0.5);
			glVertex3f(photos[i].camVis[0][0], photos[i].camVis[0][1], photos[i].camVis[0][2]);
			glColor4f(0.5, 0.5, 0.5, 0.0);
			glVertex3f(photos[i].camVis[3][0], photos[i].camVis[3][1], photos[i].camVis[3][2]);
			glVertex3f(photos[i].camVis[1][0], photos[i].camVis[1][1], photos[i].camVis[1][2]);
		glEnd();
	}
}
