/**
 * class for 3dscan projects
 * provides file parsing and writing
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef SCANPROJECT_HEADER_INCLUDED
#define SCANPROJECT_HEADER_INCLUDED

#include <vector>
using namespace std;
#include "math3d.h"
#include "CamTrafo.h"
#include "MarchingCubesVertexExt.h"
#include <windows.h>

class ScanProject {
  public:
	ScanProject(char* fileName);
	~ScanProject();

	bool read(); ///< check 'lastErrorLine' if not successful (0 means: file couldn't be read)
	bool write();
	void showInfo(HWND hwnd);
//	int photoNum, photosTaken; // number of photos to take / already taken
	MarchingCubesVertexExt* mcPtr;
	void visualizeCams();

	class CamDef {
	  public:
		CamDef();
		CamDef(const CamDef& cd);
		~CamDef();

		void setName(char* newName);
		char* name;

		CamTrafo camTrafo;
		VEC4 rotationAxis;
/*
		struct ScanParams {
			float camPos[3]; // in world coords
			float camVAngle; // angle in degrees (0 means horizontal; positive means pointing down)
			float fxy; // distance cam origin - image plane [in pixel]
		} scanParams;
*/
	};

	class Photo {
	  public:
		Photo();
		Photo(const Photo& ph);
		~Photo();

		void setMaskFileName(char* newName);
		char* maskFileName;
		void setFileName(char* newName);
		char* fileName;

		void calcTrafo(CamDef* camDefs);
		CamTrafo camTrafo; ///< not stored to project file

		int camDefInd; ///< index in camDefs
		struct ScanParams {
			float camRotationAngle; // angle in degrees of turntable rotation (clockwise rotation of object seen from above)
		} scanParams;

		double camVis[5][3]; ///< for visualisation only
	};

	vector<Photo> photos;
	vector<CamDef> camDefs;

	float minX, maxX, minY, maxY, minZ, maxZ; ///< volume borders (world metric)
	int gridSizeX, gridSizeY, gridSizeZ; ///< number of cubes in each direction
	float resolution; ///< cube edge length in meter
	static const int CLIP_XY_PLANE_ZSUB = 2; ///< # cubes in negative z-axis for smooth object clipping to xy plane

	int lastErrorLine;
	bool modified;
	bool isTextured;

	char* projectPath; ///< path, e.g. "c:\\scans\\"
	char* projectName; ///< name without extension, e.g. "test object"

  protected:
	char* fileName; ///< *.3dscan

	enum ParseState {
		PSTATE_NULL, PSTATE_3DSCAN, PSTATE_3DSCAN_CAM, PSTATE_3DSCAN_PHOTO
	};
};

#endif // SCANPROJECT_HEADER_INCLUDED
