/**
 * \mainpage 3DScanner Documentation
 *
 * This software is a complete application for scanning 3d models using
 * the 'shape from silhouettes' approach.
 *
 * The main class is ScannerApp and is called by the frontend.
 *
 * For many classes you find test programs in the source directory.
 * The doxygen doc is poor, check the files themselves for more details.
 *
 * The software is freeware, distributed under GNU general public license and
 * provided as is.
 * Marcel Lancelle 2002-2003.
 */

/**
 * application providing command processing for the frontend
 * for 3dscanner project
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef SCANNERAPP_HEADER_INCLUDED
#define SCANNERAPP_HEADER_INCLUDED

#include <windows.h>
#include <gl/glu.h>

#include "MarchingCubesVertexExt.h"
#include "CamAccess.h"
#include "CalibrateCam.h"
#include "StepperMotorParPort.h"
#include "SimpleIni.h"
#include "MBmpImg.h"
#include "ScanProject.h"
#include "QuaternionRotation.h"
#include "math3d.h"

#define GET_IMG_START		0
#define GET_IMG_CONTINUE	1

class ScannerApp {
  public:
	void init(HWND hwnd, HINSTANCE hInstance);
	void initGl();
	void deinit();
	bool preTranslateMessage(LPMSG msgPtr);
	void draw(); ///< draws OpenGL scene
	void readIniFile();
	void writeIniFile();

	void fileNew();
	void fileOpen();
	void fileOpen(char* fileName);
	void fileSave();
	void fileMergeWithProjectPrepare();
	void fileMergeWithProjectApply(bool cancel);
	void fileShowProjectInfo();
	void fileExportModel2VRML();
	void fileExportModel2Obj();

	void scanRotateTurnTable(int steps);

	void scanCalibrateCam();
	void scanCalibrateCam1();
	void scanGetImages();
	void scanSegmentImages();
	void scanPrepareManualSegmentation();
	void scanGenerate3DModel(ScanProject* prj);
	void scanFitScanningVolume2LastScan(ScanProject* prj);
	void scanSimplifyModel(ScanProject* prj);
	void scanTexture(ScanProject* prj);

	void configPortTalkInstall();
	void configPortTalkUninstall();
	void configSMReset();
	void configSMSave1Rot();
	void configSMSetPort(int portAdr);
	void configCamSelectSource();
	void configCamOptions();
	void configCamShowPreviewWindow();
	void configSelectCalibrationPattern();
	void configSetIAT(CamAccess::ImageAquisitionType iat);

	ScanProject* project;
	ScanProject* mergeProject;
	QuaternionRotation mergeRotation;
	VEC4 mergeTranslation;
	MAT4x4 mergeMatrix;
	char* exePath; ///< path of 3dscanner.exe, e.g. "c:\prg\3dscanner\"

	bool showGrid;
	bool showTexture;
	bool showSmoothShaded;
	bool showMesh;
	bool showNormals;
	bool showViews;

	bool rotateManually;
	CamAccess::ImageAquisitionType imageAquisitionType;

	enum AppState{APP_STATE_NORMAL, APP_STATE_MERGE} appState;
	CalibrateCam::CalibrationPattern calibrationPattern;

  protected:
	bool portDriverOk(); ///< open PortTalk if not yet opened
	void rotateTurnTableDeg(int angle); ///< beware of numerical errors when using this repeatedly!
	void calcMaxScanningVolume();

	float lastScanMinX, lastScanMaxX, lastScanMinY, lastScanMaxY, lastScanMinZ, lastScanMaxZ;

	CamAccess camAccess;
	StepperMotorParPort smp;
	bool stepperMotorUsed; ///< is it allready been used? for PortTalk
	int stepCounter; ///< counter for stepper motor, used in scanRotateTurnTable() and for calibrating
	float stepNum360; ///< steps for 360 degree rotation of turn table
	int portAdr;
	bool debugOutput; ///< output images for debugging purposes

	HWND hwnd;
	HINSTANCE hInstance;
};

#endif // SCANNERAPP_HEADER_INCLUDED
