/*
 * class for a basic ini configuration file support
 * for 3dscanner project
 *
 * !! lines have max size (see below)
 * !! space characters are NOT whitespace characters
 * !! lines starting with ' are ignored (comment lines)
 * !! when writing to file, comment lines are deleted
 *
 *		Marcel Lancelle 2002
 */

#include "SimpleIni.h"
#include <stdio.h>

const int BUFFERSIZE = 200; // max size of a single line

// reads ini file into array
SimpleIni::SimpleIni(char* fileName) {
	this->fileName = fileName;
	FILE* iniFile;
	if ((iniFile = fopen(fileName, "r")) == NULL) {
		fprintf(stderr, "couldn't open file %s!\n", fileName);
		return;
	}

	char buffer[BUFFERSIZE]; // buffer for a single line
	char* bufferStr = (char *) &buffer;

	while (fgets(bufferStr, BUFFERSIZE, iniFile) != NULL) { // while not error / eof
		if ((strcmp(bufferStr, "\n") != 0)		// ignore empty lines
  			&& (strcmp(bufferStr, "'") != 0)) { // and comment lines
			// seek for '=' character
			char* tmpPtr = bufferStr;
			while ((*tmpPtr) && (*tmpPtr != '=')) tmpPtr++;

			if (*tmpPtr == '=') { // if '=' is in line
				*tmpPtr = 0; // replace '=' with string separator character
				tmpPtr++;
				char* valPtr = tmpPtr;
				// delete '\n' at end of line
				while ((*tmpPtr) && (*tmpPtr != '\n')) tmpPtr++;
				*tmpPtr = 0;

				IniEntry iniEntry;

				// allocate mem and copy strings
				iniEntry.key = new char[strlen(bufferStr)+1];
    			strcpy(iniEntry.key, bufferStr);
				iniEntry.value = new char[strlen(valPtr)+1];
    			strcpy(iniEntry.value, valPtr);

				entries.push_back(iniEntry);
			}
		}
	}

	fclose(iniFile);
}

SimpleIni::~SimpleIni() {
	for (int i=0; i<entries.size(); i++) {
		delete[] entries[i].key;
		delete[] entries[i].value;
	}
}

// compares every string in array until first occurance is found
// returns NULL if not existant
char* SimpleIni::getEntry(char* key) {
	for (int i=0; i<entries.size(); i++) {
		if (strcmp(key, entries[i].key) == 0) return entries[i].value;
	}
	return NULL;
}

// returns true if value could be read
bool SimpleIni::getIntEntry(char* key, int *val) {
	char* str = getEntry(key);
	if (str == NULL) return false;
	return sscanf(str, "%i", val);
}

// returns true if value could be read
bool SimpleIni::getFloatEntry(char* key, float *val) {
	char* str = getEntry(key);
	if (str == NULL) return false;
	return sscanf(str, "%f", val);
}
