#include <windows.h>
#include "SplashWindow.h"

#define TIMER_ID 42

// static member vars
HWND SplashWindow::hwnd;
WNDCLASS SplashWindow::wndCl;
HGDIOBJ SplashWindow::hbmSplash;
BITMAP SplashWindow::splashBmp;
DWORD SplashWindow::threadID;
bool SplashWindow::active;

SplashWindow::SplashWindow(HINSTANCE hInstance, WORD bmpId) {
	hbmSplash = LoadBitmap(hInstance, MAKEINTRESOURCE(bmpId));
	GetObject(hbmSplash, sizeof(splashBmp), (LPSTR) &splashBmp);

	char* className = "SplashWndClass";
	wndCl.style = 0;
	wndCl.lpfnWndProc = &wndProc;
	wndCl.cbClsExtra = 0;
	wndCl.cbWndExtra = 0;
	wndCl.hInstance = hInstance;
	wndCl.hIcon = NULL;
	wndCl.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndCl.hbrBackground = (HBRUSH) GetStockObject(NULL_BRUSH);
	wndCl.lpszMenuName = NULL;
	wndCl.lpszClassName = className;
	RegisterClass(&wndCl);

	int scrWidth = GetSystemMetrics(SM_CXSCREEN);
	int scrHeight = GetSystemMetrics(SM_CYSCREEN);

	hwnd = CreateWindowEx(WS_EX_TOPMOST, className, "SplashWnd",
				WS_POPUP, //  | WS_VISIBLE
				(scrWidth-splashBmp.bmWidth)/2, (scrHeight-splashBmp.bmHeight)/2,
				splashBmp.bmWidth, splashBmp.bmHeight,
				NULL, NULL, hInstance, NULL);
}

void SplashWindow::show(int milliSeconds) {
	threadID = 0;
	HANDLE hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) messageLoop, NULL, 0, &threadID);
	SetThreadPriority(hThread, THREAD_PRIORITY_BELOW_NORMAL);
	SetTimer(hwnd, TIMER_ID, milliSeconds, NULL);
}

DWORD SplashWindow::messageLoop(void* paramPtr) {
	ShowWindow(hwnd, SW_SHOWNORMAL/*SW_SHOW*/);
	SetForegroundWindow(hwnd);

	MSG msg;
	active = true;
	while (active && GetMessage(&msg, hwnd, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return 0;
}

LRESULT CALLBACK SplashWindow::wndProc(HWND hwnd, UINT msg,
	WPARAM wParam, LPARAM lParam) {

	switch (msg) {
		case WM_LBUTTONDOWN:
		case WM_KILLFOCUS:
		case WM_NCLBUTTONDOWN:
		case WM_KEYDOWN:
		case WM_CLOSE:
			KillTimer(hwnd, TIMER_ID);
		case WM_TIMER:
			ShowWindow(hwnd, SW_HIDE);
			active = false;

//			PostQuitMessage(0);
			return 0;
/*
		case WM_PAINT: {
			PAINTSTRUCT paintStruct;
			HDC hdc = BeginPaint(hwnd, &paintStruct);
			HDC hMemDC = CreateCompatibleDC(hdc);
			HGDIOBJ hbmOld = SelectObject(hMemDC, hbmSplash);
			BitBlt(hdc,0,0,splashBmp.bmWidth,splashBmp.bmHeight,hMemDC,0,0,SRCCOPY);
			SelectObject(hMemDC, hbmOld);
			DeleteDC(hMemDC);
			//TextOut(hdc, 20, 20, ":-) bla bla", 11);
			EndPaint(hwnd, &paintStruct);
			return 0;}
*/

		case WM_ERASEBKGND:{
			HDC hMemDC = CreateCompatibleDC((HDC) wParam);
			HGDIOBJ hbmOld = SelectObject(hMemDC, hbmSplash);
			BitBlt((HDC) wParam,0,0,splashBmp.bmWidth,splashBmp.bmHeight,hMemDC,0,0,SRCCOPY);
			SelectObject(hMemDC, hbmOld);
			DeleteDC(hMemDC);
			return 0;}

		default:
			return DefWindowProc(hwnd,msg,wParam,lParam);
	}

	return 0;
}
