/**
 * stepper motor controlling software for own stepper motor interface
 * for 3dscanner project
 *
 * parallel port interface (low level)
 * bipolar stepper motor (two coils)
 *
 * \warning be sure to enable port access before! see PortAccess.*
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef STEPPERMOTORPARPORT_HEADER_INCLUDED
#define STEPPERMOTORPARPORT_HEADER_INCLUDED

/*
#ifndef bool
enum {false, true};
typedef int	bool;
#endif // bool
*/

#pragma warning (disable:4244)		/* disable bogus conversion warnings */

#define POWER_OFF		0
#define POWER_NORMAL	1	///< one winding powered at a time
#define POWER_HIGH		2	/// two windings powered at a time
							/// (needs twice energy; multiplies torque by about 1.4)
#define POWER_HALFSTEP	3	///< half stepping mode

class StepperMotorParPort {

  public:
	StepperMotorParPort();
	~StepperMotorParPort();

	/// setPower: sets one of the modes above
	/// think at turning off power after use!
	void setPower(int powerMode);

	/// setPortAdr: 0x0278, 0x0378 are common adresses for the parallel port
	void setPortAdr(int port);

	/** rotateDirection: rotates in one direction
	 * when you don't specify a delay think at leaving enough time
	 * between rotations to stay in valid rotation frequencies
	 * direction: real direction depends on how motor is connected
	 * msDelay: delay in ms between two rotation states
	 * stepNum: number of steps to rotate (negative values allowed)
	 */
	void rotate(bool direction);
	void rotate(bool direction, int msDelay, int stepNum = 1);

	/** rotateAccelerated: linear speed changes
	 * direction: real direction depends on how motor is connected
	 * stepNum: number of steps to rotate (negative values allowed)
	 * minMsDelay: shortest delay specifies max rotation frequency (in ms)
	 * accMsTime: accelerate during this period (in ms)
	 *    ______________
	 *   /              \
	 *  /                \  time - speed diagram
	 * /                  \
	 */
	void rotateAccelerated(bool direction, int stepNum, int minMsDelay, int accMsTime);

  protected:
	int powerMode;		///< one of the modes above, initially: 0 (off)
	int port;			///< port adress, initially: -1 (invalid)
	char rotationState;	///< lower 2 bits are counter for state

	void updatePort();
};

#endif // STEPPERMOTORPARPORT_HEADER_INCLUDED
