#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "QuaternionRotation.h"
#include "math3d.h"

#define PI 3.14159

QuaternionRotation q;
int oldMouseX, oldMouseY;
int actButton;

struct CamParams {
	float ox, oy; // Bildursprung in Pixeln
	float t[3]; // translation
	float r[3][3]; // rotation
	float fxy; // distance origin - image plane [in pixel]
};
CamParams camParams;

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0, 0, -1.5);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

{
glPushMatrix();
MAT4x4 m;
resetMat(&m);
rotateX(&m, 80); // beta
VEC4 t;
resetVec(&t);
t.v[1] = -0.3;
t.v[2] = 0.2;
translate(&m, &t);
rotateZ(&m, 0); // alpha

	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_QUADS);
		resetVec(&t);
		t.v[0] = 0.1;
		t.v[1] = 0.1;
		multMatVec(&m, &t);
		glVertex3fv(&t.v[0]);

		resetVec(&t);
		t.v[0] = -0.1;
		t.v[1] = 0.1;
		multMatVec(&m, &t);
		glVertex3fv(&t.v[0]);

		resetVec(&t);
		t.v[0] = -0.1;
		t.v[1] = -0.1;
		multMatVec(&m, &t);
		glVertex3fv(&t.v[0]);

		resetVec(&t);
		t.v[0] = 0.1;
		t.v[1] = -0.1;
		multMatVec(&m, &t);
		glVertex3fv(&t.v[0]);
	glEnd();
glPopMatrix();
}

	glColor3f(0.6, 0.6, 1.0);

	for (int i=-10; i<=10; i++) {
		glBegin(GL_LINES);
			glVertex3f(i*0.1, -1.0, 0.0);
			glVertex3f(i*0.1, +1.0, 0.0);
		glEnd();
	}

	for (int i=-10; i<=10; i++) {
		glBegin(GL_LINES);
			glVertex3f(-1, i*0.1, 0.0);
			glVertex3f(+1, i*0.1, 0.0);
		glEnd();
	}

	glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 0.0, 1.0);
	glEnd();

	glColor3f(1.0, 1.0, 1.0);

	glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(camParams.t[0], camParams.t[1], 0);
	glEnd();

	glBegin(GL_LINES);
		glVertex3f(camParams.t[0], camParams.t[1], 0);
		glVertex3f(camParams.t[0], camParams.t[1], camParams.t[2]);
	glEnd();

	// draw cam
	glPushMatrix();
	glTranslatef(camParams.t[0], camParams.t[1], camParams.t[2]);
//	glTranslatef(0, 0.2, 0.3);
glutSolidCube(0.1);

	glRotatef(-30, 0, 1, 0);

	float l = 0.5;
	glBegin(GL_LINES);
		glVertex3f(0, 0, 0);
		glVertex3f(-l, 640 * 0.5 * l / camParams.fxy, 480 * 0.5 * l / camParams.fxy);
		glVertex3f(0, 0, 0);
		glVertex3f(-l, -640 * 0.5 * l / camParams.fxy, 480 * 0.5 * l / camParams.fxy);
		glVertex3f(0, 0, 0);
		glVertex3f(-l, 640 * 0.5 * l / camParams.fxy, -480 * 0.5 * l / camParams.fxy);
		glVertex3f(0, 0, 0);
		glVertex3f(-l, -640 * 0.5 * l / camParams.fxy, -480 * 0.5 * l / camParams.fxy);
	glEnd();

	glPopMatrix();

	glutSwapBuffers();
}

void init() {
/*
	// lighting
	GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
	//	light_position is NOT default value
	GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
*/
	glEnable(GL_DEPTH_TEST);
//	glEnable(GL_CULL_FACE);

	// init
	glClearColor (0.0, 0.0, 0.0, 0.0);




	// Bildursprung in Pixeln
	camParams.ox = 640 * 0.5;
	camParams.oy = 480 * 0.5;

		// translation
//		camParams.t[0] = 0.20;
camParams.t[0] = 0.2;
//		camParams.t[1] = 0.0;
camParams.t[1] = 0;
//		camParams.t[2] = 0.17;
camParams.t[2] = 0.3;

	// rotation
	float a = 30 * PI / 180;
	float b = 0;//i*60 * PI / 180;
	camParams.r[0][0] = cos(b);
	camParams.r[0][1] = -sin(a) * sin(b);
	camParams.r[0][2] = cos(a) * sin(b);
	camParams.r[1][0] = 0;
	camParams.r[1][1] = cos(a);
	camParams.r[1][2] = sin(a);
	camParams.r[2][0] = -sin(b);
	camParams.r[2][1] = -sin(a) * cos(b);
	camParams.r[2][2] = cos(a) * cos(b);

	// distance origin - image plane [in pixel]
//		camParams.fxy = 480;
camParams.fxy = 480;

}

// Buchstaben; einige andere Tasten
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
/*		case '+':
			glutPostRedisplay();
			break;
		case '-':
			glutPostRedisplay();
			break;
*/
		case 27: // escape
		case 'q': exit(0);
   }
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	switch (actButton) {
		case GLUT_LEFT_BUTTON:
			q.rotate((x - oldMouseX) * 0.01, 0.0, -1.0, 0.0);
			q.rotate((y - oldMouseY) * 0.01, 1.0, 0.0, 0.0);
			break;
		case GLUT_MIDDLE_BUTTON:
			break;
	}

	oldMouseX = x;
	oldMouseY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y) {
	if (state == GLUT_DOWN) actButton = button;
	oldMouseX = x;
	oldMouseY = y;
}

void reshape(GLsizei width, GLsizei height) {
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0, // field of view in degree
		(GLfloat)(width)/(GLfloat)(height), // aspect ratio
		0.01, 50.0); // z near, z far
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

int main(int argc, char* argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(40, 40);
	glutCreateWindow("camParamVisu");
	glutKeyboardFunc(keyboardFunc);
//	glutSpecialFunc(keyboardSpecialFunc);
	glutReshapeFunc(reshape);
	glutMotionFunc(mouseDrag);
	glutMouseFunc(mouseButton);
//	glutIdleFunc(update);

	init();

	glutDisplayFunc(display);
	glutMainLoop();

	return 0;   /* ANSI C requires main to return int. */
}
