/*
 * frontend providing WinMain, window control and menu
 * for 3dscanner project
 *
 * partly copied from NeHe's Base Code
 *
 *		Marcel Lancelle 2002
 */

#include <windows.h>
#include "frontend.h"
#include "ScannerApp.h"
#include "QuaternionRotation.h"
#include "SplashWindow.h"
#include <math.h>
#include <commctrl.h>
#include "math3d.h"

char* HELP_FILE_NAME = "help\\index.htm";

ScannerApp scannerApp;
QuaternionRotation q;
int oldMouseX, oldMouseY;
bool mbDown = false; // mouse button pressed here (clicked in this window, not in another)
SplashWindow* splashWnd;
HWND hwndProgressBar;
// STATUSBAR
//HWND hStatusBar;
char* startupFileName = NULL;

int camPosZoom = 70;
float camPosZ;

void calcCamPosZ() {
	if (camPosZoom < 2) camPosZoom = 2;
	if (camPosZoom > 260) camPosZoom = 260;
	camPosZ = -pow(2.0, camPosZoom/20.0) * 0.02;
}

// recalculates merge Matrix
void calcMergeMatrix() {
	float m[4][4];
	scannerApp.mergeRotation.getRotMatrix(m);
	for (int j=0; j<4; j++) for (int i=0; i<4; i++)
		scannerApp.mergeMatrix.m[i][j] = m[j][i];

	for (int i=0; i<scannerApp.mergeProject->photos.size(); i++)
		scannerApp.mergeProject->photos[i].calcTrafo(&scannerApp.mergeProject->camDefs[0]);

/*
	MAT4x4 translMatr;
	resetMat(&translMatr);
	translate(&translMatr, &scannerApp.mergeTranslation);
	multMatMat(&scannerApp.mergeMatrix, &translMatr);
*/
	translate(&scannerApp.mergeMatrix, &scannerApp.mergeTranslation);
}

typedef struct {									// Contains Information Vital To Applications
	HINSTANCE		hInstance;						// Application Instance
	const char*		className;						// Application ClassName
} Application;										// Application

typedef struct {									// Window Creation Info
	Application*		application;				// Application Structure
	char*				title;						// Window Title
	int					width;						// Width
	int					height;						// Height
	int					bitsPerPixel;				// Bits Per Pixel
} GL_WindowInit;									// GL_WindowInit

typedef struct {									// Contains Information Vital To A Window
	HWND				hWnd;						// Window Handle
	HDC					hDC;						// Device Context
	HGLRC				hRC;						// Rendering Context
	GL_WindowInit		init;						// Window Init
	BOOL				isVisible;					// Window Visible?
} GL_Window;										// GL_Window

HMENU menu;			  /* Handle of the menu */

////////////////////////////////////////////////////////////////////////////////

bool getFileNameOpen(HWND hwnd, char* fileNameBuffer, int fnbuffLength, char* filter, char* title, char* defaultExtension) {
	OPENFILENAME ofn;
	ZeroMemory(&ofn, sizeof (OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = filter;
	ofn.nFilterIndex = 1;
	fileNameBuffer[0] = 0;
	ofn.lpstrFile = (LPTSTR) fileNameBuffer;
	ofn.nMaxFile = fnbuffLength;
	ofn.lpstrInitialDir = NULL; // current directory
	ofn.lpstrTitle = title;
	ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY; //OFN_PATHMUSTEXIST; //OFN_HIDEREADONLY; OFN_OVERWRITEPROMPT
	ofn.lpstrDefExt = defaultExtension;
	return GetOpenFileName(&ofn); // return if dialog was cancelled
}

bool getFileNameSave(HWND hwnd, char* fileNameBuffer, int fnbuffLength, char* filter, char* title, char* defaultExtension) {
	OPENFILENAME ofn;
	ZeroMemory(&ofn, sizeof (OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = filter;
	ofn.nFilterIndex = 1;
	fileNameBuffer[0] = 0;
	ofn.lpstrFile = (LPTSTR) fileNameBuffer;
	ofn.nMaxFile = fnbuffLength;
	ofn.lpstrInitialDir = NULL; // current directory
	ofn.lpstrTitle = title;
	ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY; //OFN_PATHMUSTEXIST; //OFN_HIDEREADONLY; OFN_FILEMUSTEXIST
	ofn.lpstrDefExt = defaultExtension;
	if (strlen(defaultExtension) > 3) ofn.lpstrDefExt = NULL; // only first 3 characters are appended -> do it later manually
	if (!GetOpenFileName(&ofn)) return false; // return if dialog was cancelled

	// append default extension if no extension was chosen
	if ((defaultExtension) && (ofn.nFileExtension == 0)) {
		strcat(fileNameBuffer, ".");
		strcat(fileNameBuffer, defaultExtension);
	}

	return true;
}

// splits e.g. "c:\\dir\\abc.exe" to "c:\\dir\\", "abc", "exe" - don't forget to delete[] last 3 char* !
void splitPathNameExt(char* fullName, char** path, char** name, char** ext) {
	int nPos = strlen(fullName);
	while ((nPos) && (fullName[nPos - 1] != '\\')) nPos--;
	int ePos = strlen(fullName);
	while ((ePos) && (fullName[ePos - 1] != '.')) ePos--;
	if (ePos <= nPos) ePos = strlen(fullName)+1;

	*path = new char[nPos + 1];
	strncpy(*path, fullName, nPos); (*path)[nPos] = '\0';

	*name = new char[ePos - nPos];
	strncpy(*name, (char*) &(fullName[nPos]), ePos - nPos - 1); (*name)[ePos - nPos - 1] = '\0';

	if (ePos > strlen(fullName)) ePos = strlen(fullName);
	*ext = new char[strlen(fullName) - ePos + 1];
	strncpy(*ext, (char*) &(fullName[ePos]), strlen(fullName) - ePos); (*ext)[strlen(fullName) - ePos] = '\0';
}

void setupProgressBar(HWND hwnd, HINSTANCE hInstance, int rangeStart, int rangeEnd) {
	// set up progress controll
	RECT rcClient;  // client area of parent window
	int cyVScroll;  // height of scroll bar arrow

	// Ensure that the common control DLL is loaded and create a
	// progress bar along the bottom of the client area of the
	// parent window. Base the height of the progress bar on
	// the height of a scroll bar arrow.
	InitCommonControls();
	GetClientRect(hwnd, &rcClient);
	cyVScroll = GetSystemMetrics(SM_CYVSCROLL);
	hwndProgressBar = CreateWindowEx(0, PROGRESS_CLASS, (LPSTR) NULL,
		WS_CHILD | WS_VISIBLE, rcClient.left,
		rcClient.bottom - cyVScroll,
		rcClient.right, cyVScroll,
		hwnd, (HMENU) 0, hInstance, NULL);
	// Set the range and increment of the progress bar.
	SendMessage(hwndProgressBar, PBM_SETRANGE, 0, MAKELPARAM(rangeStart, rangeEnd));
	SendMessage(hwndProgressBar, PBM_SETSTEP, (WPARAM) 1, 0);
}

void incrementProgressBar() {
	// advance the current position of the progress bar by the increment
	SendMessage(hwndProgressBar, PBM_STEPIT, 0, 0);
}

void closeProgressBar() {
	// close progress controll
	DestroyWindow(hwndProgressBar);
}

////////////////////////////////////////////////////////////////////////////////

void TerminateApplication(GL_Window* window);		// Terminate The Application
LRESULT CALLBACK WindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

BOOL Initialize(GL_Window* window) {	// Performs All Your Initialization
	q.rotate(-0.7, 1.0, 0.0, 0.0);

	calcCamPosZ();

	scannerApp.initGl();
	return true;
}

void Deinitialize (void) {							// Performs All Your DeInitialization
}

BOOL g_isProgramLooping;

void Draw(GL_Window* window) {									// Perform All Your Scene Drawing
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear Screen And Depth Buffer

	glLoadIdentity();
//	glTranslatef(0, 0, camPosZ);
	glTranslatef(0, 0, -1);
	glScalef(-1/camPosZ, -1/camPosZ, -1/camPosZ);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

	scannerApp.draw();

	glFlush();													// Flush The GL Rendering Pipeline
	SwapBuffers(window->hDC);					// Swap Buffers (Double Buffering)
}

GL_Window			window;												// Window Structure

void TerminateApplication (GL_Window* window) {							// Terminate The Application
	PostMessage (window->hWnd, WM_QUIT, 0, 0);							// Send A WM_QUIT Message
	g_isProgramLooping = FALSE;											// Stop Looping Of The Program
}

void ReshapeGL(int width, int height) {									// Reshape The Window When It's Moved Or Resized
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));				// Reset The Current Viewport
	glMatrixMode(GL_PROJECTION);										// Select The Projection Matrix
	glLoadIdentity();													// Reset The Projection Matrix
	gluPerspective(60.0f, (GLfloat)(width)/(GLfloat)(height),			// Calculate The Aspect Ratio Of The Window
					0.1f, 20.0f);
	glMatrixMode(GL_MODELVIEW);											// Select The Modelview Matrix
	glLoadIdentity();													// Reset The Modelview Matrix
}

BOOL CreateWindowGL (GL_Window* window) {								// This Code Creates Our OpenGL Window
	DWORD windowStyle = WS_OVERLAPPEDWINDOW;							// Define Our Window Style
	DWORD windowExtendedStyle = WS_EX_ACCEPTFILES;	// Define The Window's Extended Style

	PIXELFORMATDESCRIPTOR pfd = {										// pfd Tells Windows How We Want Things To Be
		sizeof(PIXELFORMATDESCRIPTOR),									// Size Of This Pixel Format Descriptor
		1,																// Version Number
		PFD_DRAW_TO_WINDOW |											// Format Must Support Window
		PFD_SUPPORT_OPENGL |											// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,												// Must Support Double Buffering
		PFD_TYPE_RGBA,													// Request An RGBA Format
		window->init.bitsPerPixel,										// Select Our Color Depth
		0, 0, 0, 0, 0, 0,												// Color Bits Ignored
		0,																// No Alpha Buffer
		0,																// Shift Bit Ignored
		0,																// No Accumulation Buffer
		0, 0, 0, 0,														// Accumulation Bits Ignored
		16,																// 16Bit Z-Buffer (Depth Buffer)
		0,																// No Stencil Buffer
		0,																// No Auxiliary Buffer
		PFD_MAIN_PLANE,													// Main Drawing Layer
		0,																// Reserved
		0, 0, 0															// Layer Masks Ignored
	};

	RECT windowRect = {0, 0, window->init.width, window->init.height};	// Define Our Window Coordinates

	GLuint PixelFormat;													// Will Hold The Selected Pixel Format

	// Adjust Window, Account For Window Borders
	AdjustWindowRectEx(&windowRect, windowStyle, TRUE, windowExtendedStyle);
	int sysWidth = GetSystemMetrics(SM_CXSCREEN);
	int sysHeight = GetSystemMetrics(SM_CYSCREEN);

	// Create The OpenGL Window
	window->hWnd = CreateWindowEx (windowExtendedStyle,					// Extended Style
								   window->init.application->className,	// Class Name
								   window->init.title,					// Window Title
								   windowStyle,							// Window Style
								   (sysWidth-window->init.width)/2, // Window X Position
								   (sysHeight-window->init.height)/2, // Window Y Position
								   window->init.width,	// Window Width
								   window->init.height,	// Window Height
								   HWND_DESKTOP,						// Desktop Is Window's Parent
								   0,									// No Menu
								   window->init.application->hInstance, // Pass The Window Instance
								   window);

	if (window->hWnd == 0)												// Was Window Creation A Success?
	{
		return FALSE;													// If Not Return False
	}

	window->hDC = GetDC(window->hWnd);									// Grab A Device Context For This Window
	if (window->hDC == 0)												// Did We Get A Device Context?
	{
		// Failed
		DestroyWindow(window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
		return FALSE;													// Return False
	}

	PixelFormat = ChoosePixelFormat (window->hDC, &pfd);				// Find A Compatible Pixel Format
	if (PixelFormat == 0)												// Did We Find A Compatible Format?
	{
		// Failed
		ReleaseDC(window->hWnd, window->hDC);							// Release Our Device Context
		window->hDC = 0;												// Zero The Device Context
		DestroyWindow(window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
		return FALSE;													// Return False
	}

	if (SetPixelFormat(window->hDC, PixelFormat, &pfd) == FALSE)		// Try To Set The Pixel Format
	{
		// Failed
		ReleaseDC(window->hWnd, window->hDC);							// Release Our Device Context
		window->hDC = 0;												// Zero The Device Context
		DestroyWindow(window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
		return FALSE;													// Return False
	}

	window->hRC = wglCreateContext(window->hDC);						// Try To Get A Rendering Context
	if (window->hRC == 0)												// Did We Get A Rendering Context?
	{
		// Failed
		ReleaseDC(window->hWnd, window->hDC);							// Release Our Device Context
		window->hDC = 0;												// Zero The Device Context
		DestroyWindow(window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
		return FALSE;													// Return False
	}

	// Make The Rendering Context Our Current Rendering Context
	if (wglMakeCurrent (window->hDC, window->hRC) == FALSE)
	{
		// Failed
		wglDeleteContext (window->hRC);									// Delete The Rendering Context
		window->hRC = 0;												// Zero The Rendering Context
		ReleaseDC (window->hWnd, window->hDC);							// Release Our Device Context
		window->hDC = 0;												// Zero The Device Context
		DestroyWindow (window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
		return FALSE;													// Return False
	}

	ShowWindow(window->hWnd, SW_NORMAL);								// Make The Window Visible
	window->isVisible = TRUE;											// Set isVisible To True

	ReshapeGL (window->init.width, window->init.height);				// Reshape Our GL Window

	return TRUE;														// Window Creating Was A Success
																		// Initialization Will Be Done In WM_CREATE
}

BOOL DestroyWindowGL (GL_Window* window)								// Destroy The OpenGL Window & Release Resources
{
	if (window->hWnd != 0)												// Does The Window Have A Handle?
	{
		if (window->hDC != 0)											// Does The Window Have A Device Context?
		{
			wglMakeCurrent (window->hDC, 0);							// Set The Current Active Rendering Context To Zero
			if (window->hRC != 0)										// Does The Window Have A Rendering Context?
			{
				wglDeleteContext (window->hRC);							// Release The Rendering Context
				window->hRC = 0;										// Zero The Rendering Context

			}
			ReleaseDC (window->hWnd, window->hDC);						// Release The Device Context
			window->hDC = 0;											// Zero The Device Context
		}
		DestroyWindow (window->hWnd);									// Destroy The Window
		window->hWnd = 0;												// Zero The Window Handle
	}

	return TRUE;														// Return True
}

BOOL RegisterWindowClass(Application* application) {					// Register A Window Class For This Application.
																		// TRUE If Successful
	// Register A Window Class
	WNDCLASSEX windowClass;												// Window Class
	ZeroMemory (&windowClass, sizeof (WNDCLASSEX));						// Make Sure Memory Is Cleared
	windowClass.cbSize			= sizeof (WNDCLASSEX);					// Size Of The windowClass Structure
	windowClass.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraws The Window For Any Movement / Resizing
	windowClass.lpfnWndProc		= (WNDPROC)(WindowProcedure);				// WindowProc Handles Messages
	windowClass.hInstance		= application->hInstance;				// Set The Instance
//	windowClass.hbrBackground	= (HBRUSH)(COLOR_APPWORKSPACE);			// Class Background Brush Color
	windowClass.hbrBackground	= (HBRUSH) GetStockObject(NULL_BRUSH);
	windowClass.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	windowClass.hIcon 			= LoadIcon(application->hInstance, MAKEINTRESOURCE(IDI_3DSCANNER)); // LoadIcon(NULL, IDI_APPLICATION);
	windowClass.hIconSm 		= LoadIcon(application->hInstance, MAKEINTRESOURCE(IDI_3DSCANNER));
	windowClass.lpszClassName	= application->className;				// Sets The Applications Classname
	if (RegisterClassEx(&windowClass) == 0) {							// Did Registering The Class Fail?
		// NOTE: Failure, Should Never Happen
		MessageBox (HWND_DESKTOP, "RegisterClassEx Failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;													// Return False (Failure)
	}
	return TRUE;														// Return True (Success)
}

void updateMenu(HWND hwnd, bool draw = true) {
	// main menu items
	EnableMenuItem(menu, ENUM_MENU_SCAN, (scannerApp.project) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);

	// file submenu items
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_FILE), ENUM_MENU_FILESAVE, scannerApp.project ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_FILE), ENUM_MENU_FILE_MERGE_WITH_PROJECT, scannerApp.project ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_FILE), ENUM_MENU_FILE_PROJECT_INFO, scannerApp.project ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_FILE), ENUM_MENU_FILE_EXPORT_MODEL_VRML, (scannerApp.project && scannerApp.project->mcPtr) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_FILE), ENUM_MENU_FILE_EXPORT_MODEL_OBJ, (scannerApp.project && scannerApp.project->mcPtr) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);

	// scan submenu items
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_SCAN), ENUM_MENU_SCAN_SEGMENTIMG, (scannerApp.project && scannerApp.project->photos.size()) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_SCAN), ENUM_MENU_SCAN_PREPMANSEG, (scannerApp.project && scannerApp.project->photos.size()) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_SCAN), ENUM_MENU_SCAN_GENERATE3D, (scannerApp.project && scannerApp.project->photos.size()) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_SCAN), ENUM_MENU_SCAN_FITVOLUME, (scannerApp.project && scannerApp.project->mcPtr) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableMenuItem(GetSubMenu(menu, ENUM_MENU_SCAN), ENUM_MENU_SCAN_TEXTURE, (scannerApp.project && scannerApp.project->mcPtr) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);

	if (draw) DrawMenuBar(hwnd);
}

/* This function is called by the Windowsfunction DispatchMessage( ) */
LRESULT CALLBACK WindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	// Get The Window Context
	GL_Window* window = (GL_Window*)(GetWindowLong (hwnd, GWL_USERDATA));

	switch (message) {
		case WM_CREATE: {
// STATUSBAR
/*			// status bar
			InitCommonControls();
			hStatusBar = CreateWindowEx(0, STATUSCLASSNAME, NULL,
				WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP, 0, 0, 0, 0,
				hwnd, (HMENU)IDC_MAIN_STATUS, GetModuleHandle(NULL), NULL);
SetWindowText(hStatusBar, ":-)");
*/
			CREATESTRUCT* creation = (CREATESTRUCT*)(lParam);			// Store Window Structure Pointer
			window = (GL_Window*)(creation->lpCreateParams);
			SetWindowLong(hwnd, GWL_USERDATA, (LONG)(window));

   			scannerApp.init(hwnd, window->init.application->hInstance);

//			Sleep(500); // delay creation so that splash screen is shown first
//			ShowWindow(hwnd, SW_RESTORE);
//			ShowWindow(hwnd, SW_NORMAL);
   			return 0; }

		case WM_PAINT:
//			Draw(window);
			break;

		case WM_ERASEBKGND:
			Draw(window);
			return 1;

		case WM_USER_REDRAW:
			Draw(window);
			break;

		case WM_USER_OPENFILE:
			if (!startupFileName) break; // should never happen
			scannerApp.fileOpen(startupFileName);
			if (scannerApp.project) SetCurrentDirectory(scannerApp.project->projectPath);
			updateMenu(hwnd);
			delete[] startupFileName;
			return 0;
			break;

		case WM_DROPFILES: {
			HDROP hDrop = (HDROP) wParam;
			int buffSize = DragQueryFile(hDrop, 0, NULL, 0) + 1;
			char* fileName = new char[buffSize];
			DragQueryFile(hDrop, 0, fileName, buffSize);
			scannerApp.fileOpen(fileName);
			if (scannerApp.project) SetCurrentDirectory(scannerApp.project->projectPath);
			updateMenu(hwnd);
			delete[] fileName;
			break;}

		case WM_COMMAND:
		  switch (wParam) {
			//////////////  File  //////////////
			case IDM_FILENEW:
				scannerApp.fileNew();
				updateMenu(hwnd);
				return 0;
			case IDM_FILEOPEN:
				scannerApp.fileOpen();
				updateMenu(hwnd);
				return 0;
			case IDM_FILESAVE:
				scannerApp.fileSave();
				return 0;

			case IDM_FILE_MERGE_WITH_PROJECT:
				scannerApp.fileMergeWithProjectPrepare();
				return 0;

			case IDM_FILE_PROJECT_INFO:
				scannerApp.fileShowProjectInfo();
				return 0;

			case IDM_FILE_EXPORT_MODEL_VRML:
				scannerApp.fileExportModel2VRML();
				return 0;

			case IDM_FILE_EXPORT_MODEL_OBJ:
				scannerApp.fileExportModel2Obj();
				return 0;

			case IDM_FILEEXIT:
				SendMessage(hwnd, WM_CLOSE, 0, 0L);
				return 0;

			//////////////  Scan  //////////////
			case IDM_SCAN_CALCAM:
				scannerApp.scanCalibrateCam();
				return 0;
			case IDM_SCAN_CALCAM1:
				scannerApp.scanCalibrateCam1();
				return 0;
			case IDM_SCAN_GETIMG:
//				PostMessage(hwnd, WM_USER_GET_IMG, GET_IMG_START, 0);
				scannerApp.scanGetImages();
				return 0;
			case IDM_SCAN_SEGMENTIMG:
				scannerApp.scanSegmentImages();
				return 0;
			case IDM_SCAN_PREPMANSEG:
				scannerApp.scanPrepareManualSegmentation();
				return 0;
			case IDM_SCAN_GENERATE3D:
				scannerApp.scanGenerate3DModel(scannerApp.project);
				updateMenu(hwnd);
				return 0;
			case IDM_SCAN_FITVOLUME:
				scannerApp.scanFitScanningVolume2LastScan(scannerApp.project);
				return 0;
			case IDM_SCAN_SIMPLIFY:
				scannerApp.scanSimplifyModel(scannerApp.project);
				return 0;
			case IDM_SCAN_TEXTURE:
				scannerApp.scanTexture(scannerApp.project);
				return 0;

			//////////////  View  //////////////
			case IDM_VIEW_GRID:
				scannerApp.showGrid = !scannerApp.showGrid;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_GRID, scannerApp.showGrid ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;
			case IDM_VIEW_TEXTURE:
				scannerApp.showTexture = !scannerApp.showTexture;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_TEXTURE, scannerApp.showTexture ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;
			case IDM_VIEW_SMOOTHSHADED:
				scannerApp.showSmoothShaded = !scannerApp.showSmoothShaded;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_SMOOTHSHADED, scannerApp.showSmoothShaded ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;
			case IDM_VIEW_MESH:
				scannerApp.showMesh = !scannerApp.showMesh;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_MESH, scannerApp.showMesh ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;
			case IDM_VIEW_NORMALS:
				scannerApp.showNormals = !scannerApp.showNormals;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_NORMALS, scannerApp.showNormals ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;
			case IDM_VIEW_VIEWS:
				scannerApp.showViews = !scannerApp.showViews;
				CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_VIEWS, scannerApp.showViews ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
				DrawMenuBar(hwnd);
				InvalidateRect(hwnd, NULL, FALSE);
				return 0;

			//////////////  Config  //////////////
			case IDM_CONFIGSMDRVINST:
				scannerApp.configPortTalkInstall();
				return 0;
			case IDM_CONFIGSMDRVUNINST:
				scannerApp.configPortTalkUninstall();
				return 0;
			case IDM_CONFIGSMRESET:
				scannerApp.configSMReset();
				return 0;
			case IDM_CONFIGSMSAVE1ROT:
				scannerApp.configSMSave1Rot();
				return 0;
			case IDM_CONFIGSMSETPORT278H:
				scannerApp.configSMSetPort(0x0278);
				return 0;
			case IDM_CONFIGSMSETPORT378H:
				scannerApp.configSMSetPort(0x0378);
				return 0;
			case IDM_CONFIGCAMMETHODTWAIN:
				scannerApp.configSetIAT(CamAccess::IAT_TWAIN);
				CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_TWAIN, MF_BYPOSITION);
				return 0;
			case IDM_CONFIGCAMMETHODVFW:
				scannerApp.configSetIAT(CamAccess::IAT_VFW);
				CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_VFW, MF_BYPOSITION);
				return 0;
			case IDM_CONFIGCAMMETHODFILE:
				scannerApp.configSetIAT(CamAccess::IAT_FILE);
				CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_FILE, MF_BYPOSITION);
				return 0;
			case IDM_CONFIGCAMSELECTSOURCE:
				scannerApp.configCamSelectSource();
				return 0;
			case IDM_CONFIGCAMOPTIONS:
				scannerApp.configCamOptions();
				return 0;
			case IDM_CONFIGCAMSHOWPREVIEWWIN:
				scannerApp.configCamShowPreviewWindow();
				return 0;
			case IDM_CONFIGSELECTCALPAT:
				scannerApp.configSelectCalibrationPattern();
				return 0;

			//////////////  Help  //////////////
			case IDM_HELPCONTENTS: {
				//WinHelp( hwnd, (LPSTR) "HELPFILE.HLP", HELP_CONTENTS, 0L);
				if (MessageBox(NULL, "Show html help file?", "3DScanner", MB_ICONQUESTION | MB_YESNOCANCEL) != IDYES) return 0;

				// generate help file name
				char* hfn = new char[strlen(scannerApp.exePath) + strlen(HELP_FILE_NAME) + 1];
				strcpy(hfn, scannerApp.exePath);
				strcat(hfn, HELP_FILE_NAME);

				ShellExecute(NULL, "open", hfn, NULL, "", 0);
				delete[] hfn;
				return 0;}
			case IDM_HELPABOUT:
				//MessageBox(NULL, "3DScanner by Marcel Lancelle, 2002\nTU Braunschweig, Germany", "About...", MB_ICONINFORMATION | MB_OK);
				splashWnd->show(10000);
				return 0;
		  } // WM_COMMAND: switch (wParam)
		break; // WM_COMMAND

/*
	  case WM_USER_GET_IMG:
//MessageBox (NULL, "WM_USER + 0" , "3DScanner", MB_OK);
		   scannerApp.scanGetImages(wParam);
*/
	  case WM_KEYDOWN:
			switch ((int)wParam) {
				//case VK_ESCAPE:
				case VK_RIGHT:
					if (!scannerApp.rotateManually) scannerApp.scanRotateTurnTable(1);
					return 0;
				case VK_LEFT:
					if (!scannerApp.rotateManually) scannerApp.scanRotateTurnTable(-1);
					return 0;
				case VK_F1:
					PostMessage(hwnd, WM_COMMAND, IDM_HELPCONTENTS, 0);
					return 0;
				case VK_RETURN:
					if (scannerApp.appState == scannerApp.APP_STATE_MERGE) {
						scannerApp.fileMergeWithProjectApply(false);
						//InvalidateRect(hwnd, NULL, FALSE);
						Draw(window);
					}
					return 0;
				case VK_ESCAPE:
					if (scannerApp.appState == scannerApp.APP_STATE_MERGE) {
						scannerApp.fileMergeWithProjectApply(true);
						//InvalidateRect(hwnd, NULL, FALSE);
						Draw(window);
					}
					return 0;
			}
			break;
	  case WM_SYSKEYDOWN: // ALT pressed
			switch ((int)wParam) {
				case VK_RIGHT:
					if (!scannerApp.rotateManually) scannerApp.scanRotateTurnTable(50);
					return 0;
				case VK_LEFT:
					if (!scannerApp.rotateManually) scannerApp.scanRotateTurnTable(-50);
					return 0;
			}
			break;

		case WM_SIZE:													// Size Action Has Taken Place
			switch (wParam)												// Evaluate Size Action
			{
				case SIZE_MINIMIZED:									// Was Window Minimized?
					window->isVisible = FALSE;							// Set isVisible To False
				return 0;												// Return

				case SIZE_MAXIMIZED:									// Was Window Maximized?
					window->isVisible = TRUE;							// Set isVisible To True
					ReshapeGL(LOWORD (lParam), HIWORD (lParam));		// Reshape Window - LoWord=Width, HiWord=Height
					InvalidateRect(hwnd, NULL, FALSE);
				return 0;												// Return

				case SIZE_RESTORED:										// Was Window Restored?
					window->isVisible = TRUE;							// Set isVisible To True
/* STATUSBAR
			HWND hStatus = GetDlgItem(hwnd, IDC_MAIN_STATUS);
			SendMessage(hStatus, WM_SIZE, 0, 0);
//SendMessage(hStatusBar, WM_SIZE, 0, 0);
*/
					ReshapeGL(LOWORD (lParam), HIWORD (lParam));		// Reshape Window - LoWord=Width, HiWord=Height
					InvalidateRect(hwnd, NULL, FALSE);
				return 0;												// Return
			}
		break;

	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		SetCapture(hwnd); // capture mouse
		mbDown = true;
		return 0;

	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
		ReleaseCapture(); // release mouse
		mbDown = false;
		return 0;

	case WM_MOUSEMOVE: {
		int fwKeys = wParam; // key flags
		int xPos = (SHORT) LOWORD(lParam); // horizontal position of cursor
		int yPos = (SHORT) HIWORD(lParam); // vertical position of cursor

		if (mbDown) switch (fwKeys) {
			case MK_LBUTTON:
				q.rotate((xPos - oldMouseX) * 0.01, 0.0, 1.0, 0.0);
				q.rotate((yPos - oldMouseY) * 0.01, 1.0, 0.0, 0.0);
				Draw(window);
				break;
			case MK_CONTROL | MK_LBUTTON:
				camPosZoom += yPos - oldMouseY;
				calcCamPosZ();
				Draw(window);
				break;
			case MK_RBUTTON:
				if (scannerApp.appState == scannerApp.APP_STATE_MERGE) {
					// rotate merge object
					float m[4][4];
					q.getRotMatrix(m);
					scannerApp.mergeRotation.rotate((xPos - oldMouseX) * 0.01, m[0][1], m[1][1], m[2][1]);
					scannerApp.mergeRotation.rotate((yPos - oldMouseY) * 0.01, m[0][0], m[1][0], m[2][0]);
					calcMergeMatrix();
					Draw(window);
				}
				break;
			case MK_CONTROL | MK_RBUTTON:
				if (scannerApp.appState == scannerApp.APP_STATE_MERGE) {
					float m[4][4];
					q.getRotMatrix(m);
					scannerApp.mergeTranslation.v[0] += (m[0][0]*(xPos - oldMouseX) - m[0][1]*(yPos - oldMouseY)) * 0.002;
					scannerApp.mergeTranslation.v[1] += (m[1][0]*(xPos - oldMouseX) - m[1][1]*(yPos - oldMouseY)) * 0.002;
					scannerApp.mergeTranslation.v[2] += (m[2][0]*(xPos - oldMouseX) - m[2][1]*(yPos - oldMouseY)) * 0.002;
					calcMergeMatrix();
					Draw(window);
				}
				break;
		}

		oldMouseX = xPos;
		oldMouseY = yPos;
	break;}

	  case WM_CLOSE:
		   scannerApp.deinit();
			TerminateApplication(window);								// Terminate The Application
//		   DestroyWindow( hwnd );
		   return 0;
/*
	  case WM_DESTROY:
		   PostQuitMessage(0);
		   return 0;
*/

	  break;
//	  default:	//for messages that we don't deal with
//	  return DefWindowProc(hwnd, message, wParam, lParam);
   }
	  return DefWindowProc(hwnd, message, wParam, lParam);
//  return 0;
}

// Program Entry (WinMain)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	splashWnd = new SplashWindow(hInstance, IDB_SPLASH);
	splashWnd->show(3000);

	Application			application;									// Application Structure
	BOOL				isMessagePumpActive;							// Message Pump Active?
	MSG					msg;											// Window Message Structure

	// Fill Out Application Data
	application.className = "3dscCl";									// Application Class Name
	application.hInstance = hInstance;									// Application Instance

	// Fill Out Window
	ZeroMemory (&window, sizeof (GL_Window));							// Make Sure Memory Is Zeroed
	window.init.application		= &application;							// Window Application
	window.init.title			= "3DScanner";							// Window Title
	window.init.width			= 640;									// Window Width
	window.init.height			= 480;									// Window Height
	window.init.bitsPerPixel	= 16;									// Bits Per Pixel

	// Register A Class For Our Window To Use
	if (RegisterWindowClass (&application) == FALSE) {					// Did Registering A Class Fail?
		// Failure
		MessageBox(HWND_DESKTOP, "Error Registering Window Class!", "Error", MB_OK | MB_ICONEXCLAMATION);
		return -1;														// Terminate Application
	}

	g_isProgramLooping = TRUE;											// Program Looping Is Set To TRUE
	while (g_isProgramLooping)											// Loop Until WM_QUIT Is Received
	{
		// Create A Window
		if (CreateWindowGL (&window) == TRUE)							// Was Window Creation Successful?
		{
			menu = LoadMenu(application.hInstance, MAKEINTRESOURCE(ID_MENU));
			SetMenu(window.hWnd, menu);

			// check view menu items
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_GRID, scannerApp.showGrid ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_TEXTURE, scannerApp.showTexture ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_SMOOTHSHADED, scannerApp.showSmoothShaded ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_MESH, scannerApp.showMesh ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_NORMALS, scannerApp.showNormals ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );
			CheckMenuItem(GetSubMenu(menu, ENUM_MENU_VIEW), ENUM_MENU_VIEW_VIEWS, scannerApp.showViews ? MF_BYPOSITION|MF_CHECKED : MF_BYPOSITION|MF_UNCHECKED );

			// check config cam menu items
			if (scannerApp.imageAquisitionType == CamAccess::IAT_TWAIN) CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_TWAIN, MF_BYPOSITION);
			if (scannerApp.imageAquisitionType == CamAccess::IAT_VFW) CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_VFW, MF_BYPOSITION);
			if (scannerApp.imageAquisitionType == CamAccess::IAT_FILE) CheckMenuRadioItem(GetSubMenu(GetSubMenu(GetSubMenu(menu, ENUM_MENU_CONFIG), ENUM_MENU_CONFIG_CAM), ENUM_MENU_CONFIG_CAM_METHOD), ENUM_MENU_CONFIG_CAM_METHOD_FIRST, ENUM_MENU_CONFIG_CAM_METHOD_LAST, ENUM_MENU_CONFIG_CAM_METHOD_FILE, MF_BYPOSITION);

			updateMenu(window.hWnd, false);

			// At This Point We Should Have A Window That Is Setup To Render OpenGL
			if (Initialize (&window) == FALSE)					// Call User Intialization
			{
				// Failure
				TerminateApplication (&window);							// Close Window, This Will Handle The Shutdown
			}
			else														// Otherwise (Start The Message Pump)
			{	// Initialize was a success

/*
	// Run the message loop. It will run until GetMessage() returns 0
	while (GetMessage (&msg, NULL, 0, 0))
	{
		if (!scannerApp.preTranslateMessage(&msg)) { // for twain
			// Translate virtual-key messages into character messages
			TranslateMessage(&msg);
		}

		// Send message to WindowProcedure
		DispatchMessage(&msg);
	}
*/

//Sleep(3000); // delay creation so that splash screen is shown first
//ShowWindow(window.hWnd, SW_NORMAL);

				// open file on start up
				if (strlen(lpCmdLine) > 0) {
					startupFileName = new char[strlen(lpCmdLine)+1];
					strcpy(startupFileName, lpCmdLine);
					// remove " characters at start and end of file name
					if ((startupFileName[0] == '\"') && (startupFileName[strlen(startupFileName)-1] == '\"')) {
						for (int i=0; i<strlen(startupFileName)-2; i++) startupFileName[i] = startupFileName[i+1];
						startupFileName[strlen(startupFileName)-2] = '\0';
					}
					SendMessage(window.hWnd, WM_USER_OPENFILE, 0, 0);
				}

				isMessagePumpActive = TRUE;								// Set isMessagePumpActive To TRUE
				while (isMessagePumpActive == TRUE)						// While The Message Pump Is Active
				{
					// Success Creating Window.  Check For Window Messages
					if (PeekMessage (&msg, NULL/*window.hWnd*/, 0, 0, PM_REMOVE) != 0)
					{
						// Check For WM_QUIT Message
						if (msg.message != WM_QUIT)						// Is The Message A WM_QUIT Message?
						{
							if (!scannerApp.preTranslateMessage(&msg)) TranslateMessage(&msg);
							DispatchMessage (&msg);						// If Not, Dispatch The Message
						}
						else											// Otherwise (If Message Is WM_QUIT)
						{
							isMessagePumpActive = FALSE;				// Terminate The Message Pump
						}
					}
					WaitMessage ();								// Application Is Minimized Wait For A Message
				}														// Loop While isMessagePumpActive == TRUE

			}															// If (Initialize (...

			// Application Is Finished
			Deinitialize ();											// User Defined DeInitialization

			DestroyWindowGL (&window);									// Destroy The Active Window
		}
		else															// If Window Creation Failed
		{
			// Error Creating Window
			MessageBox(HWND_DESKTOP, "Error Creating OpenGL Window", "Error", MB_OK | MB_ICONEXCLAMATION);
			g_isProgramLooping = FALSE;									// Terminate The Loop
		}
	}																	// While (isProgramLooping)

	UnregisterClass(application.className, application.hInstance);		// UnRegister Window Class
	if (splashWnd) delete splashWnd;
	return 0;
}																		// End Of WinMain()
