#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "QuaternionRotation.h"
#include "MBmpImg.h"
#include "mathext.h"

MBmpImg* bmpImg;

QuaternionRotation q;
int oldMouseX, oldMouseY;

//#define SIZE_HSL 16
// SHIFT_HSL = ld(256/SIZE_HSL)
//#define SHIFT_HSL 4

#define SIZE_HSL 32
// SHIFT_HSL = ld(256/SIZE_HSL)
#define SHIFT_HSL 3

unsigned int hist[SIZE_HSL][SIZE_HSL][SIZE_HSL]; // [r][g][b]
unsigned int max = 0;
float scaleCube;
int scCuInt = 20;

int actButton;

#define SCCU_MIN -200
#define SCCU_MAX 400

void calcScaleCube() {
	if (scCuInt < SCCU_MIN) scCuInt = SCCU_MIN;
	if (scCuInt > SCCU_MAX) scCuInt = SCCU_MAX;
	scaleCube = 2.0 * pow(2.0, scCuInt/20.0) / max;
}

// x, y, z: -1..1
void getHSLCoords(int h, int s, int l, float* x, float* y, float* z) {
	float norm = 1 / 255.0;
	float dist = (128-abs(l-127)) * 2 * norm * s * norm;
	*x = dist * cos(h*norm*2*PI);
	*y = dist * sin(h*norm*2*PI);
	*z = 2*norm*l - 1;
}

void getHSLModifiedCoords(int h, int s, int l, float* x, float* y, float* z) {
	float norm = 1 / 255.0;
	float dist = sqrt((128-abs(l-127)) * 2 * norm) * s * norm;
	*x = dist * cos(h*norm*2*PI);
	*y = dist * sin(h*norm*2*PI);
	*z = 0.5*(2*norm*l - 1);
}

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0, 0, -3);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

	float st = 1 / (float) SIZE_HSL;

	glColor3f(0.6, 0.6, 0.6);
	glutWireCube(2.0);

	float tx, ty, tz;
	// draw red, green, blue, black, white boxes
	glPushMatrix();
		getHSLModifiedCoords(0, 0, 0, &tx, &ty, &tz);
		glTranslatef(tx, ty, tz);
		glColor3f(0.0, 0.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		getHSLModifiedCoords(0, 255, 127, &tx, &ty, &tz);
		glTranslatef(tx, ty, tz);
		glColor3f(1.0, 0.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		getHSLModifiedCoords(85, 255, 127, &tx, &ty, &tz);
		glTranslatef(tx, ty, tz);
		glColor3f(0.0, 1.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		getHSLModifiedCoords(170, 255, 127, &tx, &ty, &tz);
		glTranslatef(tx, ty, tz);
		glColor3f(0.0, 0.0, 1.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		getHSLModifiedCoords(0, 0, 255, &tx, &ty, &tz);
		glTranslatef(tx, ty, tz);
		glColor3f(1.0, 1.0, 1.0);
		glutSolidCube(0.05);
	glPopMatrix();

	for (int h=0; h<SIZE_HSL; h++)
	 for (int s=0; s<SIZE_HSL; s++)
	  for (int l=0; l<SIZE_HSL; l++) {
		if (hist[h][s][l]) {
			glPushMatrix();

			int h255 = h << SHIFT_HSL;
			int s255 = s << SHIFT_HSL;
			int l255 = l << SHIFT_HSL;
			// translate and scale
			getHSLModifiedCoords(h255, s255, l255, &tx, &ty, &tz);
			glTranslatef(tx, ty, tz);

			float sc = scaleCube * hist[h][s][l];
			if (sc < 0.05) sc = 0.05;
			if (sc > 0.8) sc = 0.8;
			sc *= st;
			glScalef(sc, sc, sc);

			int r, g, b;
			MBmpImg::getRGB(h255, s255, l255, &r, &g, &b);

			// draw colored cube
			glBegin(GL_QUADS);
				// oben
				glColor3ub(r, g, b);
				glVertex3f(+1.0, +1.0, +1.0);
				glVertex3f(-1.0, +1.0, +1.0);
				glVertex3f(-1.0, -1.0, +1.0);
				glVertex3f(+1.0, -1.0, +1.0);
				// unten
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				// links
				glColor3ub(CLIP(r-25, 0, 255), CLIP(g-25, 0, 255), CLIP(b-25, 0, 255));
				glVertex3f(-1.0, +1.0, +1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(-1.0, -1.0, +1.0);
				// rechts
				glVertex3f(+1.0, -1.0, +1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(+1.0, +1.0, +1.0);
				// vorne
				glColor3ub(CLIP(r+25, 0, 255), CLIP(g+25, 0, 255), CLIP(b+25, 0, 255));
				glVertex3f(+1.0, +1.0, +1.0);
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				glVertex3f(-1.0, +1.0, +1.0);
				// hinten
				glVertex3f(-1.0, -1.0, +1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(+1.0, -1.0, +1.0);
			glEnd();

			glPopMatrix();
		}
	}

	glutSwapBuffers();
}

void init(char* bmpFileName) {
	for (int h=0; h<SIZE_HSL; h++)
	 for (int s=0; s<SIZE_HSL; s++)
	  for (int l=0; l<SIZE_HSL; l++) hist[h][s][l] = 0;

	MBmpImg* rgbImg = new MBmpImg(bmpFileName);
	bmpImg = new MBmpImg(rgbImg->getWidth(), rgbImg->getHeight(), 24);
	rgbImg->RGB2HSL(bmpImg);

bmpImg->HSL2RGB(rgbImg);
bmpImg->write2BmpFile("rgbouthsl.bmp");
rgbImg->write2BmpFile("rgbout2.bmp");

	delete rgbImg;

	if (bmpImg->getWidth() == 0) {
		printf("error while trying to read bitmap file\n");
		delete bmpImg;
		exit(0);
	}

	BYTE* tmpSourcePtr = bmpImg->getData();
	for (int y=0; y<bmpImg->getHeight(); y++) {
		for (int x=0; x<bmpImg->getWidth(); x++) {
			unsigned char h = *tmpSourcePtr++;
			unsigned char s = *tmpSourcePtr++;
			unsigned char l = *tmpSourcePtr++;

			h >>= SHIFT_HSL;
			s >>= SHIFT_HSL;
			l >>= SHIFT_HSL;

			hist[h][s][l]++;
			if (hist[h][s][l] > max) max = hist[h][s][l];
		}
	}

	calcScaleCube();

	delete bmpImg;

/*
	// lighting
	GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
	//	light_position is NOT default value
	GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
*/
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	// init
	glClearColor (0.0, 0.0, 0.0, 0.0);
}

// Buchstaben; einige andere Tasten
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
		case '+':
			scCuInt++;
			calcScaleCube();
			glutPostRedisplay();
			break;
		case '-':
			scCuInt--;
			calcScaleCube();
			glutPostRedisplay();
			break;
		case 27: // escape
		case 'q': exit(0);
   }
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	switch (actButton) {
		case GLUT_LEFT_BUTTON:
			q.rotate((x - oldMouseX) * 0.01, 0.0, -1.0, 0.0);
			q.rotate((y - oldMouseY) * 0.01, 1.0, 0.0, 0.0);
			break;
		case GLUT_MIDDLE_BUTTON:
			scCuInt += (y - oldMouseY);
			calcScaleCube();
			break;
	}

	oldMouseX = x;
	oldMouseY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y) {
	if (state == GLUT_DOWN) actButton = button;
	oldMouseX = x;
	oldMouseY = y;
}

void reshape(GLsizei width, GLsizei height) {
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0, // field of view in degree
		(GLfloat)(width)/(GLfloat)(height), // aspect ratio
		0.01, 50.0); // z near, z far
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

int main(int argc, char* argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(40, 40);
	glutCreateWindow("HSL Histogram");
	glutKeyboardFunc(keyboardFunc);
//	glutSpecialFunc(keyboardSpecialFunc);
	glutReshapeFunc(reshape);
	glutMotionFunc(mouseDrag);
	glutMouseFunc(mouseButton);
//	glutIdleFunc(update);

	if (argc > 1) init(argv[1]); else {
		printf("parameter: bitmap file\n");
		exit(0);
	}

	glutDisplayFunc(display);
	glutMainLoop();

	return 0;   /* ANSI C requires main to return int. */
}
