/**
 * Vector (4) and Matrix (4x4) functions
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *			00 01 02 03
 * Matrix	10 11 12 13
 *			20 21 22 23
 *			30 31 32 33
 *
 * is declared like that:
 * float a[3][3] = {{00, 01, 02, 03},
 *					 {10, 11, 12, 13},
 *					 {20, 21, 22, 23},
 *					 {30, 31, 32, 33}};
 *
 * access: a[row][column]
 * (d.h.: a[Zeile][Spalte])
 *
 * (a in memory: 00, 01, 02, 03, 10, 11, 12, 13, 20, 21, 22, 23, 30, 31, 32, 33)
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef MATH3D_HEADER_INCLUDED
#define MATH3D_HEADER_INCLUDED

struct VEC4 {
	float v[4];
};

struct MAT4x4 {
	float m[4][4];
};

void resetVec(VEC4* vPtr);
void resetMat(MAT4x4* mPtr);
void copyVec(VEC4* vSrcPtrfloat, VEC4* vDestPtrfloat); // vDest = vSrc
void copyMat(MAT4x4* mSrcPtrfloatfloat, MAT4x4* mDestPtrfloatfloat); // mDest = mSrc

void addVec(VEC4* a, VEC4* b); // a = a + b
void subVec(VEC4* a, VEC4* b); // a = a - b
void scaleVec(VEC4* a, float scale); // a = a * f
float getVec3Length(VEC4* a);
void normalizeVec3(VEC4* a);
float scalarMultVec3(VEC4* a, VEC4* b); // a * b'

void rotateX(MAT4x4* mPtr, float angle); // m = mRot * m
void rotateY(MAT4x4* mPtr, float angle); // m = mRot * m
void rotateZ(MAT4x4* mPtr, float angle); // m = mRot * m
void buildRotationMatrix(MAT4x4* mPtr, VEC4* axis, float angle); // axis is rotation unit vector
void translate(MAT4x4* mPtr, VEC4* vPtrfloat);
void multMatVec(MAT4x4* mPtr, VEC4* vPtrfloat); // vec = mat * vec
void multMatMat(MAT4x4* m1Ptr, MAT4x4* m2Ptr); // m1 = m1 * m2
void multMatMatInv(MAT4x4* m1Ptr, MAT4x4* m2Ptr); // m1 = m2 * m1
void getInvRotMat(MAT4x4* src, MAT4x4* dest);

#endif // MATH3D_HEADER_INCLUDED
