#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "QuaternionRotation.h"
#include "MBmpImg.h"

MBmpImg* bmpImg;

QuaternionRotation q;
int oldMouseX, oldMouseY;

#define SIZE_RGB 16
// SHIFT_RGB = ld(256/SIZE_RGB)
#define SHIFT_RGB 4

unsigned int hist[SIZE_RGB][SIZE_RGB][SIZE_RGB]; // [r][g][b]
unsigned int max = 0;
float scaleCube;
int scCuInt = 20;

int actButton;

#define SCCU_MIN -200
#define SCCU_MAX 400

void calcScaleCube() {
	if (scCuInt < SCCU_MIN) scCuInt = SCCU_MIN;
	if (scCuInt > SCCU_MAX) scCuInt = SCCU_MAX;
	scaleCube = 2.0 * pow(2.0, scCuInt/20.0) / max;
}

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0, 0, -3);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

	float st = 1 / (float) SIZE_RGB;

	glColor3f(0.6, 0.6, 0.6);
	glutWireCube(2.0);

	// draw red, green, blue, black, white boxes
	glPushMatrix();
		glTranslatef(-1, -1, -1);
		glColor3f(0.0, 0.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		glTranslatef(1, -1, -1);
		glColor3f(1.0, 0.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		glTranslatef(-1, 1, -1);
		glColor3f(0.0, 1.0, 0.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		glTranslatef(-1, -1, 1);
		glColor3f(0.0, 0.0, 1.0);
		glutSolidCube(0.05);
	glPopMatrix();
	glPushMatrix();
		glTranslatef(1, 1, 1);
		glColor3f(1.0, 1.0, 1.0);
		glutSolidCube(0.05);
	glPopMatrix();

	glTranslatef(-1, -1, -1);

	for (int r=0; r<SIZE_RGB; r++)
	 for (int g=0; g<SIZE_RGB; g++)
	  for (int b=0; b<SIZE_RGB; b++) {
		if (hist[r][g][b]) {
			glPushMatrix();

			// translate and scale
			glTranslatef((2*r+1)*st, (2*g+1)*st, (2*b+1)*st);
			float sc = scaleCube * hist[r][g][b];
			if (sc < 0.05) sc = 0.05;
			if (sc > 0.8) sc = 0.8;
			sc *= st;
			glScalef(sc, sc, sc);

			// draw colored cube
			glBegin(GL_QUADS);
				// oben
				glColor3f(r*st, g*st, b*st);
				glVertex3f(+1.0, +1.0, +1.0);
				glVertex3f(-1.0, +1.0, +1.0);
				glVertex3f(-1.0, -1.0, +1.0);
				glVertex3f(+1.0, -1.0, +1.0);
				// unten
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				// links
				glColor3f(r*st-0.1, g*st-0.1, b*st-0.1);
				glVertex3f(-1.0, +1.0, +1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(-1.0, -1.0, +1.0);
				// rechts
				glVertex3f(+1.0, -1.0, +1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(+1.0, +1.0, +1.0);
				// vorne
				glColor3f(r*st+0.1, g*st+0.1, b*st+0.1);
				glVertex3f(+1.0, +1.0, +1.0);
				glVertex3f(+1.0, +1.0, -1.0);
				glVertex3f(-1.0, +1.0, -1.0);
				glVertex3f(-1.0, +1.0, +1.0);
				// hinten
				glVertex3f(-1.0, -1.0, +1.0);
				glVertex3f(-1.0, -1.0, -1.0);
				glVertex3f(+1.0, -1.0, -1.0);
				glVertex3f(+1.0, -1.0, +1.0);
			glEnd();

			glPopMatrix();
		}
	}

	glutSwapBuffers();
}

void init(char* bmpFileName) {
	for (int r=0; r<SIZE_RGB; r++)
	 for (int g=0; g<SIZE_RGB; g++)
	  for (int b=0; b<SIZE_RGB; b++) hist[r][g][b] = 0;

	bmpImg = new MBmpImg(bmpFileName);
	if (bmpImg->getWidth() == 0) {
		printf("error while trying to read bitmap file\n");
		delete bmpImg;
		exit(0);
	}

	BYTE* tmpSourcePtr = bmpImg->getData();
	for (int y=0; y<bmpImg->getHeight(); y++) {
		for (int x=0; x<bmpImg->getWidth(); x++) {
			unsigned char r = *tmpSourcePtr++;
			unsigned char g = *tmpSourcePtr++;
			unsigned char b = *tmpSourcePtr++;

			b >>= SHIFT_RGB;
			g >>= SHIFT_RGB;
			r >>= SHIFT_RGB;

			hist[r][g][b]++;
			if (hist[r][g][b] > max) max = hist[r][g][b];
		}
	}

	calcScaleCube();

	delete bmpImg;

/*
	// lighting
	GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
	//	light_position is NOT default value
	GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
*/
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	// init
	glClearColor (0.0, 0.0, 0.0, 0.0);
}

// Buchstaben; einige andere Tasten
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
		case '+':
			scCuInt++;
			calcScaleCube();
			glutPostRedisplay();
			break;
		case '-':
			scCuInt--;
			calcScaleCube();
			glutPostRedisplay();
			break;
		case 27: // escape
		case 'q': exit(0);
   }
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	switch (actButton) {
		case GLUT_LEFT_BUTTON:
			q.rotate((x - oldMouseX) * 0.01, 0.0, -1.0, 0.0);
			q.rotate((y - oldMouseY) * 0.01, 1.0, 0.0, 0.0);
			break;
		case GLUT_MIDDLE_BUTTON:
			scCuInt += (y - oldMouseY);
			calcScaleCube();
			break;
	}

	oldMouseX = x;
	oldMouseY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y) {
	if (state == GLUT_DOWN) actButton = button;
	oldMouseX = x;
	oldMouseY = y;
}

void reshape(GLsizei width, GLsizei height) {
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0, // field of view in degree
		(GLfloat)(width)/(GLfloat)(height), // aspect ratio
		0.01, 50.0); // z near, z far
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

int main(int argc, char* argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(40, 40);
	glutCreateWindow("RGB Histogram");
	glutKeyboardFunc(keyboardFunc);
//	glutSpecialFunc(keyboardSpecialFunc);
	glutReshapeFunc(reshape);
	glutMotionFunc(mouseDrag);
	glutMouseFunc(mouseButton);
//	glutIdleFunc(update);

	if (argc > 1) init(argv[1]); else {
		printf("parameter: bitmap file\n");
		exit(0);
	}

	glutDisplayFunc(display);
	glutMainLoop();

	return 0;   /* ANSI C requires main to return int. */
}
