#include <stdio.h>
#include <GL/glut.h>
#include "MarchingCubes.h"
#include "QuaternionRotation.h"

#define SIZE_X 30
#define SIZE_Y 30
#define SIZE_Z 30

#define sqr(x) ((x)*(x))

QuaternionRotation q;
int oldMouseX, oldMouseY;

MarchingCubes mc;
char slices[2][SIZE_Y * SIZE_X];

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0, 0, -5);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

	glColor3f(0.0, 0.0, 0.0);
//	glutSolidCube(1.0);

//	mc.draw();
	mc.drawFlatShaded();

glDisable(GL_LIGHTING);
glutWireCube(2);
glEnable(GL_LIGHTING);

	for (int z=0; z<2; z++)
		for (int y=0; y<2; y++)
			for (int x=0; x<2; x++) if (slices[z % 2][y*2 + x] > 0) {
				glPushMatrix();
				glTranslatef((x==0)?-1:1, (y==0)?-1:1, (z==0)?-1:1);
				glutSolidSphere(0.2, 6, 6);
				glPopMatrix();
			}

	glutSwapBuffers();
}

void init (void) {

	// lighting
	GLfloat light_ambient[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };

	GLfloat light_position[] = { 5.0, 2.0, 0.5, 0.0 };

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	// init
	glClearColor (1.0, 1.0, 1.0, 0.0);

	glMatrixMode(GL_PROJECTION);
	gluPerspective( /* field of view in degree */ 60.0,
	/* aspect ratio */ 1.0,
	/* Z near */ 0.1, /* Z far */ 20.0);
	glMatrixMode(GL_MODELVIEW);


// fr Wireframe:
//    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); // GL_FILL
}

// Buchstaben; einige andere Tasten
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
		case '+':
			q.rotate(-0.1, 0.0, 1.0, 0.0);
			glutPostRedisplay();
			break;
		case '-':
			q.rotate( 0.1, 0.0, 1.0, 0.0);
			glutPostRedisplay();
			break;
		case 27:
		case 'q': exit(0);
   }
}

// Sondertasten
void keyboardSpecialFunc(int key, int x, int y) {
	switch (key) {
		case GLUT_KEY_LEFT:
			q.rotate( 0.1, 0.0, 1.0, 0.0);
			glutPostRedisplay();
			break;
		case GLUT_KEY_RIGHT:
			q.rotate(-0.1, 0.0, 1.0, 0.0);
			glutPostRedisplay();
			break;
		case GLUT_KEY_UP:
			q.rotate(-0.1, 1.0, 0.0, 0.0);
			glutPostRedisplay();
			break;
		case GLUT_KEY_DOWN:
			q.rotate( 0.1, 1.0, 0.0, 0.0);
			glutPostRedisplay();
			break;
	}
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	q.rotate((x - oldMouseX) * 0.01, 0.0, -1.0, 0.0);
	q.rotate((y - oldMouseY) * 0.01, 1.0, 0.0, 0.0);

	oldMouseX = x;
	oldMouseY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y) {
	oldMouseX = x;
	oldMouseY = y;
}

int main(int argc, char** argv) {
	printf("mc.init():\n");

	mc.init(SIZE_X, SIZE_Y, SIZE_Z);

	printf("filling data array:\n");
	for (int z=0; z<SIZE_Z; z++) {
		// fill data array
		for (int y=0; y<SIZE_Y; y++) {
			for (int x=0; x<SIZE_X; x++) {
/*
				int v = sqr(x - SIZE_X/2) + sqr(y - SIZE_Y/2) + sqr(z - SIZE_Z/2);
				v -= (int) (0.7*sqr(SIZE_X/2));
				v /= -2;
*/
int v = -100;
if ((x>0) && (x<SIZE_X-1) && (y>0) && (y<SIZE_Y-1) && (z>0) && (z<SIZE_Z-1)) v = rand()%200-100;
				if (v < -100) v = -100;
				if (v > 100) v = 100;
				slices[z % 2][y*SIZE_X + x] = v;
				if (v == 0) printf("O"); else printf((v>0)?"+":"-");
			}
			printf("\n");
		}
		if (z > 0) {		
			printf("processing slice %i - %i ...", z-1, z);
printf("\n");
			mc.processSlice(z-1, slices[(z-1) % 2], slices[z % 2]);
			printf(" finished (num coords: %i, num tris: %i)\n", mc.coords.size(), mc.triangles.size());
			//system("PAUSE");
		}
	}
	mc.terminate(); // to be called after the loop with processSlice


	printf("number of triangles: %i\n", mc.triangles.size());
	printf("number of coords: %i\n", mc.coords.size());

	if (mc.write2ObjFile("testMarchingCubes.obj")) {
		printf("obj file written successfully\n");
	} else {
		printf("obj file could not be written\n");
	}

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(450, 450);
	glutInitWindowPosition(10, 10);
	glutCreateWindow("Quaternion rotation demo");
	glutKeyboardFunc(keyboardFunc);
	glutSpecialFunc(keyboardSpecialFunc);
	glutMotionFunc(mouseDrag);
	glutMouseFunc(mouseButton);
	init();
	glutDisplayFunc(display); 
	glutMainLoop();
	return 0; // ANSI C requires main to return int
}
