// displays a bitmap image as gray level mountain


#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "QuaternionRotation.h"
#include "MBmpImg.h"
#include "mathext.h"

MBmpImg* bmpImg;

QuaternionRotation q;
int oldMouseX, oldMouseY;
int actButton;

int imgWidth, imgHeight;
float* coordPtr;
BYTE* colorPtr;
BYTE* lightingPtr;

bool showColor, showLighting, showFlat;

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0, 0, -1.5);

	// apply Quaternion rotation with OpenGL matrix multiplication
	float m[4][4];
	q.getRotMatrix(m);
	glMultMatrixf(&m[0][0]);

	// display bitmap image as gray level mountain
//showLighting
//	glEnable(GL_COLOR_MATERIAL);
	if ((!showColor) && (!showLighting)) {
		glColor3f(0.7, 0.7, 0.7);
	} else {
		if (showColor) glColorPointer(3, GL_UNSIGNED_BYTE, 0, &colorPtr[0]); // 3 colors, type (color format), stride = 0 (-> tightly packed), data pointer
			else glColorPointer(3, GL_UNSIGNED_BYTE, 0, &lightingPtr[0]); // 3 colors, type (color format), stride = 0 (-> tightly packed), data pointer
		glEnableClientState(GL_COLOR_ARRAY);
	}
	glVertexPointer(3, GL_FLOAT, 0, &coordPtr[0]); // 3 coords, type, stride = 0 (-> tightly packed), data pointer
	glEnableClientState(GL_VERTEX_ARRAY);

	if (showFlat) glScalef(1.0, 1.0, 0.0);

	glBegin(GL_TRIANGLES);
	for (int y=0; y<imgHeight-1; y++)
		for (int x=0; x<imgWidth-1; x++) {
//			glNormal3fv(&faceNormals[t].n[0]);
			glArrayElement(x + y*imgWidth);
			glArrayElement(x+1 + y*imgWidth);
			glArrayElement(x+1 + (y+1)*imgWidth);

			glArrayElement(x+1 + (y+1)*imgWidth);
			glArrayElement(x + (y+1)*imgWidth);
			glArrayElement(x + y*imgWidth);
		}
	glEnd();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

//	glDisable(GL_COLOR_MATERIAL);

	glutSwapBuffers();
}

void init(char* bmpFileName) {
	MBmpImg* grayImg = new MBmpImg(bmpFileName);

	imgWidth = grayImg->getWidth();
	imgHeight = grayImg->getHeight();

	if ((imgWidth == 0) || (grayImg->getBitsPerPixel() != 8)) {
		printf("error while trying to read bitmap file (only gray level images allowed!)\n");
		delete bmpImg;
		exit(0);
	}

	coordPtr = new float[imgWidth * imgHeight * 3];
	colorPtr = new BYTE[imgWidth * imgHeight * 3];
	lightingPtr = new BYTE[imgWidth * imgHeight * 3];

	int maxdim = (imgWidth > imgHeight) ? imgWidth : imgHeight;
	int ind = 0;
	for (int y=0; y<imgHeight; y++)
		for (int x=0; x<imgWidth; x++) {
			BYTE col = *grayImg->getDataXY(imgWidth-x-1, y);
			int l = *grayImg->getDataXY(CLIP(imgWidth-x-2, 0, imgWidth-1), y) + *grayImg->getDataXY(imgWidth-x-1, CLIP(y+1, 0, imgHeight-1))
   					- *grayImg->getDataXY(CLIP(imgWidth-x, 0, imgWidth-1), y) - *grayImg->getDataXY(imgWidth-x-1, CLIP(y-1, 0, imgHeight-1));
   			l = CLIP(l/2 + 128, 0, 255);
			colorPtr[ind] = col;
			lightingPtr[ind] = l;
			coordPtr[ind++] = (x - imgWidth/2) / (float) maxdim;
			colorPtr[ind] = col;
			lightingPtr[ind] = l;
			coordPtr[ind++] = (y - imgHeight/2) / (float) maxdim;
			colorPtr[ind] = col;
			lightingPtr[ind] = l;
			coordPtr[ind++] = col * 0.001;
		}

	delete bmpImg;

/*
	// lighting
	GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
	//	light_position is NOT default value
	GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
*/
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	// init
	glClearColor (0.0, 0.0, 0.4, 0.0);

	showColor = true;
	showLighting = false;
	showFlat = false;
}

// Buchstaben; einige andere Tasten
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
/*		case '+':
			glutPostRedisplay();
			break;
		case '-':
			glutPostRedisplay();
			break;
*/
		case 'c':
			showColor = !showColor;
			glutPostRedisplay();
			break;
		case 'f':
			showFlat = !showFlat;
			glutPostRedisplay();
			break;
		case 'l':
			showLighting= !showLighting;
			glutPostRedisplay();
			break;

		case 27: // escape
		case 'q': exit(0);
   }
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	switch (actButton) {
		case GLUT_LEFT_BUTTON:
			q.rotate((x - oldMouseX) * 0.01, 0.0, 1.0, 0.0);
			q.rotate((y - oldMouseY) * 0.01, 1.0, 0.0, 0.0);
			break;
/*		case GLUT_MIDDLE_BUTTON:
			break;
*/	}

	oldMouseX = x;
	oldMouseY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y) {
	if (state == GLUT_DOWN) actButton = button;
	oldMouseX = x;
	oldMouseY = y;
}

void reshape(GLsizei width, GLsizei height) {
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0, // field of view in degree
		(GLfloat)(width)/(GLfloat)(height), // aspect ratio
		0.01, 10.0); // z near, z far
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

int main(int argc, char* argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(40, 40);
	glutCreateWindow("gray level mountain");
	glutKeyboardFunc(keyboardFunc);
//	glutSpecialFunc(keyboardSpecialFunc);
	glutReshapeFunc(reshape);
	glutMotionFunc(mouseDrag);
	glutMouseFunc(mouseButton);
//	glutIdleFunc(update);

	if (argc > 1) init(argv[1]); else {
		printf("parameter: bitmap file\n");
		exit(0);
	}

	glutDisplayFunc(display);
	glutMainLoop();

	// clean up
	delete[] coordPtr;
	delete[] colorPtr;
	delete[] lightingPtr;

	return 0;   /* ANSI C requires main to return int. */
}
